
	//package org.bhr.project;

	/** 
	 * This Class contains methods for storing and retrieving a 4*4 matrix 
	 * of joint probability distribution values along each branch
	 *
	 * @author Vivek Jayaswal
	 *
	 */
	 
	/**
	 * Mar. 26, 2008 
	 * add two vectors to hold the frequencies for two ends
	 */
	public class BranchDetails {	
		double[][] jointProbability = new double[4][4];
		/**
		 * add on Mar.26, 2008
		 * parent: rows
		 * child: columns
		 */
		double[] frequencyParent = new double[4];
		double[] frequencyChild  = new double[4];

		BranchDetails() {
			
		}
		
		/**
		 * add in Feb, 2010
		 * @param d
		 */
		BranchDetails(BranchDetails inputObj)
		{
			for(int i=0; i<4; i++) 
			{
				for(int j=0; j<4; j++)
				{
					jointProbability[i][j] = inputObj.jointProbability[i][j];
				}
				frequencyChild[i] = inputObj.frequencyChild[i];
				frequencyParent[i] = inputObj.frequencyParent[i];
			}			

		}
		
		BranchDetails(double[][] d) {
			for(int i=0; i<4; i++) {
				for(int j=0; j<4; j++) {
					jointProbability[i][j] = d[i][j];
				}
			}			
		}
		/**
		 * add on Mar.26, 2008
		 * @param d
		 * @param p
		 * @param c
		 */
		BranchDetails(double[][] d, double[] p, double[] c)
		{
			int j, k;
			
			for(j=0; j<4; j++) {
				for(k=0; k<4; k++) {
					jointProbability[j][k] = d[j][k];
				}
			}			
			
			for(j=0; j<4; j++)
			{
				frequencyParent[j] = p[j];
				frequencyChild[j]  = c[j];
			}
		}
		/**
		 * add on Mar. 26, 2008
		 * @param p
		 * @param c
		 */
		BranchDetails(double[] p, double[] c)
		{
			int j;
			
			for(j=0; j<4; j++)
			{
				frequencyChild[j] = c[j];
				frequencyParent[j] = p[j];
			}
		}
		
		/* Rows represent values at nodeA's parent and the columns represent
		 * values at nodeA
		 */
		public void setJointProbability() 
		{
			int j, k;
			for(j=0; j<4; j++) {
				for(k=0; k<4; k++) {
					if(j == k) 
						jointProbability[j][k] = 1.0/8.0;
					else 
						jointProbability[j][k] = 1.0/24.0;
				}
			}
			
			/*double[] initialValues= {1.0/8.0, 2.0/48.0, 1.0/48.0, 3.0/48.0, 
					                 1.0/48.0, 1.0/8.0, 3.0/48.0, 2.0/48.0, 
					                 2.0/48.0, 3.0/48.0, 1.0/8.0, 3.0/48.0, 
					                 2.0/48.0, 1.0/48.0, 1.0/ 48, 1.0/8.0};*/
			/*int counter=0; 
			for(int j = 0; j<4; j++)
			{
				for(int k=0; k<4; k++)
				{
					jointProbability[j][k] = initialValues[counter++]; 
				}
			}*/
		}
	/**
	 * added on Mar.26, 2008
	 * @param dummy
	 */	
		public void setJointProbability(String dummy) 
		{
			int j, k;
			for(j=0; j<4; j++) {
				for(k=0; k<4; k++) {
					if(j == k) 
						jointProbability[j][k] = 1.0/8.0;
					else 
						jointProbability[j][k] = 1.0/24.0;
				}
			}
			
			for(j=0; j<4; j++)
			{
				frequencyChild[j] = sumOverParent(j);
				frequencyParent[j] = sumOverChild(j);
			}
			/*double[] initialValues= {1.0/8.0, 2.0/48.0, 1.0/48.0, 3.0/48.0, 
					                 1.0/48.0, 1.0/8.0, 3.0/48.0, 2.0/48.0, 
					                 2.0/48.0, 3.0/48.0, 1.0/8.0, 3.0/48.0, 
					                 2.0/48.0, 1.0/48.0, 1.0/ 48, 1.0/8.0};*/
			/*int counter=0; 
			for(int j = 0; j<4; j++)
			{
				for(int k=0; k<4; k++)
				{
					jointProbability[j][k] = initialValues[counter++]; 
				}
			}*/
		}
		/**
		 * add on Mar.26, 2008
		 * @param parentIndex
		 * @return
		 * compute the frequency for a character base of the parent
		 */
		double sumOverChild(int parentIndex)
		{
			int j;
			double result = 0.0;
			
			for(j=0; j<4; j++)
				result += jointProbability[parentIndex][j];
			return result;
		}
		/**
		 * add on Mar.26, 2008
		 * @param d
		 * @param parentIndex
		 * @return
		 */
		double sumOverChild(double d[][], int parentIndex)
		{
			int j;
			double result = 0.0;
			
			for(j=0; j<4; j++)
				result += d[parentIndex][j];
			return result;
		}
		/**
		 * add on Mar.26, 2008
		 * @param d
		 * @param childIndex
		 * @return
		 */
		double sumOverParent(double d[][], int childIndex)
		{
			int j;
			double result = 0.0;
			
			for(j=0; j<4; j++)
				result += d[j][childIndex];
			return result;
		}
		
		/**
		 * add on Mar.26, 2008
		 * @param childIndex
		 * @return
		 */
		double sumOverParent(int childIndex)
		{
			int j;
			double result = 0.0;
			
			for(j=0; j<4; j++)
				result += jointProbability[j][childIndex];
			return result;
		}
		
		public void setJointProbability(int x0, int x1, double sum) {
			jointProbability[x0][x1] = sum;
		}
		/**
		 * add on Mar.26, 2008
		 * @param index
		 * @param value
		 */
		public void setFrequencyParent(int index, double value) {
			frequencyParent[index] = value;
		}
		/**
		 * add on Mar.26, 2008
		 * @param index
		 * @param value
		 */
		public void setFrequencyChild(int index, double value) {
			frequencyChild[index] = value;
		}
		
		public void setJointProbability(double d[][]) 
		{
			for(int i=0; i<4; i++) {
				for(int j=0; j<4; j++) {
					jointProbability[i][j] = d[i][j];
				}
			}	
		}	
		/**
		 * add on Mar.26, 2008
		 * @param d
		 */
		public void setFrequencyParent(double d[][]) 
		{
			for(int j=0; j<4; j++)
			{
				frequencyParent[j] = sumOverChild(d, j);
			}	
		}	
		
		public void setFrequencyChild(double d[][]) {
			for(int j=0; j<4; j++)
			{
				frequencyChild[j] = sumOverParent(d, j);
			}	
		}		
		/**
		 * add on Mar.26, 2008
		 * @param p
		 */
		public void setFrequencyParent(double p[]) {
			for(int j=0; j<4; j++)
			{
				frequencyParent[j] = p[j];
			}	
		}		
		/**
		 * add on Mar.26, 2008
		 * @param c
		 */
		public void setFrequencyChild(double c[]) {
			for(int j=0; j<4; j++)
			{
				frequencyChild[j] = c[j];
			}	
		}		
		public void setJointProbability(double d[][], double[] lagrange) {
			for(int i=0; i<4; i++) {
				for(int j=0; j<4; j++) {
					jointProbability[i][j] = d[i][j];
				}
			}
			
		}
		
		public double[][] getJointProbability() {
			return jointProbability;
		}
		/**
		 * add on Mar.26, 2008
		 * @return
		 */
		public double[] getParentFrequency() {
			return frequencyParent;
		}
		
		public double[] getChildFrequency() {
			return frequencyChild;
		}
		public void transposeJointProbability() {
			double[][] temp = new double[4][4];
			
			for(int i=0; i<4; i++) {
				for(int j=0; j<4; j++) {
					temp[i][j] = jointProbability[j][i];
				}
			}
			
			jointProbability = temp;
		}
		
		public double computeSS(BranchDetails inputObj)
		{
			double ss = 0.0;
			
			for(int i=0; i<4; i++) 
			{
				for(int j=0; j<4; j++)
				{
					ss = ss + Math.pow((jointProbability[i][j] - inputObj.jointProbability[i][j]), 2.0);
				}
				//ss = ss + Math.pow((frequencyChild[i] - inputObj.frequencyChild[i]), 2);
				//ss = ss + Math.pow((frequencyParent[i] - inputObj.frequencyParent[i]), 2);
			}			
			return ss;
		}

}
