
public class Node 
{
	String nodeName;
	String leftChild;
	String rightChild;
	String parent;
	String nodeType;
	boolean leftVisited;
	boolean rightVisited;
	/**
	 * "-1": the node is going to be the root
	 * "0":  from parent to child
	 * "1":  from child to parent
	 */
	int evolutionaryDirection;
	int currentRowPermutationIndex;
	int previousRowPermutationIndex;
	int currentColumnPermutationIndex;
	int previousColumnPermutationIndex;
	
	double currentCumulativeDifs;
	double previousCumulativeDifs;
	
	double[] difs;
	
	boolean isUpdated;
	
	public Node() 
	{
		nodeName = null;
		leftChild = null;
		rightChild = null;
		parent = null;
		nodeType = null;
		leftVisited = false;
		rightVisited = false;
		
		evolutionaryDirection = 999999;
		currentRowPermutationIndex = 0;
		previousRowPermutationIndex = 0;
		currentColumnPermutationIndex = 0;
		previousColumnPermutationIndex = 0;
		
		currentCumulativeDifs = 999999999.0;
		previousCumulativeDifs = 999999999.0;
		
		isUpdated = false;
		difs = null;
	}
	
	/**
	 * copy constructure
	 * @param inputNode
	 */
	public Node(Node inputNode, int numOfDifs)
	{
		nodeName = inputNode.getNodeName();
		leftChild = inputNode.getLeftChild();
		rightChild = inputNode.getRightChild();
		parent = inputNode.getParent();
		nodeType = inputNode.getNodeType();
		
		/**
		 * "-1": the branch is going to put the root
		 * "0":  from parent to child
		 * "1":  from child to parent
		 */
		evolutionaryDirection = inputNode.getEvolutionaryDirection();
		currentRowPermutationIndex = inputNode.getCurrentRowPermutationIndex();
		previousRowPermutationIndex = inputNode.getPreviousRowPermutationIndex();
		currentColumnPermutationIndex = inputNode.getCurrentColumnPermutationIndex();
		previousColumnPermutationIndex = inputNode.getPreviousColumnPermutationIndex();
		
		currentCumulativeDifs = inputNode.getCurrentCumulativeDifs();
		previousCumulativeDifs = inputNode.getPreviousCumulativeDifs();
		isUpdated = inputNode.getIsUpdate();
		
		difs = null;
		//difs = new double[numOfDifs];
		
		//for(int j=0; j<numOfDifs; j++)
			//difs[j] = inputNode.getDifs()[j];
	}
	/** 
	 * set functions
	 */
	void setNodeName(String inputString)
	{
		nodeName = inputString;
	}
	
	void setLeftChild(String inputString)
	{
		leftChild = inputString;
	}
	
	void setRightChild(String inputString)
	{
		rightChild = inputString;
	}
	
	void setParent(String inputString)
	{
		parent = inputString;
	}
	
	void setNodeType(String inputString)
	{
		nodeType = inputString;
	}
	
	void setEvolutionaryDirection(int inputDirection)
	{
		evolutionaryDirection = inputDirection;
	}
	
	void setCurrentRowPermutationIndex(int inputIndex)
	{
		currentRowPermutationIndex = inputIndex;
	}
	void setPreviousRowPermutationIndex(int inputIndex)
	{
		previousRowPermutationIndex = inputIndex;
	}
	
	void setCurrentColumnPermutationIndex(int inputIndex)
	{
		currentColumnPermutationIndex = inputIndex;
	}
	void setPreviousColumnPermutationIndex(int inputIndex)
	{
		previousColumnPermutationIndex = inputIndex;
	}	
	
	void setCurrentCumulativeDif(double inputDifs)
	{
		currentCumulativeDifs = inputDifs;
	}
	
	void setPreviousCumulativeDif(double inputDifs)
	{
		previousCumulativeDifs = inputDifs;
	}

	void setIsUpdated(boolean inputIsUpdate)
	{
		isUpdated = inputIsUpdate;
	}
	
	void setDifs(double[] inputDifs, int inputLength)
	{
		int j;
		
		if(difs == null)
			difs = new double[inputLength];
		for(j=0; j<inputLength; j++)
			difs[j] = inputDifs[j];
	}
	/**
	 * get functions
	 */
	String getNodeName()
	{
		return nodeName;
	}
	
	String getLeftChild()
	{
		return leftChild;
	}
	
	String getRightChild()
	{
		return rightChild;
	}
	
	String getParent()
	{
		return parent;
	}
	
	String getNodeType()
	{
		return nodeType;
	}
	
	int getEvolutionaryDirection()
	{
		return evolutionaryDirection;
	}
	
	int getCurrentRowPermutationIndex()
	{
		return currentRowPermutationIndex;
	}
	
	int getPreviousRowPermutationIndex()
	{
		return previousRowPermutationIndex;
	}

	int getCurrentColumnPermutationIndex()
	{
		return currentColumnPermutationIndex;
	}
	
	int getPreviousColumnPermutationIndex()
	{
		return previousColumnPermutationIndex;
	}

	double getCurrentCumulativeDifs()
	{
		return currentCumulativeDifs;
	}
	
	double getPreviousCumulativeDifs()
	{
		return previousCumulativeDifs;
	}
	
	boolean getIsUpdate()
	{
		return isUpdated;
	}
	
	double[] getDifs()
	{
		return difs;
	}
	
	boolean getLeftVisited()
	{
		return leftVisited;
	}

	boolean getRightVisited()
	{
		return rightVisited;
	}

	void setLeftVisited(boolean inputValue)
	{
		leftVisited = inputValue;
	}

	void setRightVisited(boolean inputValue)
	{
		rightVisited = inputValue;
	}
	

	@Override
	public String toString() {
		String results = "";
		results += "node name: \t" + nodeName + "\n";
		results += "left child: \t" + leftChild + "\n";
		results += "right child: \t" + rightChild + "\n";
		results += "parent: \t" + parent + "\n";
		results += "node type: \t" + nodeType + "\n";
		
		/**
		 * "-1": the branch is going to put the root
		 * "0":  from parent to child
		 * "1":  from child to parent
		 */
		results += "evolutionary direction: \t" + evolutionaryDirection + "\n";
		results += "current row permutation index: \t" + currentRowPermutationIndex + "\n";
		results += "previous row permutaiton index: \t" + previousRowPermutationIndex + "\n";
		results += "current column permutation index: \t" + currentColumnPermutationIndex + "\n";
		results += "previous column permutation index: \t" + previousColumnPermutationIndex + "\n";
		
		results += "current cumulative difs: \t" + currentCumulativeDifs + "\n";
		results += "previous cumulative difs: \t: " + previousCumulativeDifs + "\n";
		
		results +=  "is updated: \t" + isUpdated + "\n";

		return results;
	}
}
