import java.util.*;

public class SearchShortestPath 
{
	TreeMap<String, Node> reconstructedTree    = null;
	ArrayList<ArrayList<String>> treeLevelList = null;
	ArrayList<String> intNodeList =null;
	
	int defaultNumberOfPermutations = 24;
	int itrCounter                  = 0; 
	boolean isInternalNode          = false;
	boolean isRootParentOldRoot     = false;
	
	String rootNode                 = null;
	String oldRootNode              = null;
	
	TreeMap<String, double[]> differences = null;
	
	/**
	 * add on Mar.28, 2009
	 * for handling rooting branch
	 */
	ArrayList<double[]> rootingBranchDifs = null;
	/** 
	 * this container has two elements
	 * 0: newRoot to the branch representative
	 * 1: newRoot to the parent of the branch representative
	 */
	double[] minSSOfRootingBranch = null;
	/**
	 * added on April 10, 2009
	 */
	ArrayList<ArrayList<String>> treeLevelListSubtree2 = null;
	/**
	 * 0: node for subtree 1
	 * 1: node for subtree 2
	 */
	ArrayList<Node> rootLevelStructureCase2 = null;
	/**
	 * 0: previous
	 * 1: current
	 */
	double[] currentAndPreviousMinSSs = null;
	
	/**
	 * added on April 17, 2009
	 */
	TreeMap<String, double[][]> bestFitNonStationaryJointProbMatrices = null;
	TreeMap<String, double[]> bestFitNonStationaryMarginalFrequencies = null;
	
	TreeMap<String, ArrayList<double[][]>> bestFitCondsMatrices = null;
	
	TreeMap<String, double[][]> bestFitNonStationaryConds = null;
	
	ArrayList<double[][]> rootBranchChildToParentPermutations = null;
	
	/**
	 * added on June 30, 2011
	 */
	TreeMap<String, ArrayList<double[][]>> bestFitBHJoints = null;
	TreeMap<String, double[][]> bestFitNonStationaryBHJoints = null;
	
	public SearchShortestPath() 
	{
		differences       = new TreeMap<String, double[]>();
		reconstructedTree = new TreeMap<String, Node>();
		treeLevelList     = new ArrayList<ArrayList<String>>();
		intNodeList       = new ArrayList<String>();
		
		/**
		 * add on Mar.28, 2009
		 * for handling rooting branch
		 */
		rootingBranchDifs = new ArrayList<double[]>();
		minSSOfRootingBranch = new double[2];
		/**
		 * added on April 10, 2009
		 */
		treeLevelListSubtree2 = new ArrayList<ArrayList<String>>();
		rootLevelStructureCase2 = new ArrayList<Node>();
		currentAndPreviousMinSSs = new double[2];
		currentAndPreviousMinSSs[0] = 99999999999.0;//in case misupdating, take current.
		currentAndPreviousMinSSs[1] = 99999999999.0;

		/**
		 * added on April 17, 2009
		 */
		bestFitNonStationaryJointProbMatrices= new TreeMap<String, double[][]>();
		bestFitNonStationaryMarginalFrequencies = new TreeMap<String, double[]>();
		bestFitNonStationaryConds = new TreeMap<String, double[][]>();
		
		bestFitCondsMatrices = new TreeMap<String, ArrayList<double[][]>>();
		
		rootBranchChildToParentPermutations = new ArrayList<double[][]>();
		
		/**
		 * added on June 30, 2011
		 */
		bestFitBHJoints = new TreeMap<String, ArrayList<double[][]>>();
		
		
	}
	
	private void reconstructTree(String                                        oldRoot, 
			                     String                                        newRoot, 
					             ArrayList<String>                             internalNodeList,
					             ArrayList<String>                             leafNodeList,
					             TreeMap<String, NodeStructure>                aTree,
					             TreeMap<String, ContainerForDifTwoDirections> inputDifferences)
	{
		/**
		 * set isInternalNode flag  and root node
		 */
		isInternalNode = isAnInternalNode(newRoot, internalNodeList);
		rootNode = new String(newRoot);
		oldRootNode = new String(oldRoot);
		
		String currentIntNode, currentIntNodesOldParent, currentIntPreviousNode;
		Node tmpNode;
		NodeStructure tmpNodeStructure;
		boolean continueClimbing = true;
		int j, conSize;//, k;
		
		if(reconstructedTree == null)
			reconstructedTree = new TreeMap<String, Node>();
		/**
		 * handle special case that newRoot's parent is oldRoot.
		 * this is the case that no tree reconstruction is necessary.
		 * actions necessary are 
		 * 1. setup root node
		 * 2. copy all nodes into reconstructed tree container
		 */
		conSize = internalNodeList.size();
		for(j=0; j<conSize; j++)
		{
			currentIntNode = (String)internalNodeList.get(j);
			tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);
			currentIntNodesOldParent = (String)tmpNodeStructure.getParentNode();
			if(currentIntNodesOldParent.equals(oldRootNode)&&currentIntNode.equals(rootNode))
			{
				/**
				 * it is the case of new root's parent is old root
				 */
				isRootParentOldRoot = true;
				tmpNode = new Node();
				tmpNode.setNodeName(newRoot);
				tmpNodeStructure = (NodeStructure)aTree.get(rootNode);
				
				tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());	
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
					
				tmpNode.setParent("");
				tmpNode.setEvolutionaryDirection(-1);
				reconstructedTree.put(newRoot, tmpNode);
				obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences);
				obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences);
				/**
				 * add on Mar.28, 2009
				 */
				obtainRootingBranchDifs(inputDifferences, newRoot);
				//printTree(internalNodeList, leafNodeList);
				return;
			}
		}
		
		/**
		 * get newRoot's parent
		 */
		tmpNodeStructure = (NodeStructure)aTree.get(newRoot);
		currentIntNode = (String)tmpNodeStructure.getParentNode();
		currentIntPreviousNode = newRoot;

		/** 
		 * set current root
		 */
		if(isInternalNode == true)
		{
			tmpNode = new Node();
			tmpNode.setNodeName(newRoot);
			tmpNode.setParent(aTree.get(newRoot).getParentNode());
			tmpNode.setLeftChild(aTree.get(newRoot).getChildNode1());
			tmpNode.setRightChild(aTree.get(newRoot).getChildNode2());
			tmpNode.setEvolutionaryDirection(-1);
			reconstructedTree.put(newRoot, tmpNode);
		}
		else
		{
			tmpNode = new Node();
			tmpNode.setNodeName(newRoot);
			tmpNode.setParent("");
			tmpNode.setLeftChild(currentIntNode);
			tmpNode.setRightChild("");
			tmpNode.setEvolutionaryDirection(-1);
			reconstructedTree.put(newRoot, tmpNode);
		}
		/**
		 * handle special parent node of new root in the sense that
		 * the parent node connected directly with old root
		 */
		tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);
		currentIntNodesOldParent = (String)tmpNodeStructure.getParentNode();
		if(currentIntNodesOldParent.equals(oldRoot))
		{
			tmpNode = new Node();
			tmpNode.setNodeName(currentIntNode);
			tmpNode.setLeftChild(oldRoot);
			if(tmpNodeStructure.getChildNode1().equals(newRoot))
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
			else
				tmpNode.setRightChild(tmpNodeStructure.getChildNode1());
			tmpNode.setEvolutionaryDirection(1);
			tmpNode.setParent(newRoot);
			tmpNode.setNodeType("Internal");
			
			reconstructedTree.put(currentIntNode, tmpNode);
			
			/**
			 * set up for old root: now it is a leaf node
			 */
			tmpNode = new Node();
			tmpNode.setNodeName(oldRoot);
			tmpNode.setNodeType("Leaf");
			tmpNode.setLeftChild("");
			tmpNode.setRightChild("");
			tmpNode.setParent(currentIntNode);
			
			reconstructedTree.put(oldRoot, tmpNode);
			
			/**
			 * set up differences for old Root
			 */ 
			differences.put(oldRoot, inputDifferences.get(currentIntNode).getChildToParent());
			
			obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences);
			obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences);
			//printTree(internalNodeList, leafNodeList);
			
			/**
			 * add on Mar.28, 2009
			 */
			obtainRootingBranchDifs(inputDifferences, newRoot);
			//printTree(internalNodeList, leafNodeList);
			return;
		}
		while(continueClimbing)
		{
			/**
			 * processing parent node
			 */
			tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);			
			/**
			 * get parent's parent
			 */
			currentIntNodesOldParent = tmpNodeStructure.getParentNode();
			
			tmpNode = new Node();
			tmpNode.setNodeName(currentIntNode);
			/**
			 * evolutionary direction: from child to parent
			 */
			tmpNode.setEvolutionaryDirection(1);
			/**
			 * if newRoot's parent is oldRoot, only this internalNode needs to be reset. 
			 * Other internal nodes are just being copied to new container.  
			 */
			if(currentIntNodesOldParent.equals(oldRoot))
			{
				if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode1()))
				{
					tmpNode.setLeftChild(oldRoot);
					tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
					tmpNode.setParent(currentIntPreviousNode);
					tmpNode.setNodeType("Internal");
				}
				else if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode2()))
				{	
					tmpNode.setLeftChild(oldRoot);
					tmpNode.setRightChild(tmpNodeStructure.getChildNode1());
					tmpNode.setParent(currentIntPreviousNode);
					tmpNode.setNodeType("Internal");
							
				}
				else
				{
					System.out.println("error 1 in reconstructTreeLeafAsRoot\n");
					System.exit(-1);
				}
						
				reconstructedTree.put(currentIntNode, tmpNode);
				differences.put(currentIntNode, inputDifferences.get(currentIntPreviousNode).getChildToParent());
				/**
				 * set up for old root: now it is leaf node
				 */
				tmpNode = new Node();
				tmpNode.setNodeName(oldRoot);
				tmpNode.setLeftChild("");
				tmpNode.setRightChild("");
				tmpNode.setNodeType("Leaf");
				tmpNode.setParent(currentIntNode);
				reconstructedTree.put(oldRoot, tmpNode);
				
				differences.put(oldRoot, inputDifferences.get(currentIntNode).getChildToParent());
			
				continueClimbing = false;
			}
			else
			{
				/**
				 * Set up current node structure
				 */
				tmpNode.setParent(currentIntPreviousNode);
				if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode1()))
				{
					tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
					tmpNode.setLeftChild(currentIntNodesOldParent);
				}
				else if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode2()))
				{
					tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());
					tmpNode.setRightChild(currentIntNodesOldParent);				
				}
				else
				{
					System.out.println("error 2 in reconstructTreeLeafAsRoot\n");
					System.exit(-1);
				}
				tmpNode.setNodeType("Internal");
				reconstructedTree.put(currentIntNode, tmpNode);
				/** 
				 *set up differences
				 */
				differences.put(currentIntNode, inputDifferences.get(currentIntPreviousNode).getChildToParent());
				/**
				 * move forward and switch parameters
				 */
				currentIntPreviousNode = currentIntNode;
				currentIntNode = currentIntNodesOldParent;
			}
		}
		
		obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences);
		obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences);	
		//printTree(internalNodeList, leafNodeList);
		/**
		 * add on Mar.28, 2009
		 */
		obtainRootingBranchDifs(inputDifferences, newRoot);
		//printTree(internalNodeList, leafNodeList);
	}
		
	private void obtainTreeLevelNodesList(ArrayList<String>     internalNodeList)
	{
		//ArrayList<ArrayList<String>> results = new ArrayList<ArrayList<String>>();
		
		String tmpLeft=null, tmpRight=null, tmpIntNodeName=null, tmpParent=null;
		Node tmpNode = null;
		ArrayList<String> currentLevel = null;
		ArrayList<String> previousLevel = new ArrayList<String>();
		int index, conSize, previousLevelSize, counter = 0;

		if(treeLevelList == null)
			treeLevelList = new ArrayList<ArrayList<String>>();
		
		conSize = internalNodeList.size();
		for(index=0; index<conSize; index++)
		{
			tmpIntNodeName = (String)internalNodeList.get(index);
			//if(reconstructedTree.containsKey(tmpIntNodeName))
			//	System.out.println("yes");
			tmpNode = (Node)reconstructedTree.get(tmpIntNodeName);
			tmpParent = tmpNode.getParent();
			if(tmpParent.equals(rootNode))
			{
				if(isInternalNode == true && isRootParentOldRoot == false)
				{
					tmpNode = (Node)reconstructedTree.get(rootNode);
					tmpParent = tmpNode.getParent();
					break;
				}
				else
					break;
			}
		}
		
		if(isInternalNode == true && isRootParentOldRoot == false)
		{
			currentLevel = new ArrayList<String>();
			currentLevel.add(rootNode);
			currentLevel.add(tmpParent);
			treeLevelList.add(currentLevel);
		}
		else if(isInternalNode == true && isRootParentOldRoot == true)
		{
			currentLevel = new ArrayList<String>();
			currentLevel.add(rootNode);
			treeLevelList.add(currentLevel);
		}
		else
		{
		/**
		 * obtain initial left and right children
		 */
			currentLevel = new ArrayList<String>();
			currentLevel.add(tmpIntNodeName);
			treeLevelList.add(currentLevel);

			//tmpLeft = tmpNode.getLeftChild();
			//tmpRight = tmpNode.getRightChild();
		
			//currentLevel = new ArrayList<String>();
			//currentLevel.add(tmpLeft);
			//currentLevel.add(tmpRight);
		}
		
		counter++;
		previousLevel.addAll(currentLevel);
		
		while(true)
		{
			previousLevelSize = previousLevel.size();
			
			currentLevel = new ArrayList<String>();
			
			for(int j=0; j<previousLevelSize; j++)
			{
				String tmpNodeName = previousLevel.get(j);
								
				tmpNode = (Node)reconstructedTree.get(tmpNodeName);
				if((isInternalNode && tmpNodeName.equals(rootNode)) || tmpNode.getNodeType().equals("Internal"))
				{
					tmpLeft = tmpNode.getLeftChild();
					tmpRight = tmpNode.getRightChild();
					currentLevel.add(tmpLeft);
					currentLevel.add(tmpRight);
					
				}
			}
			
			if(currentLevel.size() != 0)
			{
				counter++;
				treeLevelList.add(currentLevel);
				previousLevel.clear();
				previousLevel.addAll(currentLevel);
				//currentLevel = new ArrayList<String>();
			}
			else
			{
				break;
			}
		}
				
	}
	
	private void  obtainAnEntireNodeList(ArrayList<String>                             aNodeList,
			                             TreeMap<String, NodeStructure>                aTree,
			                             int                                           intOrLeaf,
			                             TreeMap<String, ContainerForDifTwoDirections> inputDifferences)
			                                           
	{
		int index, conSize;
		String tmpNodeName;
		NodeStructure tmpNodeStructure;
		//TreeMap<String, Node> results = new TreeMap<String, Node>();
		Node tmpNode = null;
		
		//results.putAll(inputCon);
		conSize = aNodeList.size();
		
		for(index=0; index<conSize; index++)
		{
			tmpNodeName = (String)aNodeList.get(index);
			if(!reconstructedTree.containsKey(tmpNodeName))
			{
				/**
				 * obtain node structure. all information in this structure
				 * will be copied into a new node
				 */
				tmpNodeStructure = (NodeStructure)aTree.get(tmpNodeName);
				tmpNode = new Node();
				
				tmpNode.setNodeName(tmpNodeName);
				tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
				tmpNode.setParent(tmpNodeStructure.getParentNode());
				
				/**
				 * evolutionary direction: from parent to child
				 */
				tmpNode.setEvolutionaryDirection(0);
				if(intOrLeaf == 1)
				{
					/**
					 * internal 
					 */
					tmpNode.setNodeType("Internal");
				}
				else
				{
					tmpNode.setNodeType("Leaf");
				}
				
				reconstructedTree.put(tmpNodeName, tmpNode);
				differences.put(tmpNodeName, inputDifferences.get(tmpNodeName).getParentToChild());
			}
		}

	}
	
	
	private void  obtainAnEntireNodeList(ArrayList<String>                             aNodeList,
			                             TreeMap<String, NodeStructure>                aTree,
                                         int                                           intOrLeaf,
                                         TreeMap<String, ContainerForDifTwoDirections> inputDifferences, 
                                         TreeMap<String, ContainerTwoArrayListDouble> inputBestFitConds
                                         )
                          
	{
		int index, conSize;
		String tmpNodeName;
		NodeStructure tmpNodeStructure;
		//TreeMap<String, Node> results = new TreeMap<String, Node>();
		Node tmpNode = null;

		//results.putAll(inputCon);
		conSize = aNodeList.size();

		for(index=0; index<conSize; index++)
		{
			tmpNodeName = (String)aNodeList.get(index);
			if(!reconstructedTree.containsKey(tmpNodeName))
			{
				/**
				 * obtain node structure. all information in this structure
				 * will be copied into a new node
				 */
				tmpNodeStructure = (NodeStructure)aTree.get(tmpNodeName);
				tmpNode = new Node();

				tmpNode.setNodeName(tmpNodeName);
				tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
				tmpNode.setParent(tmpNodeStructure.getParentNode());

				/**
				 * evolutionary direction: from parent to child
				 */
				tmpNode.setEvolutionaryDirection(0);
				if(intOrLeaf == 1)
				{
					/**
					 * internal 
					 */
					tmpNode.setNodeType("Internal");
				}
				else
				{
					tmpNode.setNodeType("Leaf");
				}

				reconstructedTree.put(tmpNodeName, tmpNode);
				differences.put(tmpNodeName, inputDifferences.get(tmpNodeName).getParentToChild());
				/**
				 * added on April 17, 2009
				 */
				bestFitCondsMatrices.put(tmpNodeName, inputBestFitConds.get(tmpNodeName).getParentToChild());
			}
		}

	}


	private void  obtainAnEntireNodeList(ArrayList<String>                             aNodeList,
										 TreeMap<String, NodeStructure>                aTree,
										 int                                           intOrLeaf,
										 TreeMap<String, ContainerForDifTwoDirections> inputDifferences, 
										 TreeMap<String, ContainerTwoArrayListDouble> inputBestFitConds, 
										 TreeMap<String, ArrayList<double[][]>> parentToChildJoints
										 )

	{
		int index, conSize;
		String tmpNodeName;
		NodeStructure tmpNodeStructure;
		//TreeMap<String, Node> results = new TreeMap<String, Node>();
		Node tmpNode = null;

		//results.putAll(inputCon);
		conSize = aNodeList.size();

		for(index=0; index<conSize; index++)
		{
			tmpNodeName = (String)aNodeList.get(index);
			if(!reconstructedTree.containsKey(tmpNodeName))
			{
				/**
				 * obtain node structure. all information in this structure
				 * will be copied into a new node
				 */
				tmpNodeStructure = (NodeStructure)aTree.get(tmpNodeName);
				tmpNode = new Node();

				tmpNode.setNodeName(tmpNodeName);
				tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
				tmpNode.setParent(tmpNodeStructure.getParentNode());

				/**
				 * evolutionary direction: from parent to child
				 */
				tmpNode.setEvolutionaryDirection(0);
				if(intOrLeaf == 1)
				{
					/**
					 * internal 
					 */
					tmpNode.setNodeType("Internal");
				}
				else
				{
					tmpNode.setNodeType("Leaf");
				}

				reconstructedTree.put(tmpNodeName, tmpNode);
				differences.put(tmpNodeName, inputDifferences.get(tmpNodeName).getParentToChild());
				/**
				 * added on April 17, 2009
				 */
				bestFitCondsMatrices.put(tmpNodeName, inputBestFitConds.get(tmpNodeName).getParentToChild());
				/**
				 * added on June 30, 2011
				 */
				bestFitBHJoints.put(tmpNodeName, parentToChildJoints.get(tmpNodeName));
			}
		}

	}


	ContainerShortestPathSearchResults findTheShortestPath(String oldRoot, 
			                                               String newRoot,
			                                               ArrayList<String> internalNodeList,
			                                               ArrayList<String> leafNodeList,
			                                               TreeMap<String, NodeStructure> aTree,
			                                               TreeMap<String, ContainerForDifTwoDirections> differences)
	{
		
		double result=0.0;
		
		ArrayList<String> aTreeLevel = null;
		int numOfLevels, j, k, numOfElementALevel, flag=1;
		String tmpNodeName;
		Node tmpNode;
		boolean intNodeListCopySuc = false;
		
		ContainerShortestPathSearchResults results = new ContainerShortestPathSearchResults();
		TreeMap<String, Double> miniSumOfSquares = new TreeMap<String, Double>();
		TreeMap<String, ArrayList<Integer>> permutationIndices = new TreeMap<String, ArrayList<Integer>>();
		ArrayList<Integer> tmpList = null;
		
		if(intNodeList == null)
			intNodeList = new ArrayList<String>();
		intNodeList.clear();
		intNodeListCopySuc = intNodeList.addAll(internalNodeList);
		
		if(intNodeListCopySuc == false)
		{
			System.err.println("Copy internal node list failed.");
			System.exit(-1);
		}
		/**
		 * create reconstructed tree
		 */
		reconstructTree(oldRoot, newRoot, internalNodeList, leafNodeList, aTree, differences);
		/**
		 * create the level list
		 */
		obtainTreeLevelNodesList(internalNodeList);
		while(true)
		{
			if(flag == 0 && isCurrentBetterThanPrevious() == false)
				break;
			else
			{
				flag = 0;
				getAShortestPath();
				itrCounter++;
			}
		}
				
		if(treeLevelList.size() > 0)
			aTreeLevel = (ArrayList<String>)treeLevelList.get(0);
		
		/**
		 * update final results
		 * 1. top level
		 */
		if(isInternalNode == false || isRootParentOldRoot == true)//a leaf node
		{
			result = reconstructedTree.get(aTreeLevel.get(0)).getPreviousCumulativeDifs();
			miniSumOfSquares.put(new String(aTreeLevel.get(0)), new Double(result));
			tmpList = new ArrayList<Integer>();
			
			tmpList.add(new Integer(reconstructedTree.get(aTreeLevel.get(0)).getPreviousRowPermutationIndex()));
			tmpList.add(new Integer(reconstructedTree.get(aTreeLevel.get(0)).getPreviousColumnPermutationIndex()));
			permutationIndices.put(new String(reconstructedTree.get(aTreeLevel.get(0)).getNodeName()), tmpList);
		}
		else
		{
			// an internal node
			result = reconstructedTree.get(aTreeLevel.get(0)).getPreviousCumulativeDifs()
			       + reconstructedTree.get(aTreeLevel.get(1)).getPreviousCumulativeDifs();
			miniSumOfSquares.put(new String(aTreeLevel.get(0)), new Double(reconstructedTree.get(aTreeLevel.get(0)).getPreviousCumulativeDifs()));
			miniSumOfSquares.put(new String(aTreeLevel.get(1)), new Double(reconstructedTree.get(aTreeLevel.get(1)).getPreviousCumulativeDifs()));
			tmpList = new ArrayList<Integer>();
			
			tmpList.add(new Integer(reconstructedTree.get(aTreeLevel.get(0)).getPreviousRowPermutationIndex()));
			tmpList.add(new Integer(reconstructedTree.get(aTreeLevel.get(0)).getPreviousColumnPermutationIndex()));
			permutationIndices.put(new String(reconstructedTree.get(aTreeLevel.get(0)).getNodeName()), tmpList);

			tmpList = new ArrayList<Integer>();
			
			tmpList.add(new Integer(reconstructedTree.get(aTreeLevel.get(1)).getPreviousRowPermutationIndex()));
			tmpList.add(new Integer(reconstructedTree.get(aTreeLevel.get(1)).getPreviousColumnPermutationIndex()));
			permutationIndices.put(new String(reconstructedTree.get(aTreeLevel.get(1)).getNodeName()), tmpList);
			
		}
		
		numOfLevels = treeLevelList.size();
		/**
		 * since the last level is always leaf nodes and all levels results are stored in internal nodes, 
		 * last level could be ignored.
		 */
		
		if(numOfLevels > 2)
		{
			for(j=1; j<(numOfLevels-1); j++)
			{
				aTreeLevel = treeLevelList.get(j);
				numOfElementALevel = aTreeLevel.size();
				for(k=0; k<numOfElementALevel; k++)
				{
					tmpNodeName = (String)aTreeLevel.get(k);
					tmpNode = (Node)reconstructedTree.get(tmpNodeName);
					if(tmpNode.getNodeType().equals("Internal"))
					{
						result += tmpNode.getPreviousCumulativeDifs();
						miniSumOfSquares.put(new String(tmpNodeName), new Double(tmpNode.getPreviousCumulativeDifs()));
						
						tmpList = new ArrayList<Integer>();
						
						tmpList.add(new Integer(tmpNode.getPreviousRowPermutationIndex()));
						tmpList.add(new Integer(tmpNode.getPreviousColumnPermutationIndex()));
						permutationIndices.put(new String(tmpNodeName), tmpList);
					}
					
				}
			}
		}
		results.setMinimumSumOfSquares(result);
		results.setMinimumAtInternals(miniSumOfSquares);
		results.setPermutationIndices(permutationIndices);
		results.setItrCounter(itrCounter);
		results.setReconstructedTree(reconstructedTree);
		
		return results;
	}

	
	ContainerShortestPathSearchResults findTheShortestPath(TreeMap<String, NodeStructure> aTree)
	{

		double result=0.0;

		int j, flag=1;
		String tmpNodeName;
		Node tmpNode;

		ContainerShortestPathSearchResults results = new ContainerShortestPathSearchResults();
		TreeMap<String, Double> miniSumOfSquares = new TreeMap<String, Double>();
		TreeMap<String, ArrayList<Integer>> permutationIndices = new TreeMap<String, ArrayList<Integer>>();
		ArrayList<Integer> tmpList = null;

		while(true)
		{
			if(flag == 0 && isCurrentBetterThanPreviousWithoutLevel() == false)
				break;
			else
			{
				flag = 0;
				getAShortestPathWithOutLevelCon(aTree);
				itrCounter++;
			}
		}

		for(j=0; j<intNodeList.size(); j++)
		{
			tmpNodeName = (String)intNodeList.get(j);
			tmpNode = (Node)reconstructedTree.get(tmpNodeName);
			result += tmpNode.getPreviousCumulativeDifs();
			miniSumOfSquares.put(new String(tmpNodeName), new Double(tmpNode.getPreviousCumulativeDifs()));

			tmpList = new ArrayList<Integer>();

			tmpList.add(new Integer(tmpNode.getPreviousRowPermutationIndex()));
			tmpList.add(new Integer(tmpNode.getPreviousColumnPermutationIndex()));
			permutationIndices.put(new String(tmpNodeName), tmpList);

		}
		
		results.setMinimumSumOfSquares(result);
		results.setMinimumAtInternals(miniSumOfSquares);
		results.setPermutationIndices(permutationIndices);
		results.setItrCounter(itrCounter);
		return results;
	}
	
	
		
	private void getAShortestPath()
	{
		int numOfLevels = treeLevelList.size();
		String nodeName, nodeType;
		ArrayList<String> aTreeLevel = null;
		Node aNode = null;
		
		int j, k, numOfElements, tmpIndex;
		
		for(j=0; j<numOfLevels; j++)
		{
			/**
			 * tmpIndex = (total levels - 1) - j
			 * total levels: Index of last level
			 * j: number of levels has been gone through
			 */
			tmpIndex = numOfLevels - 1 - j;
			aTreeLevel = (ArrayList<String>)treeLevelList.get(tmpIndex);
			numOfElements = aTreeLevel.size();
			for(k=0; k<numOfElements; k++)
			{
				nodeName = (String)aTreeLevel.get(k);
				aNode = (Node)reconstructedTree.get(nodeName);
				nodeType = aNode.getNodeType();
				
				//if((tmpIndex==0) || ((aNode.getParent().equals(rootNode))&&aNode.getNodeType()=="Internal"&&isRootParentOldRoot==false))
				if(tmpIndex==0)
				{
					//updatePermutationInformASpecialInternalNode(nodeName);
					updateRootingBranch(nodeName);
				}
				else if(nodeType.equals("Internal"))
				{
					updatePermutationInformAnInternalNode(nodeName);
				}
			}
		}
	
	}
	
	private void getAShortestPathWithOutLevelCon(TreeMap<String, NodeStructure> aTree)
	{
		
		String nodeName;
		int j;
		
		for(j=0; j<intNodeList.size(); j++)
		{
			
			nodeName = (String)intNodeList.get(j);
			if(nodeName.equals(rootNode))
			{
				updatePermutationInformASpecialInternalNode(nodeName);
			}
			else if(isInternalNode&&nodeName.equals(aTree.get(rootNode).getParentNode()))
			{
				updatePermutationInformASpecialInternalNode(nodeName);
			}
			else if(!isInternalNode&&nodeName.equals(aTree.get(rootNode).getParentNode()))
			{
				updatePermutationInformASpecialInternalNode(nodeName);
			}
			else
			{
				updatePermutationInformAnInternalNode(nodeName);
			}
		}
	}
	
	private void updatePermutationInformAnInternalNode(String nodeName)
	{
		String childNameLeft, childNameRight, childNodeTypeLeft, childNodeTypeRight=null;

		Node currentIntNode, nodeLeft, nodeRight;
	
		int permutationIndexLeft, permutationIndexRight, permutationIndexInternal;
		
		double[] difsWorkingLeft, difsWorkingRight, difsLeft, difsRight, difsInternal, difsWorkingInternal;
		//double tmpValue;
		
		currentIntNode = (Node)reconstructedTree.get(nodeName);
		childNameLeft = (String)currentIntNode.getLeftChild();
		childNameRight = (String)currentIntNode.getRightChild();
		permutationIndexInternal = currentIntNode.getCurrentRowPermutationIndex();
		
		difsLeft = differences.get(childNameLeft);
		difsRight = differences.get(childNameRight);
		difsInternal = differences.get(nodeName);
		
		/**
		 * left child
		 */
		nodeLeft = (Node)reconstructedTree.get(childNameLeft);
		childNodeTypeLeft = nodeLeft.getNodeType();
		/**
		 * right child
		 */
		nodeRight = (Node)reconstructedTree.get(childNameRight);
		childNodeTypeRight = nodeRight.getNodeType();
		
		if(childNodeTypeLeft.equals("Internal"))
		{
			permutationIndexLeft = nodeLeft.getCurrentColumnPermutationIndex();
			difsWorkingLeft = obtainWorkingDifsColumnFix(difsLeft, permutationIndexLeft);
		}
		else
		{
			difsWorkingLeft = differences.get(childNameLeft);
		}
		
		if(childNodeTypeRight.equals("Internal"))
		{
			permutationIndexRight = nodeRight.getCurrentColumnPermutationIndex();
			difsWorkingRight = obtainWorkingDifsColumnFix(difsRight, permutationIndexRight);
		}
		else
		{
			difsWorkingRight = differences.get(childNameRight);
		}
		//System.out.println("node name: " + nodeName + "\n" + "permutation index: " + permutationIndexInternal);
		
		difsWorkingInternal = obtainWorkingDifsRowFix(difsInternal, permutationIndexInternal);
		
		TreeMap<Integer, Double> shortestDifResults = obtainShortestDifASubtree(difsWorkingLeft, difsWorkingRight, difsWorkingInternal);
		double shortestDif = shortestDifResults.get(shortestDifResults.firstKey()).doubleValue();
		int permutationIndexResults = shortestDifResults.firstKey().intValue();
		//System.out.println("node name: " + nodeName + " shortest: " + shortestDif);
		/**
		 * update current internal node
		 */
		reconstructedTree.get(nodeName).setPreviousCumulativeDif(reconstructedTree.get(nodeName).getCurrentCumulativeDifs());
		reconstructedTree.get(nodeName).setCurrentCumulativeDif(shortestDif);
		reconstructedTree.get(nodeName).setPreviousColumnPermutationIndex(reconstructedTree.get(nodeName).getCurrentColumnPermutationIndex());
		reconstructedTree.get(nodeName).setCurrentColumnPermutationIndex(permutationIndexResults);
		//System.out.println("node informa" + reconstructedTree.get(nodeName).toString());
		
		/**
		 * update left and right children 
		 */
		reconstructedTree.get(childNameLeft).setPreviousRowPermutationIndex(reconstructedTree.get(childNameLeft).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameLeft).setCurrentRowPermutationIndex(permutationIndexResults);
		reconstructedTree.get(childNameRight).setPreviousRowPermutationIndex(reconstructedTree.get(childNameRight).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameRight).setCurrentRowPermutationIndex(permutationIndexResults);
		
	}

	private void updateRootingBranch(String nodeName)
	{
		String childNameLeft, childNameRight, childNodeTypeLeft, childNodeTypeRight=null;

		Node currentIntNode, nodeLeft, nodeRight;
	
		int permutationIndexLeft, permutationIndexRight;
		
		double[] difsWorkingLeft, difsWorkingRight, difsLeft, difsRight;
		//double tmpValue;
		
		currentIntNode = (Node)reconstructedTree.get(nodeName);
		childNameLeft = (String)currentIntNode.getLeftChild();
		childNameRight = (String)currentIntNode.getRightChild();
		//permutationIndexInternal = currentIntNode.getRowPermutationIndex();
		
		difsLeft = differences.get(childNameLeft);
		difsRight = differences.get(childNameRight);
		
		
		/**
		 * left child
		 */
		nodeLeft = (Node)reconstructedTree.get(childNameLeft);
		childNodeTypeLeft = nodeLeft.getNodeType();
		/**
		 * right child
		 */
		nodeRight = (Node)reconstructedTree.get(childNameRight);
		childNodeTypeRight = nodeRight.getNodeType();
		
		if(childNodeTypeLeft.equals("Internal"))
		{
			permutationIndexLeft = nodeLeft.getCurrentColumnPermutationIndex();
			difsWorkingLeft = obtainWorkingDifsColumnFix(difsLeft, permutationIndexLeft);
		}
		else
		{
			difsWorkingLeft = differences.get(childNameLeft);
		}
		
		if(childNodeTypeRight.equals("Internal"))
		{
			permutationIndexRight = nodeRight.getCurrentColumnPermutationIndex();
			difsWorkingRight = obtainWorkingDifsColumnFix(difsRight, permutationIndexRight);
		}
		else
		{
			difsWorkingRight = differences.get(childNameRight);
		}
		//difsWorkingInternal = obtainWorkingDifsRowFix(difsInternal, permutationIndexInternal);
		
		//TreeMap<Integer, Double> shortestDifResults = obtainShortestDifASubtree(difsWorkingLeft, difsWorkingRight);
		TreeMap<Integer, Double> shortestDifResults = obtainShortestDifIncludeRootingBranch(difsWorkingLeft, difsWorkingRight, nodeName);
		
		double shortestDif = shortestDifResults.get(shortestDifResults.firstKey()).doubleValue();
		int permutationIndexResults = shortestDifResults.firstKey().intValue();
		/**
		 * update current internal node
		 */
		reconstructedTree.get(nodeName).setPreviousCumulativeDif(reconstructedTree.get(nodeName).getCurrentCumulativeDifs());
		reconstructedTree.get(nodeName).setCurrentCumulativeDif(shortestDif);
		reconstructedTree.get(nodeName).setPreviousColumnPermutationIndex(reconstructedTree.get(nodeName).getCurrentColumnPermutationIndex());
		reconstructedTree.get(nodeName).setCurrentColumnPermutationIndex(permutationIndexResults);
		/**
		 * update left and right children 
		 */
		reconstructedTree.get(childNameLeft).setPreviousRowPermutationIndex(reconstructedTree.get(childNameLeft).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameLeft).setCurrentRowPermutationIndex(permutationIndexResults);
		reconstructedTree.get(childNameRight).setPreviousRowPermutationIndex(reconstructedTree.get(childNameRight).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameRight).setCurrentRowPermutationIndex(permutationIndexResults);
		
	}
	
	private void updatePermutationInformASpecialInternalNode(String nodeName)
	{
		String childNameLeft, childNameRight, childNodeTypeLeft, childNodeTypeRight=null;

		Node currentIntNode, nodeLeft, nodeRight;
	
		int permutationIndexLeft, permutationIndexRight;
		
		double[] difsWorkingLeft, difsWorkingRight, difsLeft, difsRight;
		//double tmpValue;
		
		currentIntNode = (Node)reconstructedTree.get(nodeName);
		childNameLeft = (String)currentIntNode.getLeftChild();
		childNameRight = (String)currentIntNode.getRightChild();
		//permutationIndexInternal = currentIntNode.getRowPermutationIndex();
		
		difsLeft = differences.get(childNameLeft);
		difsRight = differences.get(childNameRight);
		
		
		/**
		 * left child
		 */
		nodeLeft = (Node)reconstructedTree.get(childNameLeft);
		childNodeTypeLeft = nodeLeft.getNodeType();
		/**
		 * right child
		 */
		nodeRight = (Node)reconstructedTree.get(childNameRight);
		childNodeTypeRight = nodeRight.getNodeType();
		
		if(childNodeTypeLeft.equals("Internal"))
		{
			permutationIndexLeft = nodeLeft.getCurrentColumnPermutationIndex();
			difsWorkingLeft = obtainWorkingDifsColumnFix(difsLeft, permutationIndexLeft);
		}
		else
		{
			difsWorkingLeft = differences.get(childNameLeft);
		}
		
		if(childNodeTypeRight.equals("Internal"))
		{
			permutationIndexRight = nodeRight.getCurrentColumnPermutationIndex();
			difsWorkingRight = obtainWorkingDifsColumnFix(difsRight, permutationIndexRight);
		}
		else
		{
			difsWorkingRight = differences.get(childNameRight);
		}
		
		//difsWorkingInternal = obtainWorkingDifsRowFix(difsInternal, permutationIndexInternal);
		
		TreeMap<Integer, Double> shortestDifResults = obtainShortestDifASubtree(difsWorkingLeft, difsWorkingRight);
		double shortestDif = shortestDifResults.get(shortestDifResults.firstKey()).doubleValue();
		int permutationIndexResults = shortestDifResults.firstKey().intValue();
		/**
		 * update current internal node
		 */
		reconstructedTree.get(nodeName).setPreviousCumulativeDif(reconstructedTree.get(nodeName).getCurrentCumulativeDifs());
		reconstructedTree.get(nodeName).setCurrentCumulativeDif(shortestDif);
		reconstructedTree.get(nodeName).setPreviousColumnPermutationIndex(reconstructedTree.get(nodeName).getCurrentColumnPermutationIndex());
		reconstructedTree.get(nodeName).setCurrentColumnPermutationIndex(permutationIndexResults);
		/**
		 * update left and right children 
		 */
		reconstructedTree.get(childNameLeft).setPreviousRowPermutationIndex(reconstructedTree.get(childNameLeft).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameLeft).setCurrentRowPermutationIndex(permutationIndexResults);
		reconstructedTree.get(childNameRight).setPreviousRowPermutationIndex(reconstructedTree.get(childNameRight).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameRight).setCurrentRowPermutationIndex(permutationIndexResults);
		
	}
	
	
	private double[] obtainWorkingDifsColumnFix(double[] difs, 
			                                    int      inputFixColumnIndex)
	{
		double[] results = new double[defaultNumberOfPermutations];
		int index, tmpIndex;
		
		for(index=0; index<defaultNumberOfPermutations; index++)
		{
			tmpIndex = inputFixColumnIndex + index*defaultNumberOfPermutations;
			results[index] = difs[tmpIndex];
		}
		
		return results;
	
	}

	private double[] obtainWorkingDifsRowFix(double[] difs, 
                                             int      inputFixRowIndex)
   {
		double[] results = new double[defaultNumberOfPermutations];
		int index, tmpIndex;

		for(index=0; index<defaultNumberOfPermutations; index++)
		{
			tmpIndex = inputFixRowIndex*defaultNumberOfPermutations + index;
			//System.out.println(index + ": " + tmpIndex);
			results[index] = difs[tmpIndex];
		}

		return results;
   }
	
	
	private TreeMap<Integer, Double> obtainShortestDifASubtree(double[] difsWorkingLeft, 
			                                                   double[] difsWorkingRight,
			                                                   double[] difsWorkingInt)
	{
		TreeMap<Integer, Double> results = new TreeMap<Integer, Double>();
		double tmpDif, minDif = 9999999.0;
		int index, indexResult=0;
		
		for(index=0; index<defaultNumberOfPermutations; index++)
		{
			tmpDif = difsWorkingLeft[index] + difsWorkingRight[index] + difsWorkingInt[index];
			if(tmpDif<minDif)
			{
				minDif = tmpDif;
				indexResult = index;
			}
		}
		
		results.put(new Integer(indexResult), new Double(difsWorkingLeft[indexResult] + difsWorkingRight[indexResult]));
		
		return results;
	}

	private TreeMap<Integer, Double> obtainShortestDifASubtree(double[] difsWorkingLeft, 
                                                               double[] difsWorkingRight)
    {
		TreeMap<Integer, Double> results = new TreeMap<Integer, Double>();
		double tmpDif, minDif = 9999999.0;
		int index, indexResult=0;

		for(index=0; index<defaultNumberOfPermutations; index++)
		{
			tmpDif = difsWorkingLeft[index] + difsWorkingRight[index];
			if(tmpDif<minDif)
			{
				minDif = tmpDif;
				indexResult = index;
			}
		}

		results.put(new Integer(indexResult), new Double(minDif));

		return results;
    }

	private void obtainRootingBranchDifs(TreeMap<String, ContainerForDifTwoDirections> inputDifs, 
			                             String newRoot)
	{
		if(rootingBranchDifs == null)
			rootingBranchDifs = new ArrayList<double[]>();
		
		double[] tmpArray;
		int arraySize;
		if(isInternalNode == true && isRootParentOldRoot == false)
			arraySize = 576;
		else 
			arraySize = 24;
		//System.out.println("is root parent old root: " + isRootParentOldRoot + "\n" + "is internal node: " + isInternalNode	);
		//System.out.println("new root: " + rootNode.toString());
		tmpArray = inputDifs.get(newRoot).getParentToChild();
		tmpArray = copyArray(tmpArray, arraySize);
		rootingBranchDifs.add(tmpArray);
		tmpArray = inputDifs.get(newRoot).getChildToParent();
		tmpArray = copyArray(tmpArray, arraySize);
		rootingBranchDifs.add(tmpArray);
	}
	
	private TreeMap<Integer, Double> obtainShortestDifIncludeRootingBranch(double[] difsWorkingLeft, 
			                                                               double[] difsWorkingRight,
			                                                               String nodeName)
	{
		TreeMap<Integer, Double> results = new TreeMap<Integer, Double>();
		double[] rootingBranchDifsWorking = null, tmpDifs = null;
		double minValue = 9999999.0, tmpValue;
		TreeMap<Integer, Double> tmpResults;
		
		int numOfIterations = 0;
		int j, indexOfMinValue =0;
		
		if(nodeName.equals(rootNode))
			rootingBranchDifsWorking = rootingBranchDifs.get(0);
		else 
			rootingBranchDifsWorking = rootingBranchDifs.get(1);
		
		if(isInternalNode == true && isRootParentOldRoot == false)
			numOfIterations = 24;//24 possible row permutations
		else
			numOfIterations = 1;
		//System.out.println("node name: " + nodeName.toString());
		
		for(j=0; j<numOfIterations; j++)
		{
			//System.out.println("j: " + j);
			tmpDifs = obtainWorkingDifsRowFix(rootingBranchDifsWorking, j);
			tmpResults = obtainShortestDifASubtree(difsWorkingLeft, difsWorkingRight, tmpDifs);
			tmpValue = tmpResults.get(tmpResults.firstKey()).doubleValue();
			if(tmpValue < minValue)
			{
				minValue = tmpValue;
				indexOfMinValue = tmpResults.firstKey().intValue();
			}
		}
		
		/**
		 * AT THIS MOMENT, THE DIFFERENCE OF ROOTING BRANCH WILL NOT BE INCLUDED
		 */
		minValue = difsWorkingLeft[indexOfMinValue] + difsWorkingRight[indexOfMinValue];
		
		results.put(new Integer(indexOfMinValue), new Double(minValue));
		return results;
		
	}
	boolean isAnInternalNode(String nodeName, ArrayList<String> inputIntNodeList)
	{	
		int conSize, j;
		String tmpNodeName = null;
		
		conSize = inputIntNodeList.size();
		
		for(j=0; j<conSize; j++)
		{
			tmpNodeName = (String)inputIntNodeList.get(j);
			
			if(tmpNodeName.equals(nodeName))
				return true;
		}
		return false;
	}
	
	boolean isUpdateFinished()
	{
		Iterator<Node> itr = null;
		Collection<Node> aCollection = reconstructedTree.values();
		
		//TreeMap<String, Node> anElement;
		Node tmpNode=null;
		
		itr = aCollection.iterator();
		
		while(itr.hasNext())
		{
			tmpNode = itr.next();
			if(tmpNode.getNodeName().equals(rootNode))
				continue;
			else if(tmpNode.getIsUpdate() == true)
				return false;
		}
		
		return true;
	}
	
	boolean isCurrentBetterThanPrevious()
	{
		double resultCurrent=0.0, resultPrevious = 0.0;
		int numOfLevels, numOfElementALevel, j, k;
		ArrayList<String> aList;
		Node tmpNode;
		String tmpNodeName;
		
		if(isInternalNode == false || isRootParentOldRoot == true)//a leaf node
		{
			resultCurrent = reconstructedTree.get(treeLevelList.get(0).get(0)).getCurrentCumulativeDifs();
			resultPrevious = reconstructedTree.get(treeLevelList.get(0).get(0)).getPreviousCumulativeDifs();
		}
		else
		{
			// an internal node
			resultCurrent = reconstructedTree.get(treeLevelList.get(0).get(0)).getCurrentCumulativeDifs()
			       + reconstructedTree.get(treeLevelList.get(0).get(1)).getCurrentCumulativeDifs();
			
			resultPrevious = reconstructedTree.get(treeLevelList.get(0).get(0)).getPreviousCumulativeDifs()
		       + reconstructedTree.get(treeLevelList.get(0).get(1)).getPreviousCumulativeDifs();
		}
		
		numOfLevels = treeLevelList.size();
		/**
		 * since the last level is always leaf nodes and all levels results are stored in internal nodes, 
		 * last level could be ignored.
		 */
		
		if(numOfLevels > 2)
		{
			for(j=1; j<(numOfLevels-1); j++)
			{
				aList = treeLevelList.get(j);
				numOfElementALevel = aList.size();
				
				for(k=0; k<numOfElementALevel; k++)
				{
					tmpNodeName = (String)aList.get(k);
					tmpNode = (Node)reconstructedTree.get(tmpNodeName);
					if(tmpNode.getNodeType().equals("Internal"))
					{
						resultCurrent += tmpNode.getCurrentCumulativeDifs();
						resultPrevious += tmpNode.getPreviousCumulativeDifs();				
					}
					
				}
			}
		}
		if(resultCurrent < resultPrevious)
			return true;
		else
			return false;
	}

	boolean isCurrentBetterThanPreviousWithoutLevel()
	{
		double resultCurrent=0.0, resultPrevious = 0.0;
		int j;
		Node tmpNode;
		String tmpNodeName;
		
		for(j=1; j<intNodeList.size(); j++)
			{
				tmpNodeName = (String)intNodeList.get(j);
				tmpNode = (Node)reconstructedTree.get(tmpNodeName);
				resultCurrent += tmpNode.getCurrentCumulativeDifs();
				resultPrevious += tmpNode.getPreviousCumulativeDifs();				
			}
		
		if(resultCurrent < resultPrevious)
			return true;
		else
			return false;
	}
	
	void printTree(ArrayList<String> internalNodesList,
			       ArrayList<String> leafNodesList)
	{
		int conSize, j;//, k;
		Node tmpNode;
		String tmpNodeName;
		
		conSize = internalNodesList.size();
		
		for(j=0; j<conSize; j++)
		{
			tmpNodeName = internalNodesList.get(j);
			tmpNode = reconstructedTree.get(tmpNodeName);
			System.out.print("NodeName: " + tmpNodeName + "\t Parent: " + tmpNode.getParent() +  "\t Left: " + tmpNode.getLeftChild() + "\t Right: " + tmpNode.getRightChild() + "\n");
		}
		
		conSize = leafNodesList.size();
		
		for(j=0; j<conSize; j++)
		{
			tmpNodeName = leafNodesList.get(j);
			tmpNode = reconstructedTree.get(tmpNodeName);
			System.out.print("NodeName: " + tmpNodeName + "\t Parent: " + tmpNode.getParent() + "\t Left: " + tmpNode.getLeftChild() + "\t Right: " + tmpNode.getRightChild() + "\n");			
		}
	}
	
/*	private void printTreeLevelLists()
	{
		int j, conSize = 0;
		String aLine = "";
		
		if(treeLevelList != null)
			conSize = treeLevelList.size();
		aLine = "treeLevelList: ";
		
		for(j=0; j<conSize; j++)
			aLine += " " + treeLevelList.get(j);
		aLine += "\n";
		
		if(treeLevelListSubtree2 == null)
			return;
		
		aLine += "treeLevelListSubtree2: ";
		conSize = treeLevelListSubtree2.size();
		
		for(j=0; j<conSize; j++)
			aLine += " " + treeLevelListSubtree2.get(j);
		
		System.out.println(aLine);
	}
	*/
	
	private double[] copyArray(double[] inputArray, int arraySize)
	{
		double[] results = new double[arraySize];
		int j;
		
		for(j=0; j<arraySize; j++)
		{
			//System.out.println("j: " + j);
			results[j] = inputArray[j];
		}
		
		return results;
	}

	ContainerShortestPathSearchResults findTheShortestPathRootingAtANode(String oldRoot, 
            	                                                         String newRoot,
                                                                         ArrayList<String> internalNodeList,
                                                                         ArrayList<String> leafNodeList,
                                                                         TreeMap<String, NodeStructure> aTree,
                                                                         TreeMap<String, ContainerForDifTwoDirections> differences,
                                                                         TreeMap<String, ContainerTwoArrayListDouble> bestFitConds, 
                                                                         TreeMap<String, ArrayList<double[][]>>	 childToParentPermutations,
                                                                         TreeMap<String, ArrayList<double[][]>>	 parentToChildPermutations
                                                                         )
	{

		double result=0.0;

		ArrayList<String> aTreeLevel = null;
		int numOfLevels, j, k, numOfElementALevel, flag=1, tmpColumnPermutation;
		String tmpNodeName, tmpParentNodeName;
		Node tmpNode, tmpParentNode;
		boolean intNodeListCopySuc = false;

		ContainerShortestPathSearchResults results = new ContainerShortestPathSearchResults();
		TreeMap<String, Double> miniSumOfSquares = new TreeMap<String, Double>();
		TreeMap<String, ArrayList<Integer>> permutationIndices = new TreeMap<String, ArrayList<Integer>>();
		ArrayList<Integer> tmpList = null;

		double[][] tmpMatrix = null, rootBranchJointProb = null; 
		
		if(intNodeList == null)
			intNodeList = new ArrayList<String>();
		intNodeList.clear();
		intNodeListCopySuc = intNodeList.addAll(internalNodeList);

		if(intNodeListCopySuc == false)
		{
			System.err.println("Copy internal node list failed.");
			System.exit(-1);
		}
		/**
		 * create reconstructed tree
		 */
		reconstructTreeRootingAtANode(oldRoot, 
				                      newRoot, 
				                      internalNodeList, 
				                      leafNodeList, 
				                      aTree, 
				                      differences, 
				                      bestFitConds, 
				                      childToParentPermutations,
				                      parentToChildPermutations);
		
		if(isInternalNode == false)
		{
			/**
			 * case 1: rooting at a leaf or old root
			 */
			
			/**
			 * create the level list
			 */
			obtainTreeLevelNodesListCaseOne(internalNodeList);
			//printTreeLevelLists();
			
			while(true)
			{
				if(flag == 0 && isCurrentBetterThanPreviousCaseOne() == false)
					break;
				else
				{
					flag = 0;
					getAShortestPathCaseOne();
					itrCounter++;
				}
			}
		}
		else
		{
			/**
			 * case two: rooting at an internal node.
			 */
			/**
			 * create the level list
			 */
			obtainTreeLevelNodesListCaseTwo(internalNodeList);
			//printTreeLevelLists();
			
			while(true)
			{
				if(flag == 0 && isCurrentBetterThanPreviousCaseTwo() == false)
					break;
				else
				{
					flag = 0;
					getAShortestPathCaseTwo();
					itrCounter++;
				}
			}
			
		}
		
		
		if(bestFitNonStationaryConds == null)
			bestFitNonStationaryConds = new TreeMap<String, double[][]>();
		if(bestFitNonStationaryBHJoints == null)
			bestFitNonStationaryBHJoints = new TreeMap<String, double[][]>();
			
		numOfLevels = treeLevelList.size();
		//System.out.println("numberOfLevels: " + numOfLevels);
		
		if(treeLevelList.size() > 0)
			aTreeLevel = (ArrayList<String>)treeLevelList.get(0);

		/**
		 * update final results
		 * 1. top level
		 */
		if(isInternalNode == false)//a leaf/root node
		{
			result = reconstructedTree.get(aTreeLevel.get(0)).getPreviousCumulativeDifs();
			//miniSumOfSquares.put(new String(aTreeLevel.get(1)), new Double(result));
		
		
			/**
			 * since the last level is always leaf nodes and all levels results are stored in internal nodes, 
			 * last level could be ignored.
			 */
				
			if(numOfLevels > 1)
			{
				for(j=1; j<numOfLevels; j++)
				{
					aTreeLevel = treeLevelList.get(j);
					numOfElementALevel = aTreeLevel.size();
					for(k=0; k<numOfElementALevel; k++)
					{
						tmpNodeName = (String)aTreeLevel.get(k);
						tmpNode = (Node)reconstructedTree.get(tmpNodeName);
						if(tmpNode.getNodeType().equals("Internal"))
						{
							result += tmpNode.getPreviousCumulativeDifs();
							//System.out.println("at results: node name: " + tmpNodeName + ": " + tmpNode.getPreviousCumulativeDifs());
							miniSumOfSquares.put(new String(tmpNodeName), new Double(tmpNode.getPreviousCumulativeDifs()));
							
							tmpList = new ArrayList<Integer>();
							
							tmpList.add(new Integer(tmpNode.getPreviousRowPermutationIndex()));
							tmpList.add(new Integer(tmpNode.getPreviousColumnPermutationIndex()));
							permutationIndices.put(new String(tmpNodeName), tmpList);
							/**
							 * added on April 17, 2009
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, 
									                         bestFitCondsMatrices,
									                         tmpNode.getPreviousRowPermutationIndex(), 
									                         tmpNode.getPreviousColumnPermutationIndex());
							bestFitNonStationaryConds.put(tmpNodeName, tmpMatrix);
							
							/**
							 * added on June 30, 2011
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, 
			                         						 bestFitBHJoints,
			                         						 tmpNode.getPreviousRowPermutationIndex(), 
			                         						 tmpNode.getPreviousColumnPermutationIndex());
							bestFitNonStationaryBHJoints.put(tmpNodeName, tmpMatrix);
						}
						else
						{
							/**
							 * added on April 17, 2009
							 */
							/**
							 * obtain internal - leaf branch matrix
							 * 1. get parent node
							 * 2. get parent column permutation
							 * 3. obtain matrix using parent column permutation
							 */
							tmpParentNodeName = tmpNode.getParent();
							tmpParentNode = (Node)reconstructedTree.get(tmpParentNodeName);
							tmpColumnPermutation = tmpParentNode.getPreviousColumnPermutationIndex();
							tmpMatrix = extractBestFitMatrix(tmpNodeName, bestFitCondsMatrices, tmpColumnPermutation);
							bestFitNonStationaryConds.put(tmpNodeName, tmpMatrix);
							/**
							 * added on June 30, 2011
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, bestFitBHJoints, tmpColumnPermutation);
							bestFitNonStationaryBHJoints.put(tmpNodeName, tmpMatrix);
						}
						
						/**
						 * added on April 17, 2009
						 */
						if(j == 1)
						{
							rootBranchJointProb = extractBestFitMatrix(rootBranchChildToParentPermutations, 
									                                   tmpNode.getPreviousColumnPermutationIndex());
						}
					}
				}	
			}
		}
		else
		{
			/**
			 * get root information of both subtree 1 and subtree 2
			 */
			result = rootLevelStructureCase2.get(0).getPreviousCumulativeDifs()
				+ rootLevelStructureCase2.get(1).getPreviousCumulativeDifs();
			/**
			 * root node will only hold information of subtree 1
			 */
			miniSumOfSquares.put(new String(rootNode), new Double(rootLevelStructureCase2.get(1).getPreviousCumulativeDifs()));
			tmpList = new ArrayList<Integer>();
			
			tmpList.add(new Integer(rootLevelStructureCase2.get(0).getPreviousRowPermutationIndex()));
			tmpList.add(new Integer(rootLevelStructureCase2.get(0).getPreviousColumnPermutationIndex()));
			permutationIndices.put(new String(rootNode), tmpList);
			
			/**
			 * obtain root joint distribution
			 */
			tmpNodeName = (String)treeLevelListSubtree2.get(1).get(0);
			tmpNode = (Node)reconstructedTree.get(tmpNodeName);
			
			if(tmpNode.getNodeType().equals("Internal"))
			{
				rootBranchJointProb = extractBestFitMatrix(rootBranchChildToParentPermutations, 
						                                   rootLevelStructureCase2.get(0).getPreviousRowPermutationIndex(),
						                                   rootLevelStructureCase2.get(0).getCurrentColumnPermutationIndex());
			}
			else
			{
				rootBranchJointProb = extractBestFitMatrix(rootBranchChildToParentPermutations,
						                                   rootLevelStructureCase2.get(0).getPreviousRowPermutationIndex());
			}
                    
			/**
			 * subtree 1
			 */
		
			if(numOfLevels > 1)
			{
				for(j=1; j<numOfLevels; j++)
				{
					aTreeLevel = treeLevelList.get(j);
					numOfElementALevel = aTreeLevel.size();
					for(k=0; k<numOfElementALevel; k++)
					{
						tmpNodeName = (String)aTreeLevel.get(k);
						tmpNode = (Node)reconstructedTree.get(tmpNodeName);
						if(tmpNode.getNodeType().equals("Internal"))
						{
							//System.out.println("at results: node name: " + tmpNodeName + ": " + tmpNode.getPreviousCumulativeDifs());
							
							result += tmpNode.getPreviousCumulativeDifs();
							miniSumOfSquares.put(new String(tmpNodeName), new Double(tmpNode.getPreviousCumulativeDifs()));
						
							tmpList = new ArrayList<Integer>();
						
							tmpList.add(new Integer(tmpNode.getPreviousRowPermutationIndex()));
							tmpList.add(new Integer(tmpNode.getPreviousColumnPermutationIndex()));
							permutationIndices.put(new String(tmpNodeName), tmpList);
							/**
							 * added on April 17, 2009
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, 
									                         bestFitCondsMatrices, 
									                         tmpNode.getPreviousRowPermutationIndex(), 
									                         tmpNode.getPreviousColumnPermutationIndex());
							
							bestFitNonStationaryConds.put(tmpNodeName, tmpMatrix);
							
							/**
							 * added on April 17, 2009
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, 
									                         bestFitBHJoints, 
									                         tmpNode.getPreviousRowPermutationIndex(), 
									                         tmpNode.getPreviousColumnPermutationIndex());
							
							bestFitNonStationaryBHJoints.put(tmpNodeName, tmpMatrix);
							}
							else
							{
								/**
								 * added on April 17, 2009
								 */
								/**
								 * obtain internal - leaf branch matrix
								 * 1. get parent node
								 * 2. get parent column permutation
								 * 3. obtain matrix using parent column permutation
								 */
								if(j==1)
								{
									tmpColumnPermutation = rootLevelStructureCase2.get(0).getCurrentColumnPermutationIndex();
								}
								else
							    {
									tmpParentNodeName = tmpNode.getParent();
									tmpParentNode = (Node)reconstructedTree.get(tmpParentNodeName);
									tmpColumnPermutation = tmpParentNode.getPreviousColumnPermutationIndex();
							    }
								
								tmpMatrix = extractBestFitMatrix(tmpNodeName, bestFitCondsMatrices, tmpColumnPermutation);
								bestFitNonStationaryConds.put(tmpNodeName, tmpMatrix);
								
								/**
								 * added on June 30, 2011
								 */
								tmpMatrix = extractBestFitMatrix(tmpNodeName, bestFitBHJoints, tmpColumnPermutation);
								bestFitNonStationaryBHJoints.put(tmpNodeName, tmpMatrix);
						}
					
					}
				}	
			}
			/**
			 * subtree 2
			 */
			numOfLevels = treeLevelListSubtree2.size();
			
			if(numOfLevels > 1)
			{
				for(j=1; j<numOfLevels; j++)
				{
					aTreeLevel = treeLevelListSubtree2.get(j);
					numOfElementALevel = aTreeLevel.size();
					
					for(k=0; k<numOfElementALevel; k++)
					{
						tmpNodeName = (String)aTreeLevel.get(k);
						tmpNode = (Node)reconstructedTree.get(tmpNodeName);
						if(tmpNode.getNodeType().equals("Internal"))
						{
							result += tmpNode.getPreviousCumulativeDifs();
							miniSumOfSquares.put(new String(tmpNodeName), new Double(tmpNode.getPreviousCumulativeDifs()));
						
							tmpList = new ArrayList<Integer>();
						
							tmpList.add(new Integer(tmpNode.getPreviousRowPermutationIndex()));
							tmpList.add(new Integer(tmpNode.getPreviousColumnPermutationIndex()));
							permutationIndices.put(new String(tmpNodeName), tmpList);
							/**
							 * added on April 17, 2009
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, 
										                     bestFitCondsMatrices, 
										                     tmpNode.getPreviousRowPermutationIndex(), 
										                     tmpNode.getPreviousColumnPermutationIndex());
							
							bestFitNonStationaryConds.put(tmpNodeName, tmpMatrix);
							
							/**
							 * added on June 30, 2011
							 */
							tmpMatrix = extractBestFitMatrix(tmpNodeName, 
											                 bestFitBHJoints, 
											                 tmpNode.getPreviousRowPermutationIndex(), 
											                 tmpNode.getPreviousColumnPermutationIndex());
							
							bestFitNonStationaryBHJoints.put(tmpNodeName, tmpMatrix);
							}
							else
							{
								/**
								 * added on April 17, 2009
								 */
								/**
								 * obtain internal - leaf branch matrix
								 * 1. get parent node
								 * 2. get parent column permutation
								 * 3. obtain matrix using parent column permutation
								 */
								if(j==1)
								{
									tmpColumnPermutation = rootLevelStructureCase2.get(1).getPreviousRowPermutationIndex();
								}
								else
								{
									tmpParentNodeName = tmpNode.getParent();
									tmpParentNode = (Node)reconstructedTree.get(tmpParentNodeName);
									tmpColumnPermutation = tmpParentNode.getPreviousColumnPermutationIndex();
								}
								tmpMatrix = extractBestFitMatrix(tmpNodeName, bestFitCondsMatrices, tmpColumnPermutation);
								bestFitNonStationaryConds.put(tmpNodeName, tmpMatrix);
								/**
								 * added on June 30, 2011
								 */
								tmpMatrix = extractBestFitMatrix(tmpNodeName, bestFitBHJoints, tmpColumnPermutation);
								bestFitNonStationaryBHJoints.put(tmpNodeName, tmpMatrix);
						}
					}
				}	
			}
		}
		
		results.setMinimumSumOfSquares(result);
		results.setMinimumAtInternals(miniSumOfSquares);
		results.setPermutationIndices(permutationIndices);
		results.setItrCounter(itrCounter);
		/**
		 * Added on April 17, 2009
		 */
		computeBestFitNonStationaryDistribution(rootBranchJointProb);
		results.setBestFitNonStationaryJointProbMatricesAll(bestFitNonStationaryJointProbMatrices);
		results.setBestFitNonStationaryMarginalFrequenciesAll(bestFitNonStationaryMarginalFrequencies);
		/**
		 * added on June 30, 2011
		 */
		results.setBestFitBH(bestFitNonStationaryBHJoints);
		results.setReconstructedTree(reconstructedTree);
		results.setBestFitNonStationaryConds(bestFitNonStationaryConds);
		return results;
	}

	private void computeBestFitNonStationaryDistribution(double[][] rootBranchJointProb)
	{
		if(isInternalNode == false)
			computeBestFitNonStationaryDistributionRootingAtALeafOrRootNode(rootBranchJointProb);
		else
		{
			computeBestFitNonStationaryDistributionRootingAtAnInternalNodeSubtree1(rootBranchJointProb);
			/**
			 * since root frequencies has been computed in subtree 1, no computation is necessary for root node in
			 * subtree 2.
			 */
			computeBestFitNonStationaryDistributionRootingAtAnInternalNodeSubtree2();
		}
	}

	private void computeBestFitNonStationaryDistributionRootingAtAnInternalNodeSubtree2()
	{
		/**
		 * using multiplication and other nag interface in this class
		 */
		estimatePermutations anEstimator = new estimatePermutations();
		ArrayList<String> aLevelList = null;
		
		double[] rootFrequencies = null;
		double[] tmpFrequenices = null;
		
		double[][] tmpMatrix = null, tmpJointProb = null, tmpDiag = null;
		
		int numOfLevel = 0, numOfElementsALevel;
		int j, k;
		
		String tmpNodeName, parentNodeName;
		Node currentNode;
		
		/**
		 * subtree 2 stored in treeLevelList
		 */
		numOfLevel = treeLevelListSubtree2.size();
		for(j=1; j<numOfLevel; j++)
		{
			aLevelList = treeLevelListSubtree2.get(j);
			numOfElementsALevel = aLevelList.size();
			for(k=0; k< numOfElementsALevel; k++)
			{
				tmpNodeName = aLevelList.get(k);
				tmpMatrix = bestFitNonStationaryConds.get(tmpNodeName);
				

				currentNode = reconstructedTree.get(tmpNodeName);
				parentNodeName = currentNode.getParent();
				/**
				 * rootFrequencies is holding the frequencies of parent node
				 */
				rootFrequencies = bestFitNonStationaryMarginalFrequencies.get(parentNodeName);
				tmpDiag = anEstimator.convertVectorToDiag(rootFrequencies, anEstimator.numOfBases);
				tmpJointProb = anEstimator.computeMultiplicationOfTwoMatrices(tmpDiag, tmpMatrix);
				tmpFrequenices = anEstimator.computeNULLMarginalFrequencies(tmpJointProb, 2);
				bestFitNonStationaryJointProbMatrices.put(tmpNodeName, tmpJointProb);
				bestFitNonStationaryMarginalFrequencies.put(tmpNodeName, tmpFrequenices);
			}
		}
		
	}
	
	private void computeBestFitNonStationaryDistributionRootingAtAnInternalNodeSubtree1(double[][] rootBranchJointProb)
	{
		/**
		 * using multiplication and other nag interface in this class
		 */
		estimatePermutations anEstimator = new estimatePermutations();
		ArrayList<String> aLevelList = null;
		
		double[] rootFrequencies = anEstimator.computeNULLMarginalFrequencies(rootBranchJointProb, 1);
		double[] tmpFrequenices = null;
		
		double[][] tmpMatrix = null, tmpJointProb = null, tmpDiag = null;
		
		int numOfLevel = 0, numOfElementsALevel;
		int j, k;
		
		String tmpNodeName, parentNodeName;
		Node currentNode;
		
		/**
		 * subtree 1 stored in treeLevelList
		 */
		numOfLevel = treeLevelList.size();
		for(j=1; j<numOfLevel; j++)
		{
			aLevelList = treeLevelList.get(j);
			numOfElementsALevel = aLevelList.size();
			for(k=0; k< numOfElementsALevel; k++)
			{
				tmpNodeName = aLevelList.get(k);
				tmpMatrix = bestFitNonStationaryConds.get(tmpNodeName);
				
				if(j== 1 && k==0)
				{
					/**
					 * When handling first element of second level, the root frequencies has not been
					 * computed.  
					 */
					bestFitNonStationaryMarginalFrequencies.put(rootNode, rootFrequencies);
					tmpDiag = anEstimator.convertVectorToDiag(rootFrequencies, anEstimator.numOfBases);
					tmpJointProb = anEstimator.computeMultiplicationOfTwoMatrices(tmpDiag, tmpMatrix);
					tmpFrequenices = anEstimator.computeNULLMarginalFrequencies(tmpJointProb, 2);
					bestFitNonStationaryJointProbMatrices.put(tmpNodeName, tmpJointProb);
					bestFitNonStationaryMarginalFrequencies.put(tmpNodeName, tmpFrequenices);
				}
				else
				{
					currentNode = reconstructedTree.get(tmpNodeName);
					parentNodeName = currentNode.getParent();
					/**
					 * rootFrequencies is holding the frequencies of parent node
					 */
					rootFrequencies = bestFitNonStationaryMarginalFrequencies.get(parentNodeName);
					tmpDiag = anEstimator.convertVectorToDiag(rootFrequencies, anEstimator.numOfBases);
					tmpJointProb = anEstimator.computeMultiplicationOfTwoMatrices(tmpDiag, tmpMatrix);
					tmpFrequenices = anEstimator.computeNULLMarginalFrequencies(tmpJointProb, 2);
					bestFitNonStationaryJointProbMatrices.put(tmpNodeName, tmpJointProb);
					bestFitNonStationaryMarginalFrequencies.put(tmpNodeName, tmpFrequenices);
				}
			}
		}

	}

	private void computeBestFitNonStationaryDistributionRootingAtALeafOrRootNode(double[][] rootBranchJointProb)
	{
		/**
		 * using multiplication and other nag interface in this class
		 */
		estimatePermutations anEstimator = new estimatePermutations();
		ArrayList<String> aLevelList = null;
		
		double[] rootFrequencies = anEstimator.computeNULLMarginalFrequencies(rootBranchJointProb, 1);
		double[] tmpFrequenices = null;
		
		double[][] tmpMatrix = null, tmpJointProb = null, tmpDiag = null;
		
		int numOfLevel = 0, numOfElementsALevel;
		int j, k;
		
		String tmpNodeName, parentNodeName;
		Node currentNode;
		
		/**
		 * rooting at leaf/root to internal branch
		 */
		numOfLevel = treeLevelList.size();
		for(j=1; j<numOfLevel; j++)
		{
			aLevelList = treeLevelList.get(j);
			numOfElementsALevel = aLevelList.size();
			for(k=0; k< numOfElementsALevel; k++)
			{
				tmpNodeName = aLevelList.get(k);
				tmpMatrix = bestFitNonStationaryConds.get(tmpNodeName);
				
				if(j== 1)
				{
					bestFitNonStationaryMarginalFrequencies.put(rootNode, rootFrequencies);
					tmpDiag = anEstimator.convertVectorToDiag(rootFrequencies, anEstimator.numOfBases);
					tmpJointProb = anEstimator.computeMultiplicationOfTwoMatrices(tmpDiag, tmpMatrix);
					tmpFrequenices = anEstimator.computeNULLMarginalFrequencies(tmpJointProb, 2);
					bestFitNonStationaryJointProbMatrices.put(tmpNodeName, tmpJointProb);
					bestFitNonStationaryMarginalFrequencies.put(tmpNodeName, tmpFrequenices);
				}
				else
				{
					currentNode = reconstructedTree.get(tmpNodeName);
					parentNodeName = currentNode.getParent();
					/**
					 * rootFrequencies is holding the frequencies of parent node
					 */
					rootFrequencies = bestFitNonStationaryMarginalFrequencies.get(parentNodeName);
					tmpDiag = anEstimator.convertVectorToDiag(rootFrequencies, anEstimator.numOfBases);
					tmpJointProb = anEstimator.computeMultiplicationOfTwoMatrices(tmpDiag, tmpMatrix);
					tmpFrequenices = anEstimator.computeNULLMarginalFrequencies(tmpJointProb, 2);
					bestFitNonStationaryJointProbMatrices.put(tmpNodeName, tmpJointProb);
					bestFitNonStationaryMarginalFrequencies.put(tmpNodeName, tmpFrequenices);
				}
			}
		}

	}
	
	private double[][] extractBestFitMatrix(String inputNodeName, 
				                                           TreeMap<String, ArrayList<double[][]>> inputContainer, 
			                                                int rowPermutation, 
			                                                int colPermutation)
	{
		double[][] result = new double[4][4];
		double[][] tmpResults;
		
		ArrayList<double[][]> tmpPermutations = inputContainer.get(inputNodeName);
		
		int matrixIndex = 24*rowPermutation + colPermutation;
		int j, k;
		
		tmpResults = tmpPermutations.get(matrixIndex);
		
		for(j=0; j<4; j++)
		{
			for(k=0; k<4; k++)
			{
				result[j][k] = tmpResults[j][k];
			}
		}
		
		return result;
		
	}
	
	private double[][] extractBestFitMatrix(String inputNodeName, 
			                                                TreeMap<String, ArrayList<double[][]>> inputContainer, 
			                                                int inputColumnPermutation)
	{
		double[][] result = new double[4][4];
		
		ArrayList<double[][]> tmpPermutations = inputContainer.get(inputNodeName);
		
		double[][] tmpResult = tmpPermutations.get(inputColumnPermutation);
		
		int j, k; 
		
		for(j=0; j<4; j++)
		{
			for(k=0; k<4; k++)
			{
				result[j][k] = tmpResult[j][k];
			}
		}
		
		return result;
		

	}
	
	private double[][] extractBestFitMatrix(ArrayList<double[][]> inputContainer, 
															int inputColumnPermutation)
	{
		double[][] result = new double[4][4];
		double[][] tmpResult = inputContainer.get(inputColumnPermutation);

		int j, k; 

		for(j=0; j<4; j++)
		{
			for(k=0; k<4; k++)
			{
				result[j][k] = tmpResult[j][k];
			}
		}
		return result;
	}

	private double[][] extractBestFitMatrix(ArrayList<double[][]> inputContainer, 
			                                                int inputRowPermutation, 
			                                                int inputColumnPermutation)
	{
		double[][] result = new double[4][4];
		int tmpIndex = inputRowPermutation*24 + inputColumnPermutation;
		double[][] tmpResult = inputContainer.get(tmpIndex);

		int j, k; 

		for(j=0; j<4; j++)
		{
			for(k=0; k<4; k++)
			{
				result[j][k] = tmpResult[j][k];
			}
		}
		return result;
	}
	
	private double computeMinSSSubtree2()
	{
		double resultCurrent=0.0;
		int numOfLevels, numOfElementALevel, j, k;
		ArrayList<String> aList;
		Node tmpNode;
		String tmpNodeName;
		
		resultCurrent = rootLevelStructureCase2.get(1).getCurrentCumulativeDifs();
		
		numOfLevels = treeLevelListSubtree2.size();
		/**
		 * since the last level is always leaf nodes and all levels results are stored in internal nodes, 
		 * last level could be ignored.
		 */
		
		if(numOfLevels > 2)
		{
			for(j=1; j<(numOfLevels-1); j++)
			{
				aList = treeLevelListSubtree2.get(j);
				numOfElementALevel = aList.size();
				
				for(k=0; k<numOfElementALevel; k++)
				{
					tmpNodeName = (String)aList.get(k);
					tmpNode = (Node)reconstructedTree.get(tmpNodeName);
					if(tmpNode.getNodeType().equals("Internal"))
					{
						resultCurrent += tmpNode.getCurrentCumulativeDifs();			
					}
					
				}
			}
		}
		return resultCurrent;
	}
	
	private double computeMinSSSubtree1()
	{
		double resultCurrent=0.0;
		int numOfLevels, numOfElementALevel, j, k;
		ArrayList<String> aList;
		Node tmpNode;
		String tmpNodeName;
		
		resultCurrent = rootLevelStructureCase2.get(0).getCurrentCumulativeDifs();
		numOfLevels = treeLevelList.size();
		/**
		 * since the last level is always leaf nodes and all levels results are stored in internal nodes, 
		 * last level could be ignored.
		 */
		
		if(numOfLevels > 2)
		{
			for(j=1; j<(numOfLevels-1); j++)
			{
				aList = treeLevelList.get(j);
				numOfElementALevel = aList.size();
				
				for(k=0; k<numOfElementALevel; k++)
				{
					tmpNodeName = (String)aList.get(k);
					tmpNode = (Node)reconstructedTree.get(tmpNodeName);
					if(tmpNode.getNodeType().equals("Internal"))
					{
						resultCurrent += tmpNode.getCurrentCumulativeDifs();
					}
				}
			}
		}
		return resultCurrent;
	}
	
	private boolean isCurrentBetterThanPreviousCaseTwo()
	{
		double resultSubtree1, resultSubtree2, sumOfResults;
		
		resultSubtree1 = computeMinSSSubtree1();
		resultSubtree2 = computeMinSSSubtree2();
		
		sumOfResults = resultSubtree1 + resultSubtree2;
		
		/**
		 * 0: previous
		 * 1: current
		 */
		if(sumOfResults < currentAndPreviousMinSSs[0])
		{
			currentAndPreviousMinSSs[0] = currentAndPreviousMinSSs[1];
			currentAndPreviousMinSSs[1] = sumOfResults;
			return true;
		}
		return false;
	}
	
	private void getAShortestPathCaseTwo()
	{
		/**
		 * go through subtree 1 first. 
		 * the row index of top level will be used as top level initial row index
		 * of subtree 2.
		 */
		getAShortestPathCaseTwoSubtree1();
		/**
		 * subtree 2 will take the top level row index as top level initial row index
		 * for the iterations in subtree 2.
		 */
		getAShortestPathCaseTwoSubtree2();
		
	}

	private void updateRootingNodeSubtree2(String inputName)
	{
		String parentName;

		Node currentIntNode, nodeParent;
	
		int permutationIndexChild;
		
		double[] difsChild;
		//double tmpValue;
		
		currentIntNode = (Node)reconstructedTree.get(inputName);
		parentName = (String)currentIntNode.getParent();
		
		difsChild = differences.get(parentName);
		//System.out.println(childName);
		nodeParent = (Node)reconstructedTree.get(parentName);
		
		permutationIndexChild = nodeParent.getCurrentColumnPermutationIndex();
		if(treeLevelListSubtree2.size() > 2)
			difsChild = obtainWorkingDifsColumnFix(difsChild, permutationIndexChild);
		
		double shortestDif = difsChild[rootLevelStructureCase2.get(1).getCurrentRowPermutationIndex()];
		int permutationIndexResults = rootLevelStructureCase2.get(1).getCurrentRowPermutationIndex();
		/**
		 * update root node
		 */
		rootLevelStructureCase2.get(1).setPreviousCumulativeDif(rootLevelStructureCase2.get(1).getCurrentCumulativeDifs());
		rootLevelStructureCase2.get(1).setCurrentCumulativeDif(shortestDif);
		rootLevelStructureCase2.get(1).setPreviousColumnPermutationIndex(rootLevelStructureCase2.get(1).getCurrentColumnPermutationIndex());
		rootLevelStructureCase2.get(1).setCurrentColumnPermutationIndex(permutationIndexChild);		
		
		rootLevelStructureCase2.get(0).setPreviousRowPermutationIndex(rootLevelStructureCase2.get(0).getCurrentRowPermutationIndex());
		rootLevelStructureCase2.get(0).setCurrentRowPermutationIndex(permutationIndexResults);
		/**
		 * update left child
		 */
		if(treeLevelListSubtree2.size() > 2)
		{
			reconstructedTree.get(parentName).setPreviousRowPermutationIndex(reconstructedTree.get(parentName).getCurrentRowPermutationIndex());
			reconstructedTree.get(parentName).setCurrentRowPermutationIndex(permutationIndexResults);
		}
		
	}
	
	private void getAShortestPathCaseTwoSubtree2()
	{
		int numOfLevels = treeLevelListSubtree2.size();
		String nodeName, nodeType;
		ArrayList<String> aTreeLevel = null;
		Node aNode = null;
		
		int j, k, numOfElements, tmpIndex;
		
		for(j=1; j<numOfLevels; j++)
		{
			/**
			 * Iteration will start from last level
			 * tmpIndex = (total levels - 1) - j
			 * total levels: Index of last level
			 * j: number of levels has been gone through
			 */
			tmpIndex = numOfLevels - 1 - j;
			aTreeLevel = (ArrayList<String>)treeLevelListSubtree2.get(tmpIndex);
			numOfElements = aTreeLevel.size();
			for(k=0; k<numOfElements; k++)
			{
				nodeName = (String)aTreeLevel.get(k);
				aNode = (Node)reconstructedTree.get(nodeName);
				nodeType = aNode.getNodeType();
				
				//if((tmpIndex==0) || ((aNode.getParent().equals(rootNode))&&aNode.getNodeType()=="Internal"&&isRootParentOldRoot==false))
				if(tmpIndex==0)
				{
					//updatePermutationInformASpecialInternalNode(nodeName);
					updateRootingNodeSubtree2(nodeName);
				}
				else if(nodeType.equals("Internal"))
				{
					updatePermutationInformAnInternalNode(nodeName);
				}
			}
		}
		
	}
	
	private void updateRootingNodeSubtree1(String nodeName)
	{
		String childNameLeft, childNameRight, childNodeTypeLeft, childNodeTypeRight=null;

		Node currentIntNode, nodeLeft, nodeRight;
	
		int permutationIndexLeft, permutationIndexRight;
		
		double[] difsWorkingLeft, difsWorkingRight, difsLeft, difsRight;
		//double tmpValue;
		
		currentIntNode = (Node)reconstructedTree.get(nodeName);
		childNameLeft = (String)currentIntNode.getLeftChild();
		childNameRight = (String)currentIntNode.getRightChild();
		//permutationIndexInternal = currentIntNode.getRowPermutationIndex();
		
		difsLeft = differences.get(childNameLeft);
		difsRight = differences.get(childNameRight);
		
		
		/**
		 * left child
		 */
		nodeLeft = (Node)reconstructedTree.get(childNameLeft);
		childNodeTypeLeft = nodeLeft.getNodeType();
		/**
		 * right child
		 */
		nodeRight = (Node)reconstructedTree.get(childNameRight);
		childNodeTypeRight = nodeRight.getNodeType();
		
		if(childNodeTypeLeft.equals("Internal"))
		{
			permutationIndexLeft = nodeLeft.getCurrentColumnPermutationIndex();
			difsWorkingLeft = obtainWorkingDifsColumnFix(difsLeft, permutationIndexLeft);
		}
		else
		{
			difsWorkingLeft = differences.get(childNameLeft);
		}
		
		if(childNodeTypeRight.equals("Internal"))
		{
			permutationIndexRight = nodeRight.getCurrentColumnPermutationIndex();
			difsWorkingRight = obtainWorkingDifsColumnFix(difsRight, permutationIndexRight);
		}
		else
		{
			difsWorkingRight = differences.get(childNameRight);
		}
		//difsWorkingInternal = obtainWorkingDifsRowFix(difsInternal, permutationIndexInternal);
		
		//TreeMap<Integer, Double> shortestDifResults = obtainShortestDifASubtree(difsWorkingLeft, difsWorkingRight);
		TreeMap<Integer, Double> shortestDifResults = obtainShortestDifIncludeRootingBranch(difsWorkingLeft, difsWorkingRight, nodeName);
		
		double shortestDif = shortestDifResults.get(shortestDifResults.firstKey()).doubleValue();
		int permutationIndexResults = shortestDifResults.firstKey().intValue();
		//System.out.println("node name: " + "shortest: " + shortestDif);

		/**
		 * update root node information for subtree 1
		 */
		rootLevelStructureCase2.get(0).setPreviousCumulativeDif(rootLevelStructureCase2.get(0).getCurrentCumulativeDifs());
		rootLevelStructureCase2.get(0).setCurrentCumulativeDif(shortestDif);
		rootLevelStructureCase2.get(0).setPreviousColumnPermutationIndex(rootLevelStructureCase2.get(0).getCurrentColumnPermutationIndex());
		rootLevelStructureCase2.get(0).setCurrentColumnPermutationIndex(permutationIndexResults);
		
		/**
		 * update permutation information for subtree 2
		 */
		rootLevelStructureCase2.get(1).setPreviousRowPermutationIndex(rootLevelStructureCase2.get(1).getCurrentRowPermutationIndex());
		rootLevelStructureCase2.get(1).setCurrentRowPermutationIndex(permutationIndexResults);
		
		/**
		 * update left and right children 
		 */
		reconstructedTree.get(childNameLeft).setPreviousRowPermutationIndex(reconstructedTree.get(childNameLeft).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameLeft).setCurrentRowPermutationIndex(permutationIndexResults);
		reconstructedTree.get(childNameRight).setPreviousRowPermutationIndex(reconstructedTree.get(childNameRight).getCurrentRowPermutationIndex());
		reconstructedTree.get(childNameRight).setCurrentRowPermutationIndex(permutationIndexResults);
		
	}
	
	private void getAShortestPathCaseTwoSubtree1()
	{
		
		int numOfLevels = treeLevelList.size();
		String nodeName, nodeType;
		ArrayList<String> aTreeLevel = null;
		Node aNode = null;
		
		int j, k, numOfElements, tmpIndex;
		
		for(j=1; j<numOfLevels; j++)
		{
			/**
			 * tmpIndex = (total levels - 1) - j
			 * total levels: Index of last level
			 * j: number of levels has been gone through
			 */
			tmpIndex = numOfLevels - 1 - j;
			aTreeLevel = (ArrayList<String>)treeLevelList.get(tmpIndex);
			numOfElements = aTreeLevel.size();
			for(k=0; k<numOfElements; k++)
			{
				nodeName = (String)aTreeLevel.get(k);
				aNode = (Node)reconstructedTree.get(nodeName);
				nodeType = aNode.getNodeType();
				//System.out.println("here: " + nodeName);
				
				//if((tmpIndex==0) || ((aNode.getParent().equals(rootNode))&&aNode.getNodeType()=="Internal"&&isRootParentOldRoot==false))
				if(tmpIndex==0)
				{
					//updatePermutationInformASpecialInternalNode(nodeName);
					updateRootingNodeSubtree1(nodeName);
				}
				else if(nodeType.equals("Internal"))
				{
					updatePermutationInformAnInternalNode(nodeName);
				}
			}
		}
	
		
	}
	
	private void obtainTreeLevelNodesListCaseTwo(ArrayList<String> internalNodeList)
	{
		/**
		 * set root level nodes
		 */
		if(rootLevelStructureCase2 == null)
			rootLevelStructureCase2 = new ArrayList<Node>();
		
		Node tmpNode = null;
		
		/**
		 * first one is for subtree 1
		 */
		tmpNode = new Node(reconstructedTree.get(rootNode), 576);
		rootLevelStructureCase2.add(tmpNode);
		
		/**
		 * second one is for subtree 2
		 */
		tmpNode = new Node(reconstructedTree.get(rootNode), 576);
		rootLevelStructureCase2.add(tmpNode);
		
		obtainTreeLevelNodesListCaseTwoSubtree1(internalNodeList);
		obtainTreeLevelNodesListCaseTwoSubtree2(internalNodeList);
	}
	
	private void obtainTreeLevelNodesListCaseTwoSubtree1(ArrayList<String> internalNodeList)
	{
		String tmpLeft=null, tmpRight=null;
		Node tmpNode = null;
		ArrayList<String> currentLevel = null;
		ArrayList<String> previousLevel = new ArrayList<String>();
		int previousLevelSize, counter = 0;
		
		if(treeLevelList == null)
			treeLevelList = new ArrayList<ArrayList<String>>();
		
		/**
		 * add root and its children.
		 */
		tmpNode = (Node)reconstructedTree.get(rootNode);
		tmpLeft = tmpNode.getLeftChild();
		tmpRight = tmpNode.getRightChild();
		
		currentLevel = new ArrayList<String>();
		currentLevel.add(rootNode);
		treeLevelList.add(currentLevel);
		counter++;
		/**
		 * add root's children
		 */
		currentLevel = new ArrayList<String>();
		currentLevel.add(tmpLeft);
		currentLevel.add(tmpRight);
		
		treeLevelList.add(currentLevel);
		counter++;
		
		previousLevel.addAll(currentLevel);
		
		while(true)
		{
			previousLevelSize = previousLevel.size();
			
			currentLevel = new ArrayList<String>();
			
			for(int j=0; j<previousLevelSize; j++)
			{
				String tmpNodeName = previousLevel.get(j);
								
				tmpNode = (Node)reconstructedTree.get(tmpNodeName);
				if(tmpNode.getNodeType().equals("Internal"))
				{
					tmpLeft = tmpNode.getLeftChild();
					tmpRight = tmpNode.getRightChild();
					currentLevel.add(tmpLeft);
					currentLevel.add(tmpRight);
					
				}
			}
			
			if(currentLevel.size() != 0)
			{
				counter++;
				treeLevelList.add(currentLevel);
				previousLevel.clear();
				previousLevel.addAll(currentLevel);
				//currentLevel = new ArrayList<String>();
			}
			else
			{
				break;
			}
		}
		
	}
	
	private void obtainTreeLevelNodesListCaseTwoSubtree2(ArrayList<String> internalNodeList)
	{
		//ArrayList<ArrayList<String>> results = new ArrayList<ArrayList<String>>();
		
		String tmpLeft=null, tmpRight=null, tmpParent=null;
		Node tmpNode = null;
		ArrayList<String> currentLevel = null;
		ArrayList<String> previousLevel = new ArrayList<String>();
		int previousLevelSize, counter = 0;
		
		if(treeLevelListSubtree2 == null)
			treeLevelListSubtree2 = new ArrayList<ArrayList<String>>();
		
		/**
		 * find out the parent of the root.
		 */
		tmpNode = (Node)reconstructedTree.get(rootNode);
		tmpParent = tmpNode.getParent();
		
		/**
		 * found the root's parent
		 * put both root node and this parent node into treeLevelList
		 * level 0: root
		 * level 1: parent
		 */
		currentLevel = new ArrayList<String>();
		currentLevel.add(rootNode);
		treeLevelListSubtree2.add(currentLevel);
		counter++;
		currentLevel = new ArrayList<String>();
		currentLevel.add(tmpParent);
		treeLevelListSubtree2.add(currentLevel);
		counter++;
		
		previousLevel.addAll(currentLevel);
		
		while(true)
		{
			previousLevelSize = previousLevel.size();
			
			currentLevel = new ArrayList<String>();
			
			for(int j=0; j<previousLevelSize; j++)
			{
				String tmpNodeName = previousLevel.get(j);
								
				tmpNode = (Node)reconstructedTree.get(tmpNodeName);
				if(tmpNode.getNodeType().equals("Internal"))
				{
					tmpLeft = tmpNode.getLeftChild();
					tmpRight = tmpNode.getRightChild();
					currentLevel.add(tmpLeft);
					currentLevel.add(tmpRight);
					
				}
			}
			
			if(currentLevel.size() != 0)
			{
				counter++;
				treeLevelListSubtree2.add(currentLevel);
				previousLevel.clear();
				previousLevel.addAll(currentLevel);
				//currentLevel = new ArrayList<String>();
			}
			else
			{
				break;
			}
		}
		
	}
	
	private void updateRootingNodeCaseOne(String inputName)
	{
		String childName;

		Node currentIntNode, nodeChild;
	
		int permutationIndexChild;
		
		double[] difsChild;
		//double tmpValue;
		
		currentIntNode = (Node)reconstructedTree.get(inputName);
		childName = (String)currentIntNode.getLeftChild();
		
		difsChild = differences.get(childName);
		
		nodeChild = (Node)reconstructedTree.get(childName);
		
		permutationIndexChild = nodeChild.getCurrentColumnPermutationIndex();
			
		double shortestDif = difsChild[permutationIndexChild];
		int permutationIndexResults = permutationIndexChild;
		//System.out.println("node name: " + inputName + " shortest: " + shortestDif);

		/**
		 * update root node
		 */
		currentIntNode.setPreviousCumulativeDif(reconstructedTree.get(inputName).getCurrentCumulativeDifs());
		currentIntNode.setCurrentCumulativeDif(shortestDif);
		currentIntNode.setPreviousColumnPermutationIndex(reconstructedTree.get(inputName).getCurrentColumnPermutationIndex());
		currentIntNode.setCurrentColumnPermutationIndex(permutationIndexResults);		
		//System.out.println("node inform: " + currentIntNode.toString());
		//System.out.println("node inform: " + currentIntNode.toString());
		/**
		 * update left child
		 */
		//reconstructedTree.get(childName).setPreviousRowPermutationIndex(reconstructedTree.get(childName).getCurrentRowPermutationIndex());
		//reconstructedTree.get(childName).setCurrentRowPermutationIndex(permutationIndexResults);

	}
	
	private void getAShortestPathCaseOne()
	{
		int numOfLevels = treeLevelList.size();
		String nodeName, nodeType;
		ArrayList<String> aTreeLevel = null;
		Node aNode = null;
		
		int j, k, numOfElements, tmpIndex;
		
		for(j=0; j<numOfLevels; j++)
		{
			/**
			 * Iteration will start from last level
			 * tmpIndex = (total levels - 1) - j
			 * total levels: Index of last level
			 * j: number of levels has been gone through
			 */
			tmpIndex = numOfLevels - 1 - j;
			aTreeLevel = (ArrayList<String>)treeLevelList.get(tmpIndex);
			numOfElements = aTreeLevel.size();
			for(k=0; k<numOfElements; k++)
			{
				nodeName = (String)aTreeLevel.get(k);
				aNode = (Node)reconstructedTree.get(nodeName);
				nodeType = aNode.getNodeType();
				
				//if((tmpIndex==0) || ((aNode.getParent().equals(rootNode))&&aNode.getNodeType()=="Internal"&&isRootParentOldRoot==false))
				if(tmpIndex==0)
				{
					//updatePermutationInformASpecialInternalNode(nodeName);
					updateRootingNodeCaseOne(nodeName);
				}
				else if(nodeType.equals("Internal"))
				{
					updatePermutationInformAnInternalNode(nodeName);
				}
			}
		}
		
	}
	
	private boolean isCurrentBetterThanPreviousCaseOne()
	{
		double resultCurrent=0.0, resultPrevious = 0.0;
		int numOfLevels, numOfElementALevel, j, k;
		ArrayList<String> aList;
		Node tmpNode;
		String tmpNodeName;
		
		numOfLevels = treeLevelList.size();
		/**
		 * since the last level is always leaf nodes and all levels results are stored in internal nodes, 
		 * last level could be ignored.
		 */
		tmpNode = (Node)reconstructedTree.get(rootNode);
		
		resultCurrent += tmpNode.getCurrentCumulativeDifs();
		resultPrevious += tmpNode.getPreviousCumulativeDifs();
		
		for(j=1; j<(numOfLevels-1); j++)
		{
			aList = treeLevelList.get(j);
			numOfElementALevel = aList.size();
			
			/**
			 * last level contains all leaf nodes. 
			 * no necessary to go through this level.
			 */
			for(k=0; k<numOfElementALevel; k++)
			{
				tmpNodeName = (String)aList.get(k);
				tmpNode = (Node)reconstructedTree.get(tmpNodeName);
				if(tmpNode.getNodeType().equals("Internal"))
				{
					resultCurrent += tmpNode.getCurrentCumulativeDifs();
					resultPrevious += tmpNode.getPreviousCumulativeDifs();				
				}
						
			}		
		}
		
		if(resultCurrent < resultPrevious)
			return true;
		else
			return false;
	}
	/**
	 * this is a case that rooting at a leaf/old root. 
	 * The key is to find out the child of the root, and put the child as the first level of 
	 * the entire tree.
	 * @param internalNodeList
	 */
	private void obtainTreeLevelNodesListCaseOne(ArrayList<String>     internalNodeList)
	{
		//ArrayList<ArrayList<String>> results = new ArrayList<ArrayList<String>>();
		
		String tmpLeft=null, tmpRight=null, tmpIntNodeName=null, tmpParent=null;
		Node tmpNode = null;
		ArrayList<String> currentLevel = null;
		ArrayList<String> previousLevel = new ArrayList<String>();
		int index, conSize, previousLevelSize, counter = 0;
		
		if(treeLevelList == null)
			treeLevelList = new ArrayList<ArrayList<String>>();
		
		/**
		 * find out the child of the root.
		 */
		conSize = internalNodeList.size();
		for(index=0; index<conSize; index++)
		{
			tmpIntNodeName = (String)internalNodeList.get(index);
			//if(reconstructedTree.containsKey(tmpIntNodeName))
			//	System.out.println("yes");
			tmpNode = (Node)reconstructedTree.get(tmpIntNodeName);
			tmpParent = tmpNode.getParent();
			if(tmpParent.equals(rootNode))
			{
				/**
				 * found the root's child
				 * put both root node and this child node into treeLevelList
				 * level 0: root
				 * level 1: child
				 */
				currentLevel = new ArrayList<String>();
				currentLevel.add(rootNode);
				treeLevelList.add(currentLevel);
				counter++;
				currentLevel = new ArrayList<String>();
				currentLevel.add(tmpIntNodeName);
				treeLevelList.add(currentLevel);
				counter++;
				break;
			}
		}
		
		previousLevel.addAll(currentLevel);
		
		while(true)
		{
			previousLevelSize = previousLevel.size();
			
			currentLevel = new ArrayList<String>();
			
			for(int j=0; j<previousLevelSize; j++)
			{
				String tmpNodeName = previousLevel.get(j);
								
				tmpNode = (Node)reconstructedTree.get(tmpNodeName);
				if(tmpNode.getNodeType().equals("Internal"))
				{
					tmpLeft = tmpNode.getLeftChild();
					tmpRight = tmpNode.getRightChild();
					currentLevel.add(tmpLeft);
					currentLevel.add(tmpRight);
					
				}
			}
			
			if(currentLevel.size() != 0)
			{
				counter++;
				treeLevelList.add(currentLevel);
				previousLevel.clear();
				previousLevel.addAll(currentLevel);
				//currentLevel = new ArrayList<String>();
			}
			else
			{
				break;
			}
		}
				
	}

	private void reconstructTreeRootingAtANode(String                                        oldRoot, 
                                               String                                        newRoot, 
                                               ArrayList<String>                             internalNodeList,
                                               ArrayList<String>                             leafNodeList,
                                               TreeMap<String, NodeStructure>                aTree,
                                               TreeMap<String, ContainerForDifTwoDirections> inputDifferences,
                                               TreeMap<String, ContainerTwoArrayListDouble> bestFitConds, 
                                               TreeMap<String, ArrayList<double[][]>> childToParentPermutations,
                                               TreeMap<String, ArrayList<double[][]>> parentToChildPermutations
                                               )
	{
		/**
		 * set isInternalNode flag  and root node
		 */
		isInternalNode = isAnInternalNode(newRoot, internalNodeList);
		rootNode = new String(newRoot);
		oldRootNode = new String(oldRoot);

		String currentIntNode="", currentIntNodesOldParent="", currentIntPreviousNode="";
		Node tmpNode;
		NodeStructure tmpNodeStructure;
		boolean continueClimbing = true;
		int j, conSize;//, k;

		if(reconstructedTree == null)
			reconstructedTree = new TreeMap<String, Node>();

		conSize = internalNodeList.size();
		
		/**
		 * special case: rooting at old root
		 */
		if(oldRoot == newRoot)
		{
			tmpNode = new Node();
			tmpNode.setNodeName(oldRoot);
			
			for(j=0; j<conSize; j++)
			{
				currentIntNode = (String)internalNodeList.get(j);
				tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);
				currentIntNodesOldParent = (String)tmpNodeStructure.getParentNode();
				if(currentIntNodesOldParent.equals(oldRootNode))
				{
					break;
				}
			}
			tmpNode.setLeftChild(currentIntNode);
			tmpNode.setRightChild("");
			tmpNode.setNodeType("");
			tmpNode.setEvolutionaryDirection(-1);
			tmpNode.setParent("");
			reconstructedTree.put(oldRoot, tmpNode);
			differences.put(oldRoot, inputDifferences.get(currentIntNode).getParentToChild());
			/**
			 * added on April 17, 2009
			 */
			bestFitCondsMatrices.put(oldRoot, bestFitConds.get(currentIntNode).getParentToChild());
			/**
			 * added on June 30, 2011
			 */
			bestFitBHJoints.put(oldRoot, parentToChildPermutations.get(oldRootNode));
			rootBranchChildToParentPermutations = parentToChildPermutations.get(currentIntNode);
			
			obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences, bestFitConds, parentToChildPermutations);
			obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences, bestFitConds, parentToChildPermutations);
			
			return;
		}
		/**
		 * handle special case that newRoot's parent is oldRoot.
		 * this is the case that no tree reconstruction is necessary.
		 * actions necessary are 
		 * 1. setup root node
		 * 2. copy all nodes into reconstructed tree container
		 */
		conSize = internalNodeList.size();
		for(j=0; j<conSize; j++)
		{
			currentIntNode = (String)internalNodeList.get(j);
			tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);
			currentIntNodesOldParent = (String)tmpNodeStructure.getParentNode();
			if(currentIntNodesOldParent.equals(oldRootNode)&&currentIntNode.equals(rootNode))
			{
				/**
				 * it is the case of new root's parent is old root
				 */
				isRootParentOldRoot = true;
				tmpNode = new Node();
				tmpNode.setNodeName(newRoot);
				tmpNodeStructure = (NodeStructure)aTree.get(rootNode);

				tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());	
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
				/**
				 * april 10, 2009
				 * parent node needs this information.
				 */
				tmpNode.setParent(tmpNodeStructure.getParentNode());
				tmpNode.setEvolutionaryDirection(-1);
				
				reconstructedTree.put(newRoot, tmpNode);
				rootBranchChildToParentPermutations = childToParentPermutations.get(rootNode);
				
				/**
				 * add old root to the tree as a regular node which will have
				 * the difference of ctp of branch (old root, new root)
				 */
				tmpNode = new Node();
				tmpNode.setNodeName(oldRoot);
				tmpNode.setLeftChild("");
				tmpNode.setRightChild("");
				tmpNode.setParent(newRoot);
				tmpNode.setNodeType("Leaf");
				/**
				 * from child to parent
				 */
				tmpNode.setEvolutionaryDirection(1);
				differences.put(oldRoot, inputDifferences.get(rootNode).getChildToParent());
				/**
				 * added on April 17, 2009
				 */
				bestFitCondsMatrices.put(oldRoot, bestFitConds.get(rootNode).getChildToParent());
				
				/**
				 * added on June 30, 2011
				 */
				bestFitBHJoints.put(oldRoot, childToParentPermutations.get(rootNode));
				
				reconstructedTree.put(oldRoot, tmpNode);
				/**
				 * modified on June 30, 2011
				 */			
				obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences, bestFitConds, parentToChildPermutations);
				obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences, bestFitConds, parentToChildPermutations);
				//obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences, bestFitConds, childToParentPermutations, parentToChildPermutations);
				//obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences, bestFitConds, childToParentPermutations, parentToChildPermutations);
				/**
				 * add on Mar.28, 2009
				 */
				obtainRootingBranchDifs(inputDifferences, newRoot);
				//printTree(internalNodeList, leafNodeList);
				return;
			}
		}

		/**
		 * get newRoot's parent
		 */
		tmpNodeStructure = (NodeStructure)aTree.get(newRoot);
		currentIntNode = (String)tmpNodeStructure.getParentNode();
		currentIntPreviousNode = newRoot;

		/** 
		 * set current root
		 */
		if(isInternalNode == true)
		{
			/**
			 * set new root
			 */
			tmpNode = new Node();
			tmpNode.setNodeName(newRoot);
			tmpNode.setParent(aTree.get(newRoot).getParentNode());
			tmpNode.setLeftChild(aTree.get(newRoot).getChildNode1());
			tmpNode.setRightChild(aTree.get(newRoot).getChildNode2());
			tmpNode.setEvolutionaryDirection(-1);
			differences.put(newRoot, inputDifferences.get(newRoot).getChildToParent());
			/**
			 * added on April 17, 2009
			 */
			bestFitCondsMatrices.put(newRoot, bestFitConds.get(newRoot).getChildToParent());
			rootBranchChildToParentPermutations = childToParentPermutations.get(newRoot);
			
			reconstructedTree.put(newRoot, tmpNode);
			/**
			 * added on June 30, 2011
			 */
			bestFitBHJoints.put(newRoot, childToParentPermutations.get(newRoot));
			
			
		}
		else
		{
			tmpNode = new Node();
			tmpNode.setNodeName(newRoot);
			tmpNode.setParent(aTree.get(newRoot).getParentNode());
			tmpNode.setLeftChild(currentIntNode);
			tmpNode.setRightChild("");
			tmpNode.setEvolutionaryDirection(-1);
			reconstructedTree.put(newRoot, tmpNode);
			differences.put(newRoot, inputDifferences.get(newRoot).getChildToParent());
			/**
			 * added on April 17, 2009
			 */
			bestFitCondsMatrices.put(newRoot, bestFitConds.get(newRoot).getChildToParent());
			rootBranchChildToParentPermutations = childToParentPermutations.get(newRoot);
			/**
			 * added on June 30, 2011
			 */
			bestFitBHJoints.put(newRoot, childToParentPermutations.get(newRoot));
						
		}
		/**
		 * handle special parent node of new root in the sense that
		 * the parent node connected directly with old root
		 */
		tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);
		currentIntNodesOldParent = (String)tmpNodeStructure.getParentNode();
		if(currentIntNodesOldParent.equals(oldRoot))
		{
			tmpNode = new Node();
			tmpNode.setNodeName(currentIntNode);
			tmpNode.setLeftChild(oldRoot);
			if(tmpNodeStructure.getChildNode1().equals(newRoot))
				tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
			else
				tmpNode.setRightChild(tmpNodeStructure.getChildNode1());
			tmpNode.setEvolutionaryDirection(1);
			tmpNode.setParent(newRoot);
			tmpNode.setNodeType("Internal");
			reconstructedTree.put(currentIntNode, tmpNode);
			differences.put(currentIntNode, inputDifferences.get(newRoot).getChildToParent());
			/**
			 * added April 17, 2009
			 */
			bestFitCondsMatrices.put(currentIntNode, bestFitConds.get(newRoot).getChildToParent());
			/**
			 * added on June 30, 2011
			 */
			bestFitBHJoints.put(currentIntNode, childToParentPermutations.get(newRoot));			
			
			/**
			 * set up for old root: now it is a leaf node
			 */
			tmpNode = new Node();
			tmpNode.setNodeName(oldRoot);
			tmpNode.setNodeType("Leaf");
			tmpNode.setLeftChild("");
			tmpNode.setRightChild("");
			tmpNode.setParent(currentIntNode);			
			reconstructedTree.put(oldRoot, tmpNode);
			
			/**
			 * set up differences for old Root
			 */ 
			differences.put(oldRoot, inputDifferences.get(currentIntNode).getChildToParent());
			/**
			 * added on April 17, 2009
			 */
			bestFitCondsMatrices.put(oldRoot, bestFitConds.get(currentIntNode).getChildToParent());
			/**
			 * added on June 30, 2011
			 */
			bestFitBHJoints.put(oldRoot, childToParentPermutations.get(currentIntNode));
			
			obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences, bestFitConds, parentToChildPermutations);
			obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences, bestFitConds, parentToChildPermutations);
			//printTree(internalNodeList, leafNodeList);

			/**
			 * add on Mar.28, 2009
			 */
			obtainRootingBranchDifs(inputDifferences, newRoot);
			//printTree(internalNodeList, leafNodeList);
			
			return;
		}
		while(continueClimbing)
		{
			/**
			 * processing parent node
			 */
			tmpNodeStructure = (NodeStructure)aTree.get(currentIntNode);			
			/**
			 * get parent's parent
			 */
			currentIntNodesOldParent = tmpNodeStructure.getParentNode();

			tmpNode = new Node();
			tmpNode.setNodeName(currentIntNode);
			/**
			 * evolutionary direction: from child to parent
			 */
			tmpNode.setEvolutionaryDirection(1);
			/**
			 * if newRoot's parent is oldRoot, only this internalNode needs to be reset. 
			 * Other internal nodes are just being copied to new container.  
			 */
			if(currentIntNodesOldParent.equals(oldRoot))
			{
				if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode1()))
				{
					tmpNode.setLeftChild(oldRoot);
					tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
					tmpNode.setParent(currentIntPreviousNode);
					tmpNode.setNodeType("Internal");
				}
				else if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode2()))
				{	
					tmpNode.setLeftChild(oldRoot);
					tmpNode.setRightChild(tmpNodeStructure.getChildNode1());
					tmpNode.setParent(currentIntPreviousNode);
					tmpNode.setNodeType("Internal");
		
				}	
				else
				{
					System.out.println("error 1 in reconstructTreeLeafAsRoot\n");
					System.exit(-1);
				}
	
				reconstructedTree.put(currentIntNode, tmpNode);
				differences.put(currentIntNode, inputDifferences.get(currentIntPreviousNode).getChildToParent());
				/**
				 * added on April 17, 2009
				 */
				bestFitCondsMatrices.put(currentIntNode, bestFitConds.get(currentIntPreviousNode).getChildToParent());
				/**
				 * added on June 30, 2011
				 */
				bestFitBHJoints.put(currentIntNode, childToParentPermutations.get(currentIntPreviousNode));

				/**
				 * set up for old root: now it is leaf node
				 */
				tmpNode = new Node();
				tmpNode.setNodeName(oldRoot);
				tmpNode.setLeftChild("");
				tmpNode.setRightChild("");
				tmpNode.setNodeType("Leaf");
				tmpNode.setParent(currentIntNode);
				reconstructedTree.put(oldRoot, tmpNode);

				differences.put(oldRoot, inputDifferences.get(currentIntNode).getChildToParent());
				/**
				 * added on April 17, 2009
				 */
				bestFitCondsMatrices.put(oldRoot, bestFitConds.get(currentIntNode).getChildToParent());
				/**
				 * added on June 30, 2011
				 */
				bestFitBHJoints.put(oldRoot, childToParentPermutations.get(currentIntNode));

				continueClimbing = false;
			}
			else
			{
				/**
				 * Set up current node structure
				 */
				tmpNode.setParent(currentIntPreviousNode);
				if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode1()))
				{
					tmpNode.setRightChild(tmpNodeStructure.getChildNode2());
					tmpNode.setLeftChild(currentIntNodesOldParent);
				}
				else if(currentIntPreviousNode.equals(tmpNodeStructure.getChildNode2()))
				{
					tmpNode.setLeftChild(tmpNodeStructure.getChildNode1());
					tmpNode.setRightChild(currentIntNodesOldParent);				
				}
				else
				{
					System.out.println("error 2 in reconstructTreeLeafAsRoot\n");
					System.exit(-1);
				}
				tmpNode.setNodeType("Internal");
				reconstructedTree.put(currentIntNode, tmpNode);
				/** 
				 *set up differences
				 */
				differences.put(currentIntNode, inputDifferences.get(currentIntPreviousNode).getChildToParent());
				/**
				 * added on April 17, 2009
				 */
				bestFitCondsMatrices.put(currentIntNode, bestFitConds.get(currentIntPreviousNode).getChildToParent());
				/**
				 * added on June 30, 2011
				 */
				bestFitBHJoints.put(currentIntNode, childToParentPermutations.get(currentIntPreviousNode));

				/**
				 * move forward and switch parameters
				 */
				currentIntPreviousNode = currentIntNode;
				currentIntNode = currentIntNodesOldParent;
			}
		}

		obtainAnEntireNodeList(internalNodeList, aTree, 1, inputDifferences, bestFitConds, parentToChildPermutations);
		obtainAnEntireNodeList(leafNodeList, aTree, 2, inputDifferences, bestFitConds, parentToChildPermutations);	
		//printTree(internalNodeList, leafNodeList);
		/**
		 * add on Mar.28, 2009
		 */
		obtainRootingBranchDifs(inputDifferences, newRoot);
		//printTree(internalNodeList, leafNodeList);
	}

	/**
	 * end class SearchShortestPath
	 */
	}