//import java.io.BufferedReader;
//import java.io.InputStreamReader;
import java.util.*;
import java.io.*;

public class estimatePermutations
{

    /**
     * Native functions
     */
	protected native void callingDGEMM(char transa, char transb, int m, int n, int k, double alpha, double[] a, int lda, double[] b, int ldb, double beta, double[] c, int ldc);
    protected native void callingDGEEV(char jobvl, char jobvr, int n, double[] a, int lda, double[] wr, double[] wi, double[] vl, int ldvl, double[] vr, int ldvr, double[] work, int lwork, int info);
    protected native void callingDGETRF(int m, int n, double[] a, int lda, int[] ipiv, int info); 
    protected native void callingDSYEV(char jobz, char uplo, int n, double[] a, int lda, double[] w, double[] work, int lwork, int info);
    protected native void callingDGETRI(int n, double[] a, int lda, int[] ipiv, double[] work, int lwork, int info);
    protected native void callingZGETRF(int m, int n, double[] realPart, double[] imaginaryPart, int lda, int[] ipiv, int info);
    protected native void callingZGETRI(int n, double[] realPart, double[] imaginaryPart, int lda, int[] ipiv, int lwork, int info);

    /**
     * The library created should have the name "libcallingNAGInterface.so"
     */
   static
   {
	   System.loadLibrary("callingLapackInterface");
   }
    /**
	 * Parameters:
	 */
	/**
	 * Data file path
	 */
	String dataPath = null;
	/**
	 * Data file path
	 */	
	String treeFilePath  = null;
	/**
	 * tree notation
	 */
	String treeNotation=null;
	/**
	 * Number of taxa in the data
	 */
	int numOfTaxa   = 0; 
	/**
	 * Container which holds the tree
	 */
	 TreeMap<String, NodeStructure> phylogeneticTree = new TreeMap<String, NodeStructure>();
	 /**
	  * Internal and leaf nodes containers
	  */
	 ArrayList<String> leafNodesList = null;
	 ArrayList<String> intNodesList = null;
	 /**
	  * Root
	  */
	 String rootNode = null;
	 
	 /**
	  * Data structures
	  */
	 TreeMap<String, BranchDetails> branchList = new TreeMap<String, BranchDetails>();

	    /**
	     * handle permutations
	     */
	 /**
	  * Number of base states
	  */
	 int numOfBases = 4;
	 
	 /**
	  * Variables for permutations
	  */
    TreeMap<String, ArrayList<double[][]>> permutationsParentToChild = new TreeMap<String, ArrayList<double[][]>>();
    TreeMap<String, ArrayList<double[][]>> permutationsChildToParent = new TreeMap<String, ArrayList<double[][]>>();
    TreeMap<String, ContainerForDifTwoDirections> differences = new TreeMap<String, ContainerForDifTwoDirections>();
    TreeMap<String, ContainerTwoArrayListDouble> reconstructedConditionalProb = new TreeMap<String, ContainerTwoArrayListDouble>();
    TreeMap<String, ContainerTwoArrayListDouble> reconstructedJointProb = new TreeMap<String, ContainerTwoArrayListDouble>();
	    
    TreeMap<String, ContainerShortestPathSearchResults> shortestPathReconstructedTrees = new TreeMap<String, ContainerShortestPathSearchResults>();
    TreeMap<String,	ContainerShortestPathSearchResults> shortestPathReconstructedTreesWithoutLevel = new TreeMap<String, ContainerShortestPathSearchResults>(); 

    TreeMap<String, ArrayList<ArrayList<String>>> treeLevelList = new TreeMap<String, ArrayList<ArrayList<String>>>();
    TreeMap<String, TreeMap<String, double[][]>> bestFitPermutationMatrices = new TreeMap<String, TreeMap<String,double[][]>>();
    TreeMap<String, TreeMap<String, double[][]>> bestFitJointProbabilityMatrices = new TreeMap<String, TreeMap<String,double[][]>>();

    TreeMap<String,	ContainerShortestPathSearchResults> shortestPathReconstructedTreesRootingAtANode = new TreeMap<String, ContainerShortestPathSearchResults>(); 
    
    TreeMap<String, double[][]> bestNSGTRJoints = new TreeMap<String, double[][]>();
    
    /**
     * Edge lenghts: first is from the best root to leaf; second is the opposite direction; third one is the average.
     */
	TreeMap<String, double[]> edgeLengths = new TreeMap<String, double[]>();

	/**
	 * Options
	 */
	boolean allRootingPosition = false, defaultPosition=true;
	String selectedRootingPosition = null;
	String bestRoot = null;
	int outputOptions = 3;//both joints and nsgtr model parameters
	
    /**
     * compute the differences of permutations of all branches
     * go with the internal nodes first and then the leaf nodes
     */
	public void readParameters()
	{
		BufferedReader aReader = new BufferedReader(new InputStreamReader(System.in));
		String tmp;
		int numOfLeafNodes = 0;
		int numOfIntNodes = 0;
		int counter=0;
		int j;
		boolean tmpIndicator = false;
		/**
		 * parameters:
		 * 1. number of taxa
		 * 2. path of file: the path of input file is also assumed the path of output file
		 * 
		 */
		/**
		 * Greetings
		 */
		System.out.println();
		System.out.println();
		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		System.out.println("%  Estimate permutations for BH estimates.       %");
		System.out.println("%  This application is under GNU GPL lincense.   %");
		System.out.println("%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%");
		System.out.println();
		System.out.println();
		
		
		System.out.println("**************** Please input parameters **************");

		/**
		 * Input parameters
		 */
		try
		{
			//System.out.println("How many taxa are in the data:");
			//numOfTaxa = Integer.parseInt(aReader.readLine());
			
			System.out.println("Path of BH estimates (Format in Jayswal et al 2005 only): ");
			dataPath = aReader.readLine();
			
			System.out.println("Path of tree file (Newick format only): ");
			treeFilePath = aReader.readLine();
		
			getTreeTopology();
			numOfLeafNodes = leafNodesList.size();
			numOfIntNodes = intNodesList.size();
			/**
			 * obtain the number of taxa
			 */
			numOfTaxa = numOfIntNodes + numOfLeafNodes + 1;
			
			createJPGFile(treeNotation);
			
			//System.out.println("Estimate all rooting positions (Y/N)");
			System.out.println();
			System.out.println();
			System.out.println();
			System.out.println();
			
			System.out.println("**************************************************");
			System.out.println("* Please select root positions:                  *");
			System.out.println("*                                                *");
			System.out.println("* Please use 'tree.jpg' in working directory as  *");
			System.out.println("* a reference when selecting following options.  *");
			System.out.println("* All nodes, internal and terminal nodes, are    *");
			System.out.println("* possible rooting positions.                    *");
			System.out.println("**************************************************");
			
			System.out.println();
			while(tmpIndicator == false)
			{
				System.out.println("Estimated root or select roots from a list");
				System.out.println("(The estimated root is the one which gives the minimum SS among all rooting positions.)");
				System.out.println("1. estimated root");
				System.out.println("2. select roots from a list");
				tmp = aReader.readLine();
				
				if(tmp.equals("1"))
				{
					defaultPosition = true;//estimated root only
					tmpIndicator = true;
				}
				else if(tmp.equals("2"))
				{
					defaultPosition = false;
				
					System.out.println("Please select one or multiple root positions from the following list:");
					System.out.println(" 1. Use white space or comma as delimter when selecting rooting position such as '1 2 3'.");
					System.out.println(" 2. When finishing selecting, please press Enter key.)");
					
					System.out.println(counter + ". " + rootNode);
					counter++;
					
					for(j=0; j<numOfIntNodes; j++)
					{
						System.out.println(counter + ". " + intNodesList.get(j));
						counter++;
					}
					for(j=0; j<numOfLeafNodes; j++)
					{
						System.out.println(counter + ". " + leafNodesList.get(j));
						counter++;
					}

					selectedRootingPosition = aReader.readLine();
					tmpIndicator = true;
				}
			}
			
			System.out.println("**************************************************");
			System.out.println("* Please select the number in the following list *");
			System.out.println("* for your preferred outputs.                    *");
			System.out.println("**************************************************");

			System.out.println("");
			tmpIndicator = false;
			while(tmpIndicator == false)
			{
				System.out.println("1. Joint probability matrices along edges.");
				System.out.println("2. NSGTR model parameters along edges.");
				System.out.println("3. Both 1. and 2.");
				
				tmp = aReader.readLine();
				if(tmp.equals("1"))
				{
					outputOptions = 1;
					tmpIndicator = true;
				}
				else if(tmp.equals("2"))
				{
					outputOptions = 2;
					tmpIndicator = true;
				}
				else if(tmp.equals("3"))
				{
					outputOptions = 3;
					tmpIndicator = true;
				}
			}
			
			System.out.println("****************** End input parameters ***************"); 
		}
		catch(Exception e) 
	    {
		System.out.println("Error occured while reading input parameters " 
				   + e);
	    }	
	}
	
	/**
	 * write results
	 */
	public void writeResults()
	{
		String bestRoot = null;
		
		bestRoot = getTheBestRoot();
		
		if(allRootingPosition == false)
		{
			if(defaultPosition == true)
			{
				
				if(outputOptions == 1)// joint probability matrices only
				{
					writeBestShortestPath(bestRoot, true);//no necessary to append
				}
				else if(outputOptions == 2)//nsgtr parameters only
				{
					writeBestNSGTRModelParameters(bestRoot, true);//no necessary to append
				}
				else //both joints and nsgtr parameters
				{
					writeBestShortestPath(bestRoot, true);//no necessary to append
					writeBestNSGTRModelParameters(bestRoot, true);//no necessary to append
				}
			}
			else
			{
				ArrayList<String> selectedRoots = obtainSelectedNodes(selectedRootingPosition);
				if(outputOptions == 1)//joint probability matrices only
				{
					writeBestShortestPath(selectedRoots);
				}
				else if(outputOptions == 2)
				{
					writeBestNSGTRModelParameters(selectedRoots);
				}
				else
				{
					writeBestShortestPath(selectedRoots);
					writeBestNSGTRModelParameters(selectedRoots);
				}
			}
		}
		else
		{
			if(outputOptions == 1)
			{
				writeBestFitNonStationaryJointProbMatrices();
			}
			else if(outputOptions == 2)
			{
				writeBestNSGTRModelParametersAllNodes();
			}
			else
			{
				writeBestFitNonStationaryJointProbMatrices();
				writeBestNSGTRModelParametersAllNodes();
			}
		}
    	/**
    	 * create a new file
    	 */
 		//writeShortestPathResults(shortestPathReconstructedTrees, false);
    	/**
    	 * append results to previous results
    	 */
    	//writeShortestPathResults(shortestPathReconstructedTreesWithoutLevel, true);
    	/**
    	 * added on april 10, 2009
    	 */
    	//writeShortestPathResultsRootingAtANode(shortestPathReconstructedTreesRootingAtANode);
    	/**
    	 * added on April 17, 2009
    	 */
    	//writeBestFitNonStationaryFrequenciesVectors();
    	//writeBestFitNonStationaryJointProbMatrices();
	
	}
    public void estimatingPermutations()
    {
    	computePermutations();
		computeDifsOfPermutations();
		obtainShortestpath();
    }
    
    protected void obtainShortestpath()
    {
    	SearchShortestPath ssp; 
    	int j;
    	int numOfLeafNodes = leafNodesList.size();
    	int numOfIntNodes = intNodesList.size();
    	
    	String nodeName;
    	//String parentNodeName=null;
    	ContainerShortestPathSearchResults tmpResult;
    	//NodeStructure ns;
    	
    	if(shortestPathReconstructedTrees == null)
    		shortestPathReconstructedTrees = new TreeMap<String, ContainerShortestPathSearchResults>();
    	
    	/**
    	 * internal nodes first
    	 */
    	for(j=0; j<numOfIntNodes; j++)
    	{
    		nodeName = (String)intNodesList.get(j);
    		//System.out.println(nodeName+"\n");

    		//ns = (NodeStructure)phylogeneticTree.get(nodeName);
    		//parentNodeName = (String)ns.getParentNode();
    		ssp = new SearchShortestPath();
    		/** 
    		 * with levels
    		 */
    		tmpResult = ssp.findTheShortestPath(rootNode, nodeName, intNodesList, leafNodesList, phylogeneticTree, differences);
    		shortestPathReconstructedTrees.put(nodeName, tmpResult);
    		/**
    		 * without levels
    		 */
    		tmpResult = ssp.findTheShortestPath(phylogeneticTree);
    		shortestPathReconstructedTreesWithoutLevel.put(nodeName, tmpResult);
    		/**
    		 * added on april 10, 2009
    		 * rooting at a node. 
    		 * would like to have a fresh start
    		 */
    		ssp = new SearchShortestPath();
    		tmpResult = ssp.findTheShortestPathRootingAtANode(rootNode, nodeName, intNodesList, leafNodesList, phylogeneticTree, differences, reconstructedConditionalProb, permutationsChildToParent, permutationsParentToChild);
    		
    		shortestPathReconstructedTreesRootingAtANode.put(nodeName, tmpResult);
    	}
    	
    	/**
    	 * leaf nodes second
    	 */
    	for(j=0; j<numOfLeafNodes; j++)
    	{
    		nodeName = (String)leafNodesList.get(j);
    		//System.out.println(nodeName+"\n");
    		ssp = new SearchShortestPath();
    		/**
    		 * with levels
    		 */
    		tmpResult = ssp.findTheShortestPath(rootNode, nodeName, intNodesList, leafNodesList, phylogeneticTree, differences);	
    		shortestPathReconstructedTrees.put(nodeName, tmpResult);
    		/**
    		 * without levels
    		 */
    		tmpResult = ssp.findTheShortestPath(phylogeneticTree);	
    		shortestPathReconstructedTreesWithoutLevel.put(nodeName, tmpResult);
    		/**
    		 * added on april 10, 2009
    		 * rooting at a node. 
    		 * would like to have a fresh start
    		 */
    		ssp = new SearchShortestPath();
    		tmpResult = ssp.findTheShortestPathRootingAtANode(rootNode, nodeName, intNodesList, leafNodesList, phylogeneticTree, differences, reconstructedConditionalProb, permutationsChildToParent, permutationsParentToChild);
    		
    		shortestPathReconstructedTreesRootingAtANode.put(nodeName, tmpResult);
    		
    	}
    	
    	/**
    	 * added on april 10, 2009
    	 * root node needs to be handled specially because it is not in
    	 * either internal or leaf node.
    	 */
    	ssp = new SearchShortestPath();
		tmpResult = ssp.findTheShortestPathRootingAtANode(rootNode, rootNode, intNodesList, leafNodesList, phylogeneticTree, differences, reconstructedConditionalProb, permutationsChildToParent, permutationsParentToChild);
		
		shortestPathReconstructedTreesRootingAtANode.put(rootNode, tmpResult);
    	
    }

    protected void writeShortestPathResults(TreeMap<String, ContainerShortestPathSearchResults> inputCon,
            boolean appendOrNot)
    {
    	int j, index;
    	int numOfLeafNodes = leafNodesList.size();
    	int numOfIntNodes = intNodesList.size();
    	
    	String nodeName, parentName, tmpLine;
    	NodeStructure ns;

    	FileWriter myFw = null;

    	try
    	{

    		index = dataPath.lastIndexOf("/");
    		File myFile = new File(dataPath.substring(0,index) + "/shortestPath.txt");

    		if(myFile.exists())
    			myFw = new FileWriter(dataPath.substring(0,index) + "/shortestPath.txt", true);
    		else
    			myFw = new FileWriter(dataPath.substring(0,index) + "/shortestPath.txt");
    		/**
    		 * go through internal nodes
    		 */
    		for(j=0; j<numOfIntNodes; j++)
    		{
    			tmpLine = "";
    			nodeName = (String)intNodesList.get(j);
    			ns = (NodeStructure)phylogeneticTree.get(nodeName);
    			parentName = ns.getParentNode();

    			tmpLine += "Rooting at branch(" + parentName + ", " + nodeName + "): \n" + inputCon.get(nodeName).toString() + "\n";
    			myFw.write(tmpLine);
    		}
    		/**
    		 * go through leaf nodes
    		 */
    		for(j=0; j<numOfLeafNodes; j++)
    		{
    			tmpLine = "";
    			nodeName = (String)leafNodesList.get(j);
    			ns = (NodeStructure)phylogeneticTree.get(nodeName);
    			parentName = ns.getParentNode();

    			tmpLine += "Rooting at branch(" + parentName + ", " + nodeName + "): \n" + inputCon.get(nodeName).toString() + "\n";
    			myFw.write(tmpLine);
    		}
    		myFw.close();
    	}
    	catch(Exception e)
    	{
    		System.out.println("close file failed in writeShortestPathResults");
    		System.exit(-1);
    	}	
    }

    protected void writeShortestPathResultsRootingAtANode(TreeMap<String, ContainerShortestPathSearchResults> inputCon)
    {
    	int j, index;
    	int numOfLeafNodes = leafNodesList.size();
    	int numOfIntNodes = intNodesList.size();
    	
    	String nodeName, tmpLine;//, parentName;
    	//NodeStructure ns;

    	FileWriter myFw = null;

    	try
    	{

    		index = dataPath.lastIndexOf("/");
    		File myFile = new File(dataPath.substring(0,index) + "/shortestPathRootingAtANode.txt");
    		
    		if(myFile.exists())
    			myFw = new FileWriter(dataPath.substring(0,index) + "/shortestPathRootingAtANode.txt", true);
    		else 
    			myFw = new FileWriter(dataPath.substring(0,index) + "/shortestPathRootingAtANode.txt");
    		/**
    		 * old root first
    		 */
    		tmpLine = "";
    		tmpLine += "Rooting at node "  + rootNode + ": \n" + inputCon.get(rootNode).toString() + "\n";
			myFw.write(tmpLine);
    		/**
    		 * go through internal nodes
    		 */
    		for(j=0; j<numOfIntNodes; j++)
    		{
    			tmpLine = "";
    			nodeName = (String)intNodesList.get(j);
    			//ns = (NodeStructure)phylogeneticTree.get(nodeName);

    			tmpLine += "Rooting at node "  + nodeName + ": \n" + inputCon.get(nodeName).toString() + "\n";
    			myFw.write(tmpLine);
    		}
    		/**
    		 * go through leaf nodes
    		 */
    		for(j=0; j<numOfLeafNodes; j++)
    		{
    			tmpLine = "";
    			nodeName = (String)leafNodesList.get(j);
    			//ns = (NodeStructure)phylogeneticTree.get(nodeName);
    			//parentName = ns.getParentNode();

    			tmpLine += "Rooting at node " + nodeName + ": \n" + inputCon.get(nodeName).toString() + "\n";
    			myFw.write(tmpLine);
    		}
    		myFw.close();
    	}
    	catch(Exception e)
    	{
    		System.out.println("close file failed in writeShortestPathResultsRootingAtANode");
    		System.exit(-1);
    	}	
    }

    protected void writeBestFitNonStationaryFrequenciesVectors()
    {
        int j, k, m, index;
        int numOfLeafNodes = leafNodesList.size();
        int numOfIntNodes = intNodesList.size();

        String tmpLine, tmpRootNodeName, tmpNodeName;

        FileWriter myFw = null;

        TreeMap<String, double[]> tmpResults = null;

        double[] tmpMatrix = null;

        try
        {

                index = dataPath.lastIndexOf("/");
                File myFile = new File(dataPath.substring(0,index) + "/bestFitNonStationaryFrequencies.txt");

                if(myFile.exists())
                        myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitNonStationaryFrequencies.txt", true);
                else
                        myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitNonStationaryFrequencies.txt");
                /**
                 * go through root node first
                 * root node is not included in either internal nodes or leaf nodes.
                 * it has to be handled specially.
                 */
                        tmpResults = shortestPathReconstructedTreesRootingAtANode.get(rootNode).getBestFitNonStationaryMarginalFrequencies();
                        /**
                         * write root node frequencies
                         */
                        tmpMatrix = tmpResults.get(rootNode);
                        tmpLine = "";

                        for(m=0; m<numOfBases; m++)
                        {
                                tmpLine += tmpMatrix[m] + "\t ";
                        }
                        tmpLine += "\n";
                        myFw.write(tmpLine);

                        /**
                         * go through internal nodes first
                         */
                        for(k=0; k<numOfIntNodes; k++)
                        {
                                /**
                         * reset tmpLine
                         */
                        tmpNodeName = intNodesList.get(k);
                                tmpLine = "";


                                tmpMatrix = tmpResults.get(tmpNodeName);

                                for(m=0; m<numOfBases; m++)
                                {
                                        //System.out.println(tmpMatrix[m]);
                                        tmpLine += tmpMatrix[m] + "\t ";
                                        //System.out.println(tmpLine);
                                }
                                tmpLine += "\n";
                                myFw.write(tmpLine);
                        }
                        /**
                         * go through leaf nodes
                         */
                        for(k=0; k<numOfLeafNodes; k++)
                        {
                                /**
                         * reset tmpLine
                         */
                                tmpLine = "";
                        tmpNodeName = leafNodesList.get(k);
                                tmpMatrix = tmpResults.get(tmpNodeName);

                                for(m=0; m<numOfBases; m++)
                                {
                                        tmpLine += tmpMatrix[m] + "\t ";
                                }
                                tmpLine += "\n";
                                myFw.write(tmpLine);
                        }
 		/**
			 * go through leaf nodes
			 */
			for(k=0; k<numOfLeafNodes; k++)
			{
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
    			tmpNodeName = leafNodesList.get(k);
 				tmpMatrix = tmpResults.get(tmpNodeName);
				
				for(m=0; m<numOfBases; m++)
				{
					tmpLine += tmpMatrix[m] + "\t ";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
			}
    		
    		/**
    		 * go through internal nodes
    		 * for each of internal nodes, go through internal nodes presented branches first, then leaf nodes presented 
    		 * branches.
    		 */
    		for(j=0; j<numOfIntNodes; j++)
    		{    			
    			tmpRootNodeName = intNodesList.get(j);
    			tmpResults = shortestPathReconstructedTreesRootingAtANode.get(tmpRootNodeName).getBestFitNonStationaryMarginalFrequencies();
    			/**
    			 * go through root node first
    			 */
    			tmpLine = "";
    			tmpMatrix = tmpResults.get(rootNode);
				for(m=0; m<numOfBases; m++)
				{
					//System.out.println(tmpMatrix[m]);
					tmpLine += tmpMatrix[m] + "\t ";
					//System.out.println(tmpLine);
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
   			
     			/**
     			 * go through internal nodes
     			 */
    			for(k=0; k<numOfIntNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
        			tmpNodeName = intNodesList.get(k);
    				tmpLine = "";

  				
    				tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					//System.out.println(tmpMatrix[m]);
    					tmpLine += tmpMatrix[m] + "\t ";
    					//System.out.println(tmpLine);
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
    			/**
    			 * go through leaf nodes
    			 */
    			for(k=0; k<numOfLeafNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = leafNodesList.get(k);
     				tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					tmpLine += tmpMatrix[m] + "\t ";
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
   			
    		}
    		/**
    		 * go through internal nodes first
    		 * for each of internal nodes, go through internal nodes presented branches first, then leaf nodes presented 
    		 * branches.
    		 */

    		for(j=0; j<numOfLeafNodes; j++)
    		{    			
    			tmpRootNodeName = leafNodesList.get(j);
    			tmpResults = shortestPathReconstructedTreesRootingAtANode.get(tmpRootNodeName).getBestFitNonStationaryMarginalFrequencies();
    			/**
    			 * go through root node first
    			 */
    			tmpLine = "";
    			tmpMatrix = tmpResults.get(rootNode);
				for(m=0; m<numOfBases; m++)
				{
					tmpLine += tmpMatrix[m] + "\t ";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
    			/**
     			 * go through internal nodes
     			 */
    			for(k=0; k<numOfIntNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = intNodesList.get(k);

        			tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					tmpLine += tmpMatrix[m] + "\t ";
       				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
    			/**
    			 * go through leaf nodes
    			 */
    			for(k=0; k<numOfLeafNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = leafNodesList.get(k);

    				tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					tmpLine += tmpMatrix[m]+ "\t ";
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
   			
    		}
    		//myFw.write(tmpLine);
    		myFw.close();
    	}
     	catch(Exception e)
    	{
    		System.out.println("close file failed in writeDifferenceContainer");
    		System.exit(-1);
    	}	
   	
    }

    protected void writeBestFitNonStationaryJointProbMatrices()
    {
    	int j, k, m, n, index;
    	int numOfLeafNodes = leafNodesList.size();
    	int numOfIntNodes = intNodesList.size();
    	
    	String tmpLine, tmpRootNodeName, tmpNodeName;
    	
    	FileWriter myFw = null;
    	
    	TreeMap<String, double[][]> tmpResults = null;
    	
    	double[][] tmpMatrix = null;
    	
    	try
    	{
    		
    		index = dataPath.lastIndexOf("/");
    		File myFile = new File(dataPath.substring(0,index) + "/bestFitNonStationaryJointProbMatrices.txt");
    		
    		if(myFile.exists())
    			myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitNonStationaryJointProbMatrices.txt", true);
    		else
    			myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitNonStationaryJointProbMatrices.txt");
    		
    		/**
    		 * go through root node first
    		 * root node is not included in either internal node or leaf node.
    		 * it has to be handled specially
    		 */
			tmpResults = shortestPathReconstructedTreesRootingAtANode.get(rootNode).getBestFitNonStationaryJointProbMatrices();			
			
			/**
 			 * go through internal nodes
 			 */
			for(k=0; k<numOfIntNodes; k++)
			{
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
    			tmpNodeName = intNodesList.get(k);
				tmpMatrix = tmpResults.get(tmpNodeName);
				
				for(m=0; m<numOfBases; m++)
				{
					for(n=0; n<numOfBases; n++)
						tmpLine += tmpMatrix[m][n] + "\t ";
					tmpLine += "\n";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
			}
			/**
			 * go through leaf nodes
			 */
			for(k=0; k<numOfLeafNodes; k++)
			{
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
    			tmpNodeName = leafNodesList.get(k);
 				tmpMatrix = tmpResults.get(tmpNodeName);
				
				for(m=0; m<numOfBases; m++)
				{
					for(n=0; n<numOfBases; n++)
						tmpLine += tmpMatrix[m][n] + "\t ";
					tmpLine += "\n";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
			}
	   		
    		/**
    		 * go through internal nodes 
    		 * for each of internal nodes, go through internal nodes presented branches first, then leaf nodes presented 
    		 * branches.
    		 */

    		for(j=0; j<numOfIntNodes; j++)
    		{    			
    			tmpRootNodeName = intNodesList.get(j);
    			tmpResults = shortestPathReconstructedTreesRootingAtANode.get(tmpRootNodeName).getBestFitNonStationaryJointProbMatrices();
    			
    			/**
    			 * handle root node first
    			 */
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
				tmpMatrix = tmpResults.get(rootNode);
				
				for(m=0; m<numOfBases; m++)
				{
					for(n=0; n<numOfBases; n++)
						tmpLine += tmpMatrix[m][n] + "\t ";
					tmpLine += "\n";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
  			
     			/**
     			 * go through internal nodes
     			 */
    			for(k=0; k<numOfIntNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = intNodesList.get(k);
    				if(tmpNodeName.equals(tmpRootNodeName))
    					continue;
    				tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					for(n=0; n<numOfBases; n++)
    						tmpLine += tmpMatrix[m][n] + "\t ";
    					tmpLine += "\n";
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
    			/**
    			 * go through leaf nodes
    			 */
    			for(k=0; k<numOfLeafNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = leafNodesList.get(k);
     				tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					for(n=0; n<numOfBases; n++)
    						tmpLine += tmpMatrix[m][n] + "\t ";
    					tmpLine += "\n";
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
   			
    		}

    		/**
    		 * go through rooting at leaf nodes
    		 */
    		/**
    		 * go through internal nodes first
    		 * for each of internal nodes, go through internal nodes presented branches first, then leaf nodes presented 
    		 * branches.
    		 */
    		
    		for(j=0; j<numOfLeafNodes; j++)
    		{    			
    			tmpRootNodeName = leafNodesList.get(j);
    			tmpResults = shortestPathReconstructedTreesRootingAtANode.get(tmpRootNodeName).getBestFitNonStationaryJointProbMatrices();
       			/**
    			 * handle root node first
    			 */
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
				tmpMatrix = tmpResults.get(rootNode);
				
				for(m=0; m<numOfBases; m++)
				{
					for(n=0; n<numOfBases; n++)
						tmpLine += tmpMatrix[m][n] + "\t ";
					tmpLine += "\n";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
     			/**
     			 * go through internal nodes
     			 */
    			for(k=0; k<numOfIntNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = intNodesList.get(k);

        			tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					for(n=0; n<numOfBases; n++)
    						tmpLine += tmpMatrix[m][n] + "\t ";
    					tmpLine += "\n";
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
    			/**
    			 * go through leaf nodes
    			 */
    			for(k=0; k<numOfLeafNodes; k++)
    			{
     	   			/**
        			 * reset tmpLine
        			 */
     				tmpLine = "";
        			tmpNodeName = leafNodesList.get(k);
    				if(tmpNodeName.equals(tmpRootNodeName))
    					continue;

    				tmpMatrix = tmpResults.get(tmpNodeName);
    				
    				for(m=0; m<numOfBases; m++)
    				{
    					for(n=0; n<numOfBases; n++)
    						tmpLine += tmpMatrix[m][n] + "\t ";
    					tmpLine += "\n";
    				}
    				tmpLine += "\n";
    				myFw.write(tmpLine);
    			}
   			
    		}

    		myFw.close();
    	}
     	catch(Exception e)
    	{
    		System.out.println("close file failed in writeDifferenceContainer");
    		System.exit(-1);
    	}	
   	
    }

    protected void computeDifsOfPermutations()
    {
    	int nodeIndex;
    	String nodeName;
    	//ArrayList<double[][]> permutationsOfABranch;
    	/**
    	 * go through internal nodes
    	 */
    	int numOfIntNodes = intNodesList.size();
    	int numOfLeafNodes = leafNodesList.size();
    	
    	for(nodeIndex=0; nodeIndex<numOfIntNodes; nodeIndex++)
    	{
    		/**
    		 * obtain permutations of a branch
    		 */
    		nodeName = (String)intNodesList.get(nodeIndex);
    		//permutationsOfABranch = (ArrayList<double[][]>)permutations.get(nodeName);
    		/**
    		 * Compute differences of those permutations
    		 */
    		//computeDifsOfPermutationsOfABranch(nodeName, permutationsOfABranch);
    		computeDifsOfPermutationsOfABranch(nodeName);
    	}
    	
    	/**
    	 * go through leaf nodes
    	 */
    	for(nodeIndex=0; nodeIndex<numOfLeafNodes; nodeIndex++)
    	{
    		/**
    		 * obtain permutations of a branch
    		 */
    		nodeName = (String)leafNodesList.get(nodeIndex);
    		//permutationsOfABranch = (ArrayList<double[][]>)permutations.get(nodeName);
    		/**
    		 * Compute differences of those permutations
    		 */
    		//computeDifsOfPermutationsOfABranch(nodeName, permutationsOfABranch);
    		computeDifsOfPermutationsOfABranch(nodeName);
    	}
    	
    	/**
    	 * write out results
    	 */
    	//writeDifferences();
   }

    protected void computeDifsOfPermutationsOfABranch(String nodeName)
    {
    	ArrayList<double[][]> ptc = permutationsParentToChild.get(nodeName);
    	ArrayList<double[][]> ctp = permutationsChildToParent.get(nodeName);
    	
    	int conSize = ptc.size();
    	int j;
    	double[][] tmpMatrix, tmpResults;
    	double[] resultsParentToChild = new double[conSize];
    	double[] resultsChildToParent = new double[conSize];
    	
    	ContainerForDifTwoDirections newCon = new ContainerForDifTwoDirections();
    	ContainerReconstructedMatrices newReCon; 
    	ContainerTwoArrayListDouble newConArrayList = new ContainerTwoArrayListDouble();
    	ContainerTwoArrayListDouble newJointArrayList = new ContainerTwoArrayListDouble();
    	
    	ArrayList<double[][]> parentToChildCon = new ArrayList<double[][]>();
    	ArrayList<double[][]> childToParentCon = new ArrayList<double[][]>();
    	ArrayList<double[][]> parentToChildJoint = new ArrayList<double[][]>();
    	ArrayList<double[][]> childToParentJoint = new ArrayList<double[][]>();
    	
    	for(j=0; j<conSize; j++)
    	{
    		/**
    		 * From parent to child
    		 */
    		tmpMatrix = (double[][])ptc.get(j);

    		newReCon = reconstructMatrix(tmpMatrix);
    		tmpResults = (double[][])newReCon.getReconstructedJoint();
    		parentToChildJoint.add(tmpResults);
    		tmpResults =(double[][])newReCon.getReconstructedCond();
    		parentToChildCon.add(tmpResults);
    		resultsParentToChild[j] = newReCon.getDifference();
    		/**
    		 * from child to parent
    		 */
    		tmpMatrix = (double[][])ctp.get(j);

    		newReCon = reconstructMatrix(tmpMatrix);
       		tmpResults = (double[][])newReCon.getReconstructedJoint();
    		childToParentJoint.add(tmpResults);
    		tmpResults =(double[][])newReCon.getReconstructedCond();
    		childToParentCon.add(tmpResults);
    		resultsChildToParent[j] = newReCon.getDifference();
    	}
    	
    	/**
    	 * handle differences
    	 */
    	newCon.setParentToChild(resultsParentToChild, conSize);
    	newCon.setChildToParent(resultsChildToParent, conSize);
    	newCon.setArraySize(conSize);
    	
    	differences.put(nodeName, newCon);
    	/**
    	 * handle best fit conditional probs
    	 */
    	newConArrayList.setParentToChild(parentToChildCon);
    	newConArrayList.setChildToParent(childToParentCon);
    	reconstructedConditionalProb.put(nodeName, newConArrayList);
    	/**
    	 * handle best fit joint probs
    	 */
    	newJointArrayList.setParentToChild(parentToChildJoint);
    	newJointArrayList.setChildToParent(childToParentJoint);
    	reconstructedJointProb.put(nodeName, newJointArrayList);
    	
    }
   
    protected ContainerReconstructedMatrices reconstructMatrix(double[][] inputMatrix)
    {
   	
      	double[] pi = null;
      	/**
      	 * obtain frequencies
      	 */
      	pi = computeNULLMarginalFrequencies(inputMatrix, 1);//always from ptc since ctp has been transporsed.
    	double[][] piMatrix = null; 
    	/**
    	 * obtain frequencies matrix \Pi
    	 */
    	piMatrix = convertVectorToDiag(pi, numOfBases);
    	/**
    	 * declair temporary storage
    	 */
    	double[][] tmpMatrix1 = null;
    	double[][] tmpMatrix; 
    	ArrayList<double[]> tmpResults = null;
    	/**
    	 * \Pi^{-1}
    	 */
    	tmpMatrix1 = computeInverseOfDiagMatrix(piMatrix);
    	
    	/**
    	 * tmpMatrix = Q
    	 */
    	tmpMatrix = inputMatrix;
    	/**
    	 * compute conditional probability matrix P
    	 */
    	double[][] originalCond = null;
    	originalCond = computeMultiplicationOfTwoMatrices(tmpMatrix1, tmpMatrix);
    	/**
    	 * compute eigen decomposition of P
    	 * Note: it is possible that P gets a complex valued result
    	 */
    	ArrayList<double[]> eigenResults = null;
    	eigenResults = computeEigenDecomposition(originalCond);
    	
    	ArrayList<double[]> reorganizedEigenResults = reorganizeEigenResults(eigenResults);
    	
    	/**
    	 * obtain real part of eigen value. 
    	 * The "one" corresponds to frequencies.
    	 * The imaginary part of eigen value is not useful
    	 */
    	double[] realPartEigenValues = null; 
    	realPartEigenValues = eigenResults.get(0);
    	/**
    	 * obtain the largest eigenvalue index
    	 */
    	int indexPi;
    	indexPi = findOutIndexOfPi(realPartEigenValues);
    	/**
    	 * obtain eigenvectors U^{-1} of P
    	 */
    	double[] realPartEigenVectors = null;

    	realPartEigenVectors = reorganizedEigenResults.get(4);

    	tmpMatrix = convertFortranToJavaMatrix(realPartEigenVectors);
    	/**
    	 * obtain best fit \Pi matrix
    	 */
    	double[][] piNew = obtainNewPiVector(tmpMatrix, indexPi);
    	
    	//System.out.println("pi new: ");
    	//sumOverColumnsOfAMatrix(piNew);
    	/**
    	 * obtain symmetrized joint probablity matrix
    	 */
    	double[][] newJointProb = null;
    	/**
    	 * best fit joint = best fit \Pi * P
    	 */
    	newJointProb = computeMultiplicationOfTwoMatrices(piNew, originalCond);
    	/**
    	 * force symmetry
    	 */
    	newJointProb = symmetrizeAMatrix(newJointProb);
    	/**
    	 * compute sqrt(best fit \Pi) and inverse sqrt(best fit \Pi)
    	 */
    	double[][] piNewSqrt = null; 
    	double[][] piNewSqrtInverse = null;
    	piNewSqrt = computeSqrtOfDiagMatrix(piNew);
    	piNewSqrtInverse = computeInverseOfDiagMatrix(piNewSqrt);
     	/**
     	 * P_{sym} = \pi^{-\frac{1}{2}} %*% F %*% \pi&{-\frac{1}{2}}
     	 * 
     	 */
     	double[][] newProbSymmetric = null;
     	/*
     	 *  \pi^{-\frac{1}{2}} %*% F
     	 */
    	tmpMatrix1 = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newJointProb);
    	/*
    	 * \pi^{-\frac{1}{2}} %*% F %*% \pi&{-\frac{1}{2}}
    	 */
     	newProbSymmetric = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrtInverse);
     	
     	//System.out.println("P_sym:");
     	//(newProbSymmetric);
    	/**
    	 * computeEigenDecompositionRealSymMatrix return values
    	 * [0]: eigenvectors;
    	 * [1]: eigenvalues;
    	 */
    	ArrayList<double[]> eigenDecompositionOfReContructedCondMatrix = null;
    	eigenDecompositionOfReContructedCondMatrix = computeEigenDecompositionRealSymMatrix(newProbSymmetric);
    	/**
    	 * eigen values
    	 */
    	double[] newEigenValuesOfReconstructedCondMatrix = null;
    	newEigenValuesOfReconstructedCondMatrix = eigenDecompositionOfReContructedCondMatrix.get(0);
    	/**
    	 * if any negative eigenvalues, reset as zero.
    	 */
    	newEigenValuesOfReconstructedCondMatrix = resetEigenvalues(newEigenValuesOfReconstructedCondMatrix);
    	/**
    	 * take logarithm
    	 */
    	newEigenValuesOfReconstructedCondMatrix = takeLogarithmResetEigenvalues(newEigenValuesOfReconstructedCondMatrix);
    	//System.out.println(newEigenValuesOfReconstructedCondMatrix[0]+ " " + newEigenValuesOfReconstructedCondMatrix[1]+ " " + newEigenValuesOfReconstructedCondMatrix[2]+" " +  newEigenValuesOfReconstructedCondMatrix[3]);
    	/**
    	 * construct a new diagonal eigenvalues matrix
    	 */
    	double[][] diagonalNewEigenValuesOfReconstructedCondMatrix = null;
    	diagonalNewEigenValuesOfReconstructedCondMatrix = convertVectorToDiag(newEigenValuesOfReconstructedCondMatrix, numOfBases);
    	
    	/**
    	 * obtain right eigenvectors of P_{sym}
    	 */
    	double[] tmpVector = null;
    	tmpVector = eigenDecompositionOfReContructedCondMatrix.get(1);
    	/**
    	 * convert right eigenvectors of P_{sym} into a java style matrix
    	 */
    	double[][] newEigenVectorsOfReconstructedCondMatrixSym = null;
    	newEigenVectorsOfReconstructedCondMatrixSym = convertFortranToJavaMatrix(tmpVector);
    	/**
    	 * desymmetrizing right eigenvectors of P_{sym}
    	 */
    	double[][] newEigenVectorsOfReconstructedCondMatrix = null;
    	newEigenVectorsOfReconstructedCondMatrix = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newEigenVectorsOfReconstructedCondMatrixSym);
    	/**
    	 * preparing inversing right eigenvectors of P_{sym}
    	 * lapack asks for LU decomposited matrix for computing inverse matrix
    	 */
    	ContainerForNagArrayListForDoubleAndIntArray luOfEigenVectors = computeLURealNumber(tmpVector);
     	int[] ipiv = luOfEigenVectors.getintArray();
     	tmpResults = luOfEigenVectors.getAList();
     	realPartEigenVectors = tmpResults.get(0);//use old vector for LU results of a real matrix
     	ArrayList<double[]>inverseEigenVectors = computeInverseRealMatrix(realPartEigenVectors, ipiv);
     	/**
     	 * now tmpMatrix1 has U_{sym}^{-1}
     	 */
     	tmpMatrix1 = convertFortranToJavaMatrix(inverseEigenVectors.get(0));
     	/**
     	 * desymmetrizing left eigenvectors of P_{sym} (U^{-1}_{sym} * sqrt{\Pi})
     	 */
     	double[][] newEigenVectorsOfReconstructedCondMatrixInversed = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrt);
     	/**
     	 * compute Q*.
     	 */
     	tmpMatrix = computeMultiplicationOfTwoMatrices(newEigenVectorsOfReconstructedCondMatrix, diagonalNewEigenValuesOfReconstructedCondMatrix);
     	tmpMatrix = computeMultiplicationOfTwoMatrices(tmpMatrix, newEigenVectorsOfReconstructedCondMatrixInversed);
     	//System.out.println("Q*: ");
     	//(tmpMatrix);
     	
     	/**
     	 * set negative off-diagonal as zero
     	 */
     	tmpMatrix = handleNegativeOffDiagonalEntriesQ(tmpMatrix);
     	//System.out.println("after handle negative off-diagonal: ");
     	//(tmpMatrix);
     	
     	/**
     	 * the rate matrix should keep Q_{jj} = -\sum{j: j\neq i} Q_{i, j}
     	 * after setting negative off-diagonals as zero, the diagonal elements will
     	 * change. They should be recomputed based on Q_{jj} = -\sum{j: j\neq i} Q_{i, j}
     	 */
     	tmpMatrix = resetRateMatrix(tmpMatrix);
     	//System.out.println("recompute diagonal q: ");
     	//(tmpMatrix);
     	
     	/**
     	 * obtain normalization results: double[][]: normalized rate matrix
     	 *                               Dobule: normalization factor
     	 */
     	TreeMap<Double, double[][]> normalizedResults = null;
     	double normalizationFactor = 0.0;
     	normalizedResults = normalizeQ(tmpMatrix, piNew);
     	normalizationFactor = normalizedResults.firstKey().doubleValue();
     	//System.out.println(normalizationFactor);
    	/**
     	 * normalized Q
     	 */
     	tmpMatrix = normalizedResults.get(normalizedResults.firstKey());
     	//System.out.println("Normalized q: ");
     	//(tmpMatrix);
     	
     	/**
     	 * for debug only
     	 */
     	//(tmpMatrix);
     	/**
     	 * Symmetrizing Q
     	 */
     	tmpMatrix = computeMultiplicationOfTwoMatrices(piNewSqrt, tmpMatrix);
     	tmpMatrix = computeMultiplicationOfTwoMatrices(tmpMatrix, piNewSqrtInverse);
     	/**
     	 * take eigen decomposition of Q
     	 * WILL REUSE THE VARIABLE DEFINED FOR RECONSTRUCT P
     	 */
    	eigenDecompositionOfReContructedCondMatrix = computeEigenDecompositionRealSymMatrix(tmpMatrix);
    	newEigenValuesOfReconstructedCondMatrix = eigenDecompositionOfReContructedCondMatrix.get(0);
    	/**
    	 * because of computation error, the largest eigenvalue is not exactly equal to 0.0
    	 * this will cause the largest eigenvalue times evolutionary time not being equal to zero
    	 * It will be set the largest eigenvalue as zero by \lambda_j - \lambda_1
    	 */
    	newEigenValuesOfReconstructedCondMatrix = resetEigenValuesForceContinuousMarkovChain(newEigenValuesOfReconstructedCondMatrix);
    	/**
    	 * multiply by time
    	 */
       	newEigenValuesOfReconstructedCondMatrix = vectorMultiplyByAConstant(newEigenValuesOfReconstructedCondMatrix, normalizationFactor);
    	//System.out.println("before exponential: \n" + newEigenValuesOfReconstructedCondMatrix[0] + " " + newEigenValuesOfReconstructedCondMatrix[1] + " " + newEigenValuesOfReconstructedCondMatrix[2] + " " + newEigenValuesOfReconstructedCondMatrix[3]);
    	/**
    	 * take exponential
    	 */
    	newEigenValuesOfReconstructedCondMatrix = takeExponentialEigenvalues(newEigenValuesOfReconstructedCondMatrix);
    	//System.out.println("after exponential: \n" + newEigenValuesOfReconstructedCondMatrix[0] + " " + newEigenValuesOfReconstructedCondMatrix[1] + " " + newEigenValuesOfReconstructedCondMatrix[2] + " " + newEigenValuesOfReconstructedCondMatrix[3]);
    	
    	/**
    	 * construct a new diagonal eigenvalues matrix
    	 */
    	diagonalNewEigenValuesOfReconstructedCondMatrix = convertVectorToDiag(newEigenValuesOfReconstructedCondMatrix, numOfBases);
    	
    	/**
    	 * obtain right eigenvectors of P_{sym}
    	 */
    	tmpVector = eigenDecompositionOfReContructedCondMatrix.get(1);
    	/**
    	 * convert right eigenvectors of P_{sym} into a java style matrix
    	 */
    	newEigenVectorsOfReconstructedCondMatrixSym = convertFortranToJavaMatrix(tmpVector);
    	/**
    	 * desymmetrizing right eigenvectors of P_{sym}
    	 */
    	newEigenVectorsOfReconstructedCondMatrix = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newEigenVectorsOfReconstructedCondMatrixSym);
    	/**
    	 * preparing inversing right eigenvectors of P_{sym}
    	 * lapack asks for LU decomposited matrix for computing inverse matrix
    	 */
    	luOfEigenVectors = computeLURealNumber(tmpVector);
     	ipiv = luOfEigenVectors.getintArray();
     	tmpResults = luOfEigenVectors.getAList();
     	realPartEigenVectors = tmpResults.get(0);//use old vector for LU results of a real matrix
     	inverseEigenVectors = computeInverseRealMatrix(realPartEigenVectors, ipiv);
     	tmpMatrix1 = convertFortranToJavaMatrix(inverseEigenVectors.get(0));
     	/**
     	 * desymmetrizing left eigenvectors of P_{sym}
     	 */
     	newEigenVectorsOfReconstructedCondMatrixInversed = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrt);

     	tmpMatrix = computeMultiplicationOfTwoMatrices(newEigenVectorsOfReconstructedCondMatrix, diagonalNewEigenValuesOfReconstructedCondMatrix);
     	tmpMatrix = computeMultiplicationOfTwoMatrices(tmpMatrix, newEigenVectorsOfReconstructedCondMatrixInversed);
     	/**
     	 * for debug only
     	 */
     	//(tmpMatrix);
    	/**
    	 * compute inverse of eigenvectors.
    	 */
    	double dif = computeDifOfTwoMatrices(originalCond, tmpMatrix);
    		
    	ContainerReconstructedMatrices conMatrices = new ContainerReconstructedMatrices();
    	conMatrices.setReconstructedCond(tmpMatrix, numOfBases);
    	conMatrices.setReconstructedJointProb(newJointProb, numOfBases);
    	conMatrices.setConSize(numOfBases);
    	conMatrices.setDifference(dif);
    	
    	return conMatrices;
   }
    
    protected containerNSGTRParameters getNSGTRParemeters(double[][] inputMatrix)
    {
   	
    	double[][] tmpMatrix1 = null;
    	double[][] tmpMatrix; 
    	ArrayList<double[]> tmpResults = null;
    	double[][] originalCond = inputMatrix;

    	/**
    	 * compute eigen decomposition of P
    	 * Note: it is possible that P gets a complex valued result
    	 */
    	ArrayList<double[]> eigenResults = null;
    	eigenResults = computeEigenDecomposition(originalCond);
    	
    	ArrayList<double[]> reorganizedResults = reorganizeEigenResults(eigenResults);
    	
    	/**
    	 * obtain real part of eigen value. 
    	 * The "one" corresponds to frequencies.
    	 * The imaginary part of eigen value is not useful
    	 */
    	double[] realPartEigenValues = null; 
    	realPartEigenValues = eigenResults.get(0);
    	/**
    	 * obtain the largest eigenvalue index
    	 */
    	int indexPi;
    	indexPi = findOutIndexOfPi(realPartEigenValues);
    	/**
    	 * obtain eigenvectors U of P
    	 */
    	double[] realPartEigenVectors = null;
    	realPartEigenVectors = reorganizedResults.get(4);
    	/**
    	 * construct a numOfBases*numOfBases matrix
    	 */
    	tmpMatrix = convertFortranToJavaMatrix(realPartEigenVectors);
    	/**
    	 * obtain best fit \Pi matrix
    	 */
    	double[][] piNew = obtainNewPiVector(tmpMatrix, indexPi);
    	double[] piNewVector = new double[4];
    	for(int j=0; j<4; j++)
    	{
    		piNewVector[j] = piNew[j][j];
    	}
    	/**
    	 * set stationary frequencies
    	 */
    	containerNSGTRParameters aCon = new containerNSGTRParameters();
    	aCon.setStationaryFrequencies(piNewVector);
    	
    	//System.out.println("pi new: ");
    	//(piNew);
    	/**
    	 * obtain symmetrized joint probablity matrix
    	 */
    	double[][] newJointProb = null;
    	/**
    	 * best fit joint = best fit \Pi * P
    	 */
    	newJointProb = computeMultiplicationOfTwoMatrices(piNew, originalCond);
    	/**
    	 * force symmetry
    	 */
    	newJointProb = symmetrizeAMatrix(newJointProb);
    	/**
    	 * compute sqrt(best fit \Pi) and inverse sqrt(best fit \Pi)
    	 */
    	double[][] piNewSqrt = null; 
    	double[][] piNewSqrtInverse = null;
    	piNewSqrt = computeSqrtOfDiagMatrix(piNew);
    	piNewSqrtInverse = computeInverseOfDiagMatrix(piNewSqrt);
     	/**
     	 * P_{sym} = \pi^{-\frac{1}{2}} %*% F %*% \pi&{-\frac{1}{2}}
     	 * 
     	 */
     	double[][] newProbSymmetric = null;
     	/*
     	 *  \pi^{-\frac{1}{2}} %*% F
     	 */
    	tmpMatrix1 = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newJointProb);
    	/*
    	 * \pi^{-\frac{1}{2}} %*% F %*% \pi&{-\frac{1}{2}}
    	 */
     	newProbSymmetric = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrtInverse);
     	
     	//System.out.println("P_sym:");
     	//sumOverColumnsOfAMatrix(newProbSymmetric);
    	/**
    	 * computeEigenDecompositionRealSymMatrix return values
    	 * [0]: eigenvectors;
    	 * [1]: eigenvalues;
    	 */
    	ArrayList<double[]> eigenDecompositionOfReContructedCondMatrix = null;
    	eigenDecompositionOfReContructedCondMatrix = computeEigenDecompositionRealSymMatrix(newProbSymmetric);
    	/**
    	 * eigen values
    	 */
    	double[] newEigenValuesOfReconstructedCondMatrix = null;
    	newEigenValuesOfReconstructedCondMatrix = eigenDecompositionOfReContructedCondMatrix.get(0);
    	/**
    	 * if any negative eigenvalues, reset as zero.
    	 */
    	newEigenValuesOfReconstructedCondMatrix = resetEigenvalues(newEigenValuesOfReconstructedCondMatrix);
    	/**
    	 * take logarithm
    	 */
    	newEigenValuesOfReconstructedCondMatrix = takeLogarithmResetEigenvalues(newEigenValuesOfReconstructedCondMatrix);
    	//System.out.println(newEigenValuesOfReconstructedCondMatrix[0]+ " " + newEigenValuesOfReconstructedCondMatrix[1]+ " " + newEigenValuesOfReconstructedCondMatrix[2]+" " +  newEigenValuesOfReconstructedCondMatrix[3]);
    	/**
    	 * construct a new diagonal eigenvalues matrix
    	 */
    	double[][] diagonalNewEigenValuesOfReconstructedCondMatrix = null;
    	diagonalNewEigenValuesOfReconstructedCondMatrix = convertVectorToDiag(newEigenValuesOfReconstructedCondMatrix, numOfBases);
    	
    	/**
    	 * obtain right eigenvectors of P_{sym}
    	 */
    	double[] tmpVector = null;
    	tmpVector = eigenDecompositionOfReContructedCondMatrix.get(1);
    	/**
    	 * convert right eigenvectors of P_{sym} into a java style matrix
    	 */
    	double[][] newEigenVectorsOfReconstructedCondMatrixSym = null;
    	newEigenVectorsOfReconstructedCondMatrixSym = convertFortranToJavaMatrix(tmpVector);
    	/**
    	 * desymmetrizing right eigenvectors of P_{sym}
    	 */
    	double[][] newEigenVectorsOfReconstructedCondMatrix = null;
    	newEigenVectorsOfReconstructedCondMatrix = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newEigenVectorsOfReconstructedCondMatrixSym);
    	/**
    	 * preparing inversing right eigenvectors of P_{sym}
    	 * lapack asks for LU decomposited matrix for computing inverse matrix
    	 */
    	ContainerForNagArrayListForDoubleAndIntArray luOfEigenVectors = computeLURealNumber(tmpVector);
     	int[] ipiv = luOfEigenVectors.getintArray();
     	tmpResults = luOfEigenVectors.getAList();
     	realPartEigenVectors = tmpResults.get(0);//use old vector for LU results of a real matrix
     	ArrayList<double[]> inverseEigenVectors = computeInverseRealMatrix(realPartEigenVectors, ipiv);
     	/**
     	 * now tmpMatrix1 has U_{sym}^{-1}
     	 */
     	tmpMatrix1 = convertFortranToJavaMatrix(inverseEigenVectors.get(0));
     	/**
     	 * desymmetrizing left eigenvectors of P_{sym} (U^{-1}_{sym} * sqrt{\Pi})
     	 */
     	double[][] newEigenVectorsOfReconstructedCondMatrixInversed = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrt);
     	/**
     	 * compute Q*.
     	 */
     	tmpMatrix = computeMultiplicationOfTwoMatrices(newEigenVectorsOfReconstructedCondMatrix, diagonalNewEigenValuesOfReconstructedCondMatrix);
     	tmpMatrix = computeMultiplicationOfTwoMatrices(tmpMatrix, newEigenVectorsOfReconstructedCondMatrixInversed);
     	//System.out.println("Q*: ");
     	//sumOverColumnsOfAMatrix(tmpMatrix);
     	
     	/**
     	 * set negative off-diagonal as zero
     	 */
     	tmpMatrix = handleNegativeOffDiagonalEntriesQ(tmpMatrix);
     	//System.out.println("after handle negative off-diagonal: ");
     	//sumOverColumnsOfAMatrix(tmpMatrix);
     	
     	/**
     	 * the rate matrix should keep Q_{jj} = -\sum{j: j\neq i} Q_{i, j}
     	 * after setting negative off-diagonals as zero, the diagonal elements will
     	 * change. They should be recomputed based on Q_{jj} = -\sum{j: j\neq i} Q_{i, j}
     	 */
     	tmpMatrix = resetRateMatrix(tmpMatrix);
     	//System.out.println("recompute diagonal q: ");
     	//sumOverColumnsOfAMatrix(tmpMatrix);
     	
     	/**
     	 * obtain normalization results: double[][]: normalized rate matrix
     	 *                               Dobule: normalization factor
     	 */
     	TreeMap<Double, double[][]> normalizedResults = null;
     	normalizedResults = normalizeQ(tmpMatrix, piNew);
     	//System.out.println(normalizationFactor);
    	/**
     	 * normalized Q
     	 */
     	tmpMatrix = normalizedResults.get(normalizedResults.firstKey());
     	/**
     	 * obtain ratrix
     	 */
     	aCon.setRateMatrix(tmpMatrix);
    	
    	return aCon;
   }

    protected  double[] computeNULLMarginalFrequencies(double[][] inputMatrix, int direction)
    {
    	double[] results = initializeDoubleArray(numOfBases);
    	
    	int j, k;
     	if(direction == 1)
    	{
    		/**
    		 * parent to child
	   		 * Forward direction of traversal from parent node to child node.
	   		   So, sum over all columns i.e. child values are added together in a column.
    		 */
    		for(j=0; j<numOfBases; j++)
    		{
    			for(k=0; k<numOfBases; k++)
    				results[j] += inputMatrix[j][k];
    		}

    	}
     	else if(direction == 2)
    	{
    		/**
	   		 * Forward direction of traversal from parent node to child node.
	   		   So, sum over all columns i.e. child values are added together in a column.
    		 */
    		for(j=0; j<numOfBases; j++)
    		{
    			for(k=0; k<numOfBases; k++)
    				results[j] += inputMatrix[k][j];
    		}
    	}
     	return results;
    }

    protected ArrayList<double[]> computeEigenDecomposition(double[][] inputMatrix)
    {
    	/**
    	 * jobvl: 'v' left eigenvectors
    	 */
    	char jobvl = 'V';
    	/**
    	 * jobvr: 'v' right eigenvectors
    	 */
    	char jobvr = 'V';
    	/**
    	 * order of matrix A: numOfBases 
    	 */
    	int n = java.lang.Math.max(1, numOfBases);
    	/**
    	 *  A: matrix will be decomposited.
    	 *    dimension n*n
    	 */
    	double[] a = convertJavaToFortranMatrix(inputMatrix);
    	/**
    	 * LDA: first dimension of the array A
    	 *      (numOfBases)
    	 */
    	int lda = n;
    	/**
    	 * WR: store the real part of eigenvalues
    	 *     dimension is numOfBases
    	 */
    	double[] wr = new double[n];
    	/**
    	 * WI: store the imaginary part of eigenvalues
    	 *     dimension is numOfBases
    	 */
    	double[] wi = new double[n];
    	/**
    	 * LVDR: (numOfBases)the first dimention of the array VR.
    	 */
    	int ldvl = n;
    	/**
    	 * Vl(LDVl, *): double arrays.
    	 *                 store real part of eigenvectors.
    	 *                 LDVR: numOfBases.
    	 *                 dimension of the vector is numOfBases*numOfBases.
    	 */
    	double[] vl = new double[ldvl*n];
    	/**
    	 * LVDR: (numOfBases)the first dimention of the array VR.
    	 */
    	int ldvr = n;
    	/**
    	 * Vr(LDVI, *): double array
    	 *                 store imaginary part of eigenvectors.
    	 *                 LDVI: n
    	 *                 dimension of the vector is n*n
    	 */
    	double[] vr = new double[ldvr*n];
    	/**
    	 * LWORK: 64*N
    	 */
    	int lwork = 64*n;
    	/**
    	 * WORK: double array
    	 *     LWORK: 64*N
    	 */
    	double[] work = new double[lwork*n];
    	/**
    	 * INFO: on entry: 0
    	 */
    	int info = 0;
    	
    	callingDGEEV(jobvl, jobvr, n, a, lda, wr, wi, vl, ldvl, vr, ldvr, work, lwork, info);
    	
    	/**
    	 * handle results
    	 * eigenvalues and eigenvectors will be in an arraylist.
    	 * 1. real part of eigenvalues
    	 * 2. imaginary part of eigenvalues
    	 * 3. real part of eigenvectors
    	 * 4. imaginary part of eigenvectors
    	 */
    	ArrayList<double[]> resultList = new ArrayList<double[]>();
    	if(info == 0)
    	{
    		resultList.add(wr);
    		resultList.add(wi);
    		resultList.add(vr);
    		resultList.add(vl);
    	}
    	else
    	{
    		System.out.print("failed in ComputeEigenDecomposition.\n");
    		System.exit(-1);
    	}
    	return resultList;
    	
    }

    protected int findOutIndexOfPi(double[] inputVector)
    {
    	double max = -999999.0;
    	int j;
    	int results = 0;
    	
    	/**
    	 * find out the largest eigenvalue index
    	 */
    	for(j=0; j<numOfBases; j++)
    	{
    		if(max < inputVector[j])
    		{
    			max = inputVector[j];
    			results = j;
    		}
    	}
    	return results;
    }

    protected ContainerForNagArrayListForDoubleAndIntArray computeLUComplexNumber(double[] inputRealPart, double[] inputImaginaryPart)
    {
    	ArrayList<double[]> results = new ArrayList<double[]>();
    	
    	/**
    	 * 1. M: the number of rows of the matrix A
    	 */
    	int m = java.lang.Math.max(1, numOfBases);
    	/**
    	 * 2. n: the number of columns of the matrix A
    	 */
    	int n = java.lang.Math.max(1, numOfBases);
    	/**
    	 * 3. a complex array
    	 *    In this function, it will keep as inputRealpart and inputImaginaryPart.
    	 *    These two parts will be regrouped into structure array Complex in the native code.
    	 *    SKIP HANDLING IN THIS FUNCTIONS.
    	 */
    	/**
    	 * 4. LDA: integer:
    	 *         >= max(1, M)
    	 */
    	int lda = m;
    	/**
    	 * 5. IPIV: integer array with size at least m*n
    	 */
    	int[] ipiv = new int[m*n];
    	/**
    	 * 6. INFO: integer
    	 */
    	int info = 0;
    	
    	callingZGETRF(m, n, inputRealPart, inputImaginaryPart, lda, ipiv, info);
    
    	/**
    	 * handle on exit
    	 * When returning, inputRealPart and inputImaginaryPart have the U matrix two parts.
    	 */
		ContainerForNagArrayListForDoubleAndIntArray finalResults = new ContainerForNagArrayListForDoubleAndIntArray();
    	if(info == 0)
    	{
    		results.add(inputRealPart);
    		results.add(inputImaginaryPart);
    	
    		finalResults.setAList(results);
    		finalResults.setIntArray(ipiv, m*n);
    	}
    	else
    	{
    		System.out.print("Failed in computeLU\n");
    		System.exit(-1);
    	}
    	return finalResults;
    	
    }

    protected ArrayList<double[]> computeInverseComplexMatrix(double[] realPart, double[] imaginaryPart, int[] ipiv)
    {
    	ArrayList<double[]> results = new ArrayList<double[]>();
    	/**
    	 * 1. N: integer: the order of A
    	 */
    	int n = numOfBases;
    	/**
    	 * 2. A(LDA, *): complex array
    	 *    in this function, it will be realPart and imaginaryPart. 
    	 *    it will be handled in the native code.
    	 */
    	/**
    	 * 3: LDA(integer): the first dimension of the array A
    	 */
    	int lda = java.lang.Math.max(1,	n);
    	/**
    	 * 4. IPIV(*): return from F07ARF and has passed in as ipiv
    	 */
    	/**
    	 * 5. WORK(*) WILL BE HANDLED IN THE NATIVE CODE.
    	 */
    	/**
    	 * 6. LWORK (INTEGER): 64*n (64: blocksize from example)
    	 */
    	int lwork = 64*n;
    	/**
    	 * 7. INFO (INTEGER): INITIALIZED AS 0; 
    	 */
    	int info = 0;
    	
    	callingZGETRI(n, realPart, imaginaryPart, lda, ipiv, lwork, info);
    	
    	if(info == 0)
    	{
    		results.add(realPart);
    		results.add(imaginaryPart);
    	}
    	else
    	{
    		System.out.println("failed in ComputeInverse\n");
    		System.exit(-1);
    	}
    	
    	return results;
    }

    protected double[][] obtainNewPiVector(double[][] inputMatrix, int indexPi)
    {
    	double[][] results = initializeDoubleArray(numOfBases, numOfBases);
    	double [] tmpVector = initializeDoubleArray(numOfBases);
    	double tmpSum = 0.0;
    	
    	int j;
    	for(j=0; j<numOfBases; j++)
    	{
    		tmpVector[j] = java.lang.Math.abs(inputMatrix[indexPi][j]);
    		tmpSum += tmpVector[j];
    	}
    	for(j=0; j<numOfBases; j++)
    		results[j][j] = tmpVector[j]/tmpSum;
    	
    	return results;
    }

    protected double[][] symmetrizeAMatrix(double[][] inputMatrix)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	int j, k;
    	double tmpValue;
    	
    	for(j=0; j<(numOfBases-1); j++)
    	{
    		for(k=j+1; k<numOfBases; k++)
    		{
    			tmpValue = (inputMatrix[j][k] + inputMatrix[k][j])/2;
    			results[j][k] = results[k][j] = tmpValue;
    		}
    		results[j][j] = inputMatrix[j][j];
    	}
    	results[numOfBases-1][numOfBases-1] = inputMatrix[numOfBases-1][numOfBases-1];
    	return results;
    }

    protected double[][] computeSqrtOfDiagMatrix(double[][] inputMatrix)
    {
    	double[][] results = initializeDoubleArray(numOfBases, numOfBases);
    	int j;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		if(inputMatrix[j][j] >= 0)
    			results[j][j] = java.lang.Math.sqrt(inputMatrix[j][j]);
    		else
    		{
    			System.out.print("negative diagonal in computeSqrtOfDiagmMatrix\n");
    			System.exit(-1);
    		}
    	}
    	return results;
    }

    protected double[][] computeInverseOfDiagMatrix(double[][] inputMatrix)
    {
    	double[][] results = initializeDoubleArray(numOfBases, numOfBases);
    	int j;
    	
    	for(j=0; j<numOfBases; j++)
    	{  
    		if(inputMatrix[j][j] != 0)
    			results[j][j] = 1/inputMatrix[j][j];
    		else
    		{
    			System.out.print("zero diag in computeInverseOfDiagMatrix\n");
    			System.exit(-1);
    		}
    	}
    	return results;
    	
    }

    protected double[] resetEigenvalues(double[] inputVector)
    {
    	double[] results = new double[numOfBases];
    	int j;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		if(inputVector[j]<0)
    			results[j] = 0.0;
    		else
    			results[j] = inputVector[j];
    	}
    	return results;
    }

    protected double[] resetEigenValuesForceContinuousMarkovChain(double[] inputEigenValueVector)
    {
    	double[] results = new double[numOfBases];
    	double maxValue = inputEigenValueVector[0];
    	
    	int j;
    	/**
    	 * find out max value
    	 */
    	for(j=0; j<numOfBases; j++)
    	{
    		if(maxValue < inputEigenValueVector[j])
    			maxValue = inputEigenValueVector[j];
    	}
    	/**
    	 * \lambda_j - max value
    	 */
    	for(j=0; j<numOfBases; j++)
    		results[j] = inputEigenValueVector[j] - maxValue;
    	
    	return results;
    }

    protected double[] takeLogarithmResetEigenvalues(double[] inputVector)
    {
    	double[] results = new double[numOfBases];
    	int j;
    	
    	//System.out.println(Double.NEGATIVE_INFINITY);
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		if(inputVector[j] == 0.0)
    			results[j] = -1.0e+16;
    		else
    			results[j] = java.lang.Math.log(inputVector[j]);
    	}
    	
    	return results;
    }

    protected double[][] handleNegativeOffDiagonalEntriesQ(double[][] inputQMatrix)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	
    	int j, k;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    		{
    			if(j!=k && inputQMatrix[j][k] <= 0.0)
    			{
    				results[j][k] = results[k][j] = 0.0;
    			}
    			else
    				results[j][k] = inputQMatrix[j][k];
    		}
    	}
    	
    	return results;
    }


    protected double[][] resetRateMatrix(double[][] inputRateMatrix)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	double diagonalValue = 0.0;
    	
    	int j, k;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		diagonalValue = 0.0;
    		
    		for(k=0; k<numOfBases; k++)
    		{
    			if(j!=k)
    			{
    				diagonalValue -= inputRateMatrix[j][k];
    				results[j][k] = inputRateMatrix[j][k];
    			}
    		}
    		results[j][j] = diagonalValue;
    	}
    	return results;
    }
    

    protected TreeMap<Double, double[][]> normalizeQ(double[][] inputQMatrix, double[][] inputPiMatrix)
    {
    	TreeMap<Double, double[][]> results = new TreeMap<Double, double[][]>();
    	
    	double normalizationFactor = 0.0;
    	double[][] normalizedQ = new double[numOfBases][numOfBases];
    	int j, k;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		normalizationFactor -= inputQMatrix[j][j]*inputPiMatrix[j][j];
    	}
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    			normalizedQ[j][k] = inputQMatrix[j][k]/normalizationFactor;
    	}
    	results.put(new Double(normalizationFactor), normalizedQ);
    	
    	return results;
    }

    protected ArrayList<double[]> computeEigenDecompositionRealSymMatrix(double[][] inputMatrix)
    {
    	ArrayList<double[]> results = new ArrayList<double[]>();

    	/**
    	 * 1. jobz: 'v' both eigenvalues and eigenvectors
    	 */
    	char jobz = 'V';
    	/**
    	 * 2 UPLO: indicates whether upper or lower triangular part of A is stored. 
    	 * UPLO = "U": upper trangular part of A is stored.
    	 * UPLO = "L": lower trangular part of A is stored.
    	 */
    	char uplo = 'U';
    	/**
    	 * 3. the order of matrix A: numOfBases 
    	 */
    	int n = java.lang.Math.max(1, numOfBases);
    	/**
    	 * 4. A: matrix will be decomposited.
    	 *    dimension n*n
    	 * on exit: A contains the orthogonal matrix Z of eigenvectors.
    	 */
    	double[] a = convertJavaToFortranMatrix(inputMatrix);
    	/**
    	 * 5. LDA: first dimension of the array A
    	 *      (numOfBases)
    	 */
    	int lda = java.lang.Math.max(1, n);
    	/**
    	 * 6. WR: store the real part of eigenvalues
    	 *     dimension is numOfBases
    	 *     if info == 0, w contains eignvalues in ascending order which is the same as F02FAF
    	 */
    	double[] w = new double[java.lang.Math.max(1, n)];
    	double[] work = new double[java.lang.Math.max(1, 3*n)];
    	/**
    	 * 7: LWORK: max(1, 3*N)
    	 */
    	int lwork = java.lang.Math.max(1, 3*n);
    	/**
    	 * IFAIL: on entry: 0
    	 */
    	int info = 0;
    	
    	callingDSYEV(jobz, uplo, n, a, lda, w, work, lwork, info);
    	
    	/**
    	 * handle results
    	 * eigenvalues and eigenvectors will be in an arraylist.
    	 * 1. real part of eigenvalues
    	 * 2. imaginary part of eigenvalues
    	 * 3. real part of eigenvectors
    	 * 4. imaginary part of eigenvectors
    	 */
    	if(info == 0)
    	{
    		results.add(w);
    		results.add(a);
    	}
    	else
    	{
    		System.out.print("failed in ComputeEigenDecomposition.\n");
    		System.exit(-1);
    	}

    	return results;
    }

    protected double[] vectorMultiplyByAConstant(double[] inputVector, double inputFactor)
    {
    	double[] results = new double[numOfBases];
    	int j;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		results[j] = inputFactor*inputVector[j];
    	}
    	return results;
    }

    protected double[] takeExponentialEigenvalues(double[] inputMatrix)
    {
    	double[] results = new double[numOfBases];
    	int j;
    	for(j=0; j<numOfBases; j++)
    	{
    		results[j] = java.lang.Math.exp(inputMatrix[j]);
    		if(results[j] == Double.POSITIVE_INFINITY)
    			results[j] = 1.0e+16;
    	}
    	return results;
    	
    }

    double[][] convertVectorToDiag(double[] inputVector, int dimension)
    {
    	double[][] results = initializeDoubleArray(numOfBases, numOfBases);
    	int j;
    	
    	for(j=0; j<numOfBases; j++)
    		results[j][j] = inputVector[j];
    	
    	return results;
    }


    protected double[][] initializeDoubleArray(int numofRow, 
            int numOfCol)
    {
    	int rowIndex, colIndex;
    	double[][] results = new double[numofRow][numOfCol];
    	for(rowIndex=0; rowIndex<numofRow; rowIndex++)
    	{
    		for(colIndex=0; colIndex<numOfCol; colIndex++)
    			results[rowIndex][colIndex] = 0.0;
    	}
    	return results;
    }

    protected double[] initializeDoubleArray(int numOfElements)
    {
    	double[] tmpArray = new double[numOfElements];

    	for(int j=0; j<numOfElements; j++)
	    {
    		tmpArray[j] = 0.0;
	    }

    	return tmpArray;
    }

    protected ContainerForNagArrayListForDoubleAndIntArray computeLURealNumber(double[] inputRealMatrix)
    {
    	ArrayList<double[]> results = new ArrayList<double[]>();
    	
    	/**
    	 * 1. M: the number of rows of the matrix A
    	 */
    	int m = java.lang.Math.max(1, numOfBases);
    	/**
    	 * 2. n: the number of columns of the matrix A
    	 */
    	int n = java.lang.Math.max(1, numOfBases);
    	/**
    	 * 3. a real array
    	 */
    	double[] a = inputRealMatrix;
    	/**
    	 * 4. LDA: integer:
    	 *         >= max(1, M)
    	 */
    	int lda = java.lang.Math.max(1, m);
    	/**
    	 * 5. IPIV: integer array with size at least m*n
    	 */
    	int[] ipiv = new int[java.lang.Math.max(1, m)];
    	/**
    	 * 6. INFO: integer
    	 */
    	int info = 0;
    	
    	callingDGETRF(m, n, a, lda, ipiv, info);
    
    	/**
    	 * handle on exit
    	 * When returning, inputRealPart and inputImaginaryPart have the U matrix two parts.
    	 */
		ContainerForNagArrayListForDoubleAndIntArray finalResults = new ContainerForNagArrayListForDoubleAndIntArray();
    	if(info == 0)
    	{
    		results.add(a);
    	
    		finalResults.setAList(results);
    		finalResults.setIntArray(ipiv, java.lang.Math.max(1, m));
    	}
    	else
    	{
    		System.out.print("Failed in computeLU\n");
    		System.exit(-1);
    	}
    	return finalResults;
    	
    }

    protected ArrayList<double[]> computeInverseRealMatrix(double[] inputMatrix, int[] ipiv)
    {
    	/**
    	 * NOTES:
    	 * inputMatrix is a one-dimensional array which has the matrix in fortran 
    	 * format.
    	 * 
    	 * ipiv: is the pivot array	
    	 */
    	
    	ArrayList<double[]> results = new ArrayList<double[]>();
    	/**
    	 * 1. N: integer: the order of A
    	 */
    	int n = numOfBases;
    	/**
    	 * 2. A(LDA, *): complex array
    	 *    in this function, it will be realPart and imaginaryPart. 
    	 *    it will be handled in the native code.
    	 */
    	/**
    	 * 3: LDA(integer): the first dimension of the array A
    	 */
    	int lda = java.lang.Math.max(1,	n);
    	/**
    	 * 4. IPIV(*): return from F07ARF and has passed in as ipiv
    	 */
    	/**
    	 * 6. LWORK (INTEGER): 64*n (64: blocksize from example)
    	 */
    	int lwork = 64*n;
    	/**
    	 * 5. WORK(*) WILL BE HANDLED IN THE NATIVE CODE.
    	 */
    	double[] work = new double[lwork];
    	/**
    	 * 7. INFO (INTEGER): INITIALIZED AS 0; 
    	 */
    	int info = 0;
    	
    	callingDGETRI(n, inputMatrix, lda, ipiv, work, lwork, info);
    	
    	if(info == 0)
    	{
    		results.add(inputMatrix);
    	}
    	else
    	{
    		System.out.println("failed in ComputeInverse\n");
    		System.exit(-1);
    	}
    	
    	return results;
    }

    double[][] computeMultiplicationOfTwoMatrices(double[][] inputFirst, double[][] inputSecond)
    {
    	double[][] results;
    	/**
    	 * a: the first operand. 
    	 *    it is inputFirst
    	 */
   	
    	double[] a = convertJavaToFortranMatrix(inputFirst);
    	/**
    	 * b : the second operand. 
    	 *       it is inputSecon
    	 */
    	double[] b = convertJavaToFortranMatrix(inputSecond);
    	/**
    	 * c: the results
    	 *    for input, it is a zero matrix as required.
    	 */
    	double[] c = getAZeroMatrix(numOfBases);
    	/**
    	 * N(Integer): the first dimension of A
    	 */
    	int n = numOfBases;
    	/**
    	 * M(Integer): the second dimension of B
    	 */
    	int m = numOfBases;
    	/**
    	 * K(Integer): columns of a
    	 */
    	int k = numOfBases;
   	/**
    	 * lda: leading dimension of a
    	 */
    	int lda = numOfBases;
    	/**
    	 * ldb: leading dimension of b
    	 */
    	int ldb = numOfBases;
    	/**
    	 * ldc: leading dimension of c
    	 */
    	int ldc = numOfBases;
    	/**
    	 * dummy scalar alpha and beta
    	 */
    	double alpha = 1.0;
    	double beta = 1.0; 
    	/**
    	 * transpose of a: 'n': no transpose
    	 * transpose of b: 'n': no transpose
    	 */
    	char transa = 'n';
    	char transb = 'n';
    	
    	/**
    	 * no flag on exit which indicates the status
    	 */
    	callingDGEMM(transa, transb, m, n, k, alpha, a, lda, b, ldb, beta, c, ldc);
    	
    	results = convertFortranToJavaMatrix(c);
    	
    	return results;
    }
    
    protected double[] convertJavaToFortranMatrix(double[][] inputMatrix)
    {
    	/**
    	 * The fortran matrix is stored in columnwise in a vector.
    	 */
    	double[] results = new double[numOfBases*numOfBases];
    	int j, k, counter=0;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    			results[counter++] = inputMatrix[k][j];
    	}
    	return results;
    }

    protected double[][] convertFortranToJavaMatrix(double[] inputMatrix)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	int j, k, counter=0;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    			results[k][j] = inputMatrix[counter++];
    	}
    	
    	return results;
    }
    
    protected void sumOverColumnsOfAMatrix(double[][] inputMatrix)
    {
    	int j, k;
    	double[] tmpResult = new double[numOfBases];
    	String finalResults = "matrix: \n";
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		tmpResult[j] = 0.0;
    		for(k=0; k<numOfBases; k++)
    		{
    			tmpResult[j] += inputMatrix[j][k];
    			finalResults += inputMatrix[j][k] + "\t";
    		}
    		finalResults += "\n";
    	}
    	
    	finalResults += "sum over columns: \n";
    	
    	for(j=0; j<numOfBases; j++)
    		finalResults += tmpResult[j] + "\t";
    	
    	finalResults += "\n";
    	
    	//System.out.println(finalResults);
    }

    protected double computeDifOfTwoMatrices(double[][] inputFirst, double[][] inputSecond)
    {
    	int j, k;
    	double results = 0.0, tmp;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    		{
    			tmp = inputFirst[j][k] - inputSecond[j][k];
    			results += tmp*tmp;
    		}
    	}
    	return results;
    }

    protected void computePermutations()
    {
    	int index;
    	String nodeName, intParentName;
    	double[][] tmpMatrix;
    	NodeStructure nodeStructure; 
	    	
    	/**
    	 * the order in the permutation container is the same as in the branchList
    	 */
    	int numOfIntNodes = intNodesList.size();
    	int numOfLeafNodes = leafNodesList.size();
    	
    	for(index=0; index<numOfIntNodes; index++)
    	{
    		nodeName = (String)intNodesList.get(index);
    		tmpMatrix = (double[][])branchList.get(nodeName).getJointProbability();
    		nodeStructure = (NodeStructure)phylogeneticTree.get(nodeName);
    		intParentName = (String)nodeStructure.getParentNode();
    		if(intParentName.equals(rootNode))
    		{
    			/**
    			 * permute columns only
    			 * ptcOrctp: 0: parent to child
    			 *           1: child to parent
    			 */
    			permuteAMatrix(tmpMatrix, nodeName, 2, 0);
    			/**
    			 * child to parent
    			 */
    			tmpMatrix = obtainTranspose(tmpMatrix);
    			permuteAMatrix(tmpMatrix, nodeName, 1, 1);
    		}
    		else
    		{
    			/**
    			 * this is an internal to internal branch
    			 * both rows and columns need to be permuted
    			 */
    			permuteAMatrix(tmpMatrix, nodeName, 3, 0);
    			/**
    			 * child to parent
    			 */
    			tmpMatrix = obtainTranspose(tmpMatrix);
    			permuteAMatrix(tmpMatrix, nodeName, 3, 1);
	    	}
	    }
    	/**
    	 * work on leaf nodes
    	 */
    	for(index=0; index<numOfLeafNodes; index++)
	    {
    		nodeName = (String)leafNodesList.get(index);
    		tmpMatrix = (double[][])branchList.get(nodeName).getJointProbability();
    		/**
    		 * for all leaf nodes, only rows need to be permuted
    		 */
    		permuteAMatrix(tmpMatrix, nodeName, 1, 0);
    		
    		/**
    		 * child to parent
    		 */
    		tmpMatrix = obtainTranspose(tmpMatrix);
    		permuteAMatrix(tmpMatrix, nodeName, 2, 1);

	    	}
	    	/**
	    	 * write results
	    	 */
	    	//writePermutationsContainer();
    }
    
    protected double[][] obtainTranspose(double[][] inputMatrix)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	int j, k;
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    			results[j][k] = inputMatrix[k][j];
    	}
    	return results;
    }
 
    protected void writePermutationsContainer()
    {
    	String tmpOutputLine, tmpNodeName;
     	ArrayList<double[][]> tmpList;
     	double[][] tmpMatrix;
     	//Iterator itr;
     	
    	int index, parentIndex, childIndex, conSize, j;

    	FileWriter myFw = null;
    	
    	try
    	{
    		
    		index = dataPath.lastIndexOf("/");
    		myFw = new FileWriter(dataPath.substring(0,index) + "/permutationMatrices.txt");
    	    /**
    	     * parent to child
    	     */
       		int numOfIntNodes = intNodesList.size();
   		
       		for(index=0; index<numOfIntNodes; index++)
    		{
    			tmpNodeName = (String)intNodesList.get(index);
    			tmpList = (ArrayList<double[][]>)permutationsParentToChild.get(tmpNodeName);
    			conSize = tmpList.size();
    			for(j=0; j<conSize; j++)
    			{
    				tmpMatrix = (double[][])tmpList.get(j);
    				tmpOutputLine = "";
    				for(parentIndex=0; parentIndex<numOfBases; parentIndex++)
    				{
    					/**
    					 * add spaces between terms
    					 */
    					for(childIndex=0; childIndex<numOfBases; childIndex++)
    					{
        					if(!tmpOutputLine.equals("")) 
        						tmpOutputLine += " ";
    						tmpOutputLine += tmpMatrix[parentIndex][childIndex]; 
    					}
    					tmpOutputLine += "\n";
    					myFw.write(tmpOutputLine);
    					tmpOutputLine = "";
    				}
    			}
    		}

    		int numOfLeafNodes = leafNodesList.size();
    		for(index=0; index<numOfLeafNodes; index++)
    		{
    			tmpNodeName = (String)leafNodesList.get(index);
    			tmpList = (ArrayList<double[][]>)permutationsParentToChild.get(tmpNodeName);
    			conSize = tmpList.size();
    			for(j=0; j<conSize; j++)
    			{
    				tmpMatrix = (double[][])tmpList.get(j);
    				tmpOutputLine = "";
    				for(parentIndex=0; parentIndex<numOfBases; parentIndex++)
    				{
    					/**
    					 * add spaces between terms
    					 */
    					for(childIndex=0; childIndex<numOfBases; childIndex++)
    					{
        					if(!tmpOutputLine.equals("")) 
        						tmpOutputLine += " ";
    						tmpOutputLine += tmpMatrix[parentIndex][childIndex]; 
    					}
    					tmpOutputLine += "\n";
    					myFw.write(tmpOutputLine);
    					tmpOutputLine = "";
    				}
    			}
    		}
   		
    	    /**
    	     * child to parent
    	     */
     		for(index=0; index<numOfIntNodes; index++)
    		{
    			tmpNodeName = (String)intNodesList.get(index);
    			tmpList = (ArrayList<double[][]>)permutationsChildToParent.get(tmpNodeName);
    			conSize = tmpList.size();
    			for(j=0; j<conSize; j++)
    			{
    				tmpMatrix = (double[][])tmpList.get(j);
    				tmpOutputLine = "";
    				for(parentIndex=0; parentIndex<numOfBases; parentIndex++)
    				{
    					/**
    					 * add spaces between terms
    					 */
    					for(childIndex=0; childIndex<numOfBases; childIndex++)
    					{
        					if(!tmpOutputLine.equals("")) 
        						tmpOutputLine += " ";
    						tmpOutputLine += tmpMatrix[parentIndex][childIndex]; 
    					}
    					tmpOutputLine += "\n";
    					myFw.write(tmpOutputLine);
    					tmpOutputLine = "";
    				}
    			}
    		}

    		for(index=0; index<numOfLeafNodes; index++)
    		{
    			tmpNodeName = (String)leafNodesList.get(index);
    			tmpList = (ArrayList<double[][]>)permutationsChildToParent.get(tmpNodeName);
    			conSize = tmpList.size();
    			for(j=0; j<conSize; j++)
    			{
    				tmpMatrix = (double[][])tmpList.get(j);
    				tmpOutputLine = "";
    				for(parentIndex=0; parentIndex<numOfBases; parentIndex++)
    				{
    					/**
    					 * add spaces between terms
    					 */
    					for(childIndex=0; childIndex<numOfBases; childIndex++)
    					{
        					if(!tmpOutputLine.equals("")) 
        						tmpOutputLine += " ";
    						tmpOutputLine += tmpMatrix[parentIndex][childIndex]; 
    					}
    					tmpOutputLine += "\n";
    					myFw.write(tmpOutputLine);
    					tmpOutputLine = "";
    				}
    			}
    		}
    		myFw.close();
    	}
     	catch(Exception e)
    	{
    		System.out.println("close file failed in writePermutationsContainer");
    		System.exit(-1);
    	}	
    }

    protected void permuteAMatrix(double[][] inputMatrix, String nodeName, int rowOrColumn, int ptcOrctp)
    {
    	ArrayList<double[][]> results = new ArrayList<double[][]>();
    	ArrayList<double[][]> tmpColumnResults = new ArrayList<double[][]>();
    	
    	double[][] tmpResults;
    	int[] tmpIndex = new int[4];
    	int j, k, l; 
    	int tmpIndex1, tmpIndex2;
    	int flag = 0;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    		{
    			if(j != k)
    			{
    				tmpIndex1 =0; 
    				tmpIndex2 =0; 
    				flag = 0;
    				
    				for(l=0; l<numOfBases; l++)
    				{
    					if(l != j && l != k && flag == 0)
    					{
    						tmpIndex1 = l;
    						flag = 1;
    					}
    					else if(l != j && l != k && flag != 0)
    						tmpIndex2 = l;
    				}
    				/**
    				 * set indices
    				 */
    				tmpIndex[0] = j;
    				tmpIndex[1] = k;
    				tmpIndex[2] = tmpIndex1;
    				tmpIndex[3] = tmpIndex2;
    				
    				if(rowOrColumn == 1)
    				{
    					tmpResults = permuteAMatrixByRow(inputMatrix, tmpIndex);
    					results.add(tmpResults);
        				tmpIndex[0] = j;
        				tmpIndex[1] = k;
        				tmpIndex[2] = tmpIndex2;
        				tmpIndex[3] = tmpIndex1;
    					tmpResults = permuteAMatrixByRow(inputMatrix, tmpIndex);
    					results.add(tmpResults);
    				}
    				else if(rowOrColumn == 2)
    				{
    					tmpResults = permuteAMatrixByColumn(inputMatrix, tmpIndex);
    					results.add(tmpResults);
        				tmpIndex[0] = j;
        				tmpIndex[1] = k;
        				tmpIndex[2] = tmpIndex2;
        				tmpIndex[3] = tmpIndex1;
    					tmpResults = permuteAMatrixByColumn(inputMatrix, tmpIndex);
    					results.add(tmpResults);
    				}
    				else if(rowOrColumn == 3)
    				{
    					tmpResults = permuteAMatrixByRow(inputMatrix, tmpIndex);
    					tmpColumnResults = permuteAMatrixColumns(tmpResults);
    					results.addAll(tmpColumnResults);
        				tmpIndex[0] = j;
        				tmpIndex[1] = k;
        				tmpIndex[2] = tmpIndex2;
        				tmpIndex[3] = tmpIndex1;
    					tmpResults = permuteAMatrixByRow(inputMatrix, tmpIndex);
    					tmpColumnResults = permuteAMatrixColumns(tmpResults);
    					results.addAll(tmpColumnResults);
   				}
    			  }
    			}
    		}
    		/**
    		 * trim array size to exact what it is.
    		 */
     		results.trimToSize();
       		/**
    		 * write results into pemutations container
    		 */
     		if(ptcOrctp == 0)/* parent to child */
     			permutationsParentToChild.put(nodeName, results);
     		else /* child to parent */
     			permutationsChildToParent.put(nodeName, results);
    	}

    protected ArrayList<double[][]> permuteAMatrixColumns(double[][] inputMatrix)
    {
    	ArrayList<double[][]> results = new ArrayList<double[][]>();

    	double[][] tmpResults;
    	int[] tmpIndex = new int[4];
    	int j, k, l; 
    	int tmpIndex1, tmpIndex2, flag=0;
    	
    	for(j=0; j<numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    		{
    			if(j != k)
    			{
    				tmpIndex1 =0; 
    				tmpIndex2 =0; 
    				flag = 0;
    				for(l=0; l<numOfBases; l++)
    				{
    					if(l != j && l != k && flag == 0)
    					{
    						tmpIndex1 = l;
    						flag = 1;
    					}
    					else if(l != j && l != k && flag != 0)
    						tmpIndex2 = l;
    				}
    				/**
    				 * set indices
    				 */
    				tmpIndex[0] = j;
    				tmpIndex[1] = k;
    				tmpIndex[2] = tmpIndex1;
    				tmpIndex[3] = tmpIndex2;
    				tmpResults = permuteAMatrixByColumn(inputMatrix, tmpIndex);
    				results.add(tmpResults);
    				tmpIndex[2] = tmpIndex2;
    				tmpIndex[3] = tmpIndex1;
    				tmpResults = permuteAMatrixByColumn(inputMatrix, tmpIndex);
    				results.add(tmpResults);
    			}
    		}
    	}
    	return results;
    }
    
    protected double[][] permuteAMatrixByColumn(double[][] inputMatrix, int[] indices)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	int j, k; 
    	
    	for(j = 0; j < numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    		{
    			results[k][j]= inputMatrix[k][indices[j]];
    		}
    	}
    	
    	return results;
    }

    protected double[][] permuteAMatrixByRow(double[][] inputMatrix, int[] indices)
    {
    	double[][] results = new double[numOfBases][numOfBases];
    	int j, k; 
    	
    	for(j = 0; j < numOfBases; j++)
    	{
    		for(k=0; k<numOfBases; k++)
    		{
    			results[j][k]= inputMatrix[indices[j]][k];
    		}
    	}
    	
    	return results;
    }

	public void readJoints()
	{
		BranchDetails bd;
		double jValArray[][] = new double[4][4];
		int i,j,k;
		String rowValues;
		StringTokenizer str;

		FileReader aReader;

		try
		{
			aReader = new FileReader(dataPath);
			BufferedReader aBuffer = new BufferedReader(aReader);

			//Initialize for all internal nodes	
			/**
			 *modified for mixture model
			 */
			for(i=0; i<intNodesList.size(); i++)
			{						
				for(j=0; j<4; j++)
				{
					//Read a row of Q-matrix and identify A, C, G, T values
					rowValues = aBuffer.readLine();
					str = new StringTokenizer(rowValues, " \t");
					
					for(k=0; k<4; k++)
					{						
						jValArray[j][k] = Double.parseDouble((str.nextToken()));
					}
				}	
						
				bd = new BranchDetails();			
				bd.setJointProbability(jValArray);
				/**
				 * add on Mar.26, 2008
				 * use child vector to hold the branch frequencies
				 */
				bd.setFrequencyChild(jValArray);

				branchList.put((String)intNodesList.get(i), bd);

				//Read the blankline separating two Q-matrices
				rowValues = aBuffer.readLine();				
			}

			//Initialize for all leaf nodes
			for(i=0; i<leafNodesList.size(); i++)
				{
					//System.out.println("Reading for leaf node: " 
					//						+ leafNodesList.get(i));

					for(j=0; j<4; j++)
					{
						//Read a row of Q-matrix and identify A, C, G, T values
						rowValues = aBuffer.readLine();
						str = new StringTokenizer(rowValues, " \t");

						for(k=0; k<4; k++) 
						{
							jValArray[j][k] = Double.parseDouble((str.nextToken()));
						}
					}

					bd = new BranchDetails();
					bd.setJointProbability(jValArray);

					branchList.put((String)leafNodesList.get(i), bd);
					rowValues = aBuffer.readLine();
				}
			aReader.close();
		}//end try
		catch(Exception e)
		{
			System.out.println("Error Occured while initializing Q-values "
					+ e);				
		}		

	}
	
	public String getTreeNotation()
	{
		String treeNotation = new String();

		try
		{
			BufferedReader aBuffer = new BufferedReader(new FileReader(treeFilePath));
			
			treeNotation = aBuffer.readLine();
		} 
		catch(Exception e) 
		{
			e.printStackTrace();
		}

		return treeNotation;
				
	}
	public void getTreeTopology()
	{
	    //Helper class: TreeTraversal(boolean displayOutput) 
	    NewickTreeTraversal treeHandler = new NewickTreeTraversal(false);

		treeNotation = getTreeNotation();
		treeHandler.generateTree(treeNotation, phylogeneticTree);
		rootNode = treeHandler.getRootNode(phylogeneticTree);
		
		leafNodesList = (ArrayList<String>)treeHandler.determineLeafNodes(phylogeneticTree);	
		intNodesList = (ArrayList<String>)treeHandler.determineInternalNodes(phylogeneticTree);	
		
	}
	
	public void createJPGFile(String treeNotation)
	{
		DisplayTree aDisplayer = new DisplayTree(phylogeneticTree, rootNode, leafNodesList);
		int index = dataPath.lastIndexOf("/");
		String jpgFilePath = dataPath.substring(0,index) + "/tree.jpg";
		aDisplayer.createJPGFile(jpgFilePath); 

	}
	/**
	 * obtain the best joints
	 */
	
	protected String getTheBestRoot()
	{
			String bestRoot=null; 
			String tmpRoot; 
			
			int numOfInternalNode = intNodesList.size();
			int numOfLeafNode = leafNodesList.size();
			
			double minSumOfSquares = 10^9;
			
			double tmpMinSumOfSquares;
			int j; 
			
			ContainerShortestPathSearchResults tmpContainer; 
			
			for(j=0; j<numOfInternalNode; j++)
			{
				tmpRoot = intNodesList.get(j);
				tmpContainer = shortestPathReconstructedTreesRootingAtANode.get(tmpRoot);
				tmpMinSumOfSquares = tmpContainer.getMinimumSumOfSquares();
				
				if(tmpMinSumOfSquares < minSumOfSquares)
				{
					minSumOfSquares = tmpMinSumOfSquares;
					bestRoot = tmpRoot;
				}
			}
			
			for(j=0; j<numOfLeafNode; j++)
			{
				tmpRoot = leafNodesList.get(j);
				tmpContainer = shortestPathReconstructedTreesRootingAtANode.get(tmpRoot);
				tmpMinSumOfSquares = tmpContainer.getMinimumSumOfSquares();
				
				if(tmpMinSumOfSquares < minSumOfSquares)
				{
					minSumOfSquares = tmpMinSumOfSquares;
					bestRoot = tmpRoot;
				}
			}
			
			return bestRoot;
	}
	
/*	protected void obtainBestNSGTRJoints(String bestRoot)
	{
		
	}*/
	
	protected void writeBestShortestPath(String bestRoot, boolean isFirst)
	{
    	int k, m, n, index;   
    	int numOfIntNodes = intNodesList.size(); 
    	int numOfLeafNodes = leafNodesList.size();
    	String tmpLine, nodeName, oldParent, newParent, nodeType;
    	
    	Node newNodeStruction = null;
    	NodeStructure oldNodeStructure = null;
    	
    	FileWriter myFw = null;
    	
    	TreeMap<String, double[][]> tmpResults = null;
    	TreeMap<String, Node> reconstructedTreeAtBestRoot = null;
    	
    	double[][] tmpMatrix = null;
    	
    	try
    	{
    		
    		index = dataPath.lastIndexOf("/");
    		index = dataPath.lastIndexOf("/");

    		//File myFile = new File(dataPath.substring(0,index) + "/bestFitJointProbabilityMatrices.txt");
    		
    		if(isFirst == false)                                       
    			myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitJointProbabilityMatrices.txt", true);
    		else
    			myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitJointProbabilityMatrices.txt");
   		
     		index = bestRoot.indexOf("-");
 
     		tmpLine = "Rooting at node " + bestRoot.substring(0, index) + "\n\n";
    		myFw.write(tmpLine);
    		
    		/**
    		 * go through root node first
    		 * root node is not included in either internal node or leaf node.
    		 * it has to be handled specially
    		 */
			tmpResults = shortestPathReconstructedTreesRootingAtANode.get(bestRoot).getBestFitNonStationaryJointProbMatrices();			
			reconstructedTreeAtBestRoot = shortestPathReconstructedTreesRootingAtANode.get(bestRoot).getRecontructedTree();
			
			/**
 			 * go through internal nodes
 			 */
			for(k=0; k<numOfIntNodes; k++)
			{
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
    			nodeName = intNodesList.get(k);
			
    			newNodeStruction = reconstructedTreeAtBestRoot.get(nodeName);
    			newParent = newNodeStruction.getParent();
    			oldNodeStructure = phylogeneticTree.get(nodeName);
    			oldParent = oldNodeStructure.getParentNode();
    			
    			
    			if(nodeName.equals(bestRoot))
    			{
    				/**
    				 * get the case: 
    				 * old root is the parent of new root
    				 */
    				if(rootNode.equals(oldParent))
    				{
    					/**
    					 * rooting at the first internal node.
    					 * now rootNode holds the joints.
    					 */
    					tmpMatrix = tmpResults.get(rootNode);
    					tmpLine += "(" + nodeName.substring(0,nodeName.indexOf('-')) + ", " +  rootNode.substring(0, rootNode.indexOf('-')) + "): \n";
    				}
    				else 
    				{
    					/**
    					 * internal to internal. old parent is still the parent
    					 * of new root but it holds the joints.
    					 * 
    					 */
    					tmpLine += "(" + nodeName.substring(0,nodeName.indexOf('-')) + ", " +  oldParent.substring(0, oldParent.indexOf('-')) + "): \n";
    					tmpMatrix = tmpResults.get(oldParent);
    				}
    			}
    			else if(oldParent.equals(newParent))
    			{
    				/**
    				 * internal branch without reverse.
    				 * nodeName holds the joints
    				 */
    				tmpMatrix = tmpResults.get(nodeName);
    				tmpLine += "(" + oldParent.substring(0,oldParent.indexOf('-')) + ", " +  nodeName.substring(0, nodeName.indexOf('-')) + "): \n";
    			}
    			else 
    			{
    				 /**
    				  * Internal edge in reverse direction   				
    				  */
    				tmpLine += "(" + nodeName.substring(0, nodeName.indexOf('-')) + ", " + oldParent.substring(0, oldParent.indexOf('-')) + "):\n";
    				tmpMatrix = tmpResults.get(oldParent);
    				//tmpMatrix = obtainTranspose(tmpMatrix);
    			}
    			
				for(m=0; m<numOfBases; m++)
				{
					for(n=0; n<numOfBases; n++)
						tmpLine += tmpMatrix[m][n] + "\t ";
					tmpLine += "\n";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
			}
			/**
			 * go through leaf nodes
			 */
			for(k=0; k<numOfLeafNodes; k++)
			{
 	   			/**
    			 * reset tmpLine
    			 */
 				tmpLine = "";
    			nodeName = leafNodesList.get(k);
    			oldParent = phylogeneticTree.get(nodeName).getParentNode();
    			
    			if(nodeName.equals(bestRoot))
    			{
    				tmpLine += "(" + nodeName.substring(0, nodeName.indexOf('-')) + ", " +  oldParent.substring(0, oldParent.indexOf('-')) + "): \n";

    				/**
    				 * from old root to leaf; get Js under nodeName
    				 */
    				tmpMatrix = tmpResults.get(oldParent);
    			}
    			else
    			{
    				/**
    				 * Normal leaf node
    				 */
        			tmpLine += "(" + oldParent.substring(0, oldParent.indexOf('-')) + ", " +  nodeName.substring(0, nodeName.indexOf('-')) + "): \n";
        			tmpMatrix = tmpResults.get(nodeName);
    			}
				
				for(m=0; m<numOfBases; m++)
				{
					for(n=0; n<numOfBases; n++)
						tmpLine += tmpMatrix[m][n] + "\t ";
					tmpLine += "\n";
				}
				tmpLine += "\n";
				myFw.write(tmpLine);
			}
			
			myFw.close();
    	}
    	catch (Exception e) 
    	{
       		System.out.println("failed in writeBestShortestPath");
    		System.exit(-1);
		}
	}
	
	/**
	 * Implement nonstaionary edge length of simplified Minin and Suchard
	 */
	protected void calculateNonstationaryEdgeLengthATree()
	{
		
		int numOfIntNodes = intNodesList.size();
		int numOfLeafNodes = leafNodesList.size();
		int k;
		double[][] tmpMatrix=null, tmpTransportMatrix;
		double[] tmpLengthsAnEdge; 
		
		TreeMap<String, double[][]>tmpResults = shortestPathReconstructedTreesRootingAtANode.get(rootNode).getBestFitNonStationaryJointProbMatrices();			
		
		String tmpNodeName; 
		/**
		 * go through internal nodes
		 */
		for(k=0; k<numOfIntNodes; k++)
		{
			tmpNodeName = intNodesList.get(k);
			tmpMatrix = tmpResults.get(tmpNodeName);
			tmpLengthsAnEdge = new double[3];
			tmpLengthsAnEdge[0] = calculateNonstationaryEdgeLength(tmpMatrix);
			tmpTransportMatrix = obtainTranspose(tmpMatrix);
			tmpLengthsAnEdge[1] = calculateNonstationaryEdgeLength(tmpTransportMatrix);
			tmpLengthsAnEdge[2] = (tmpLengthsAnEdge[0] + tmpLengthsAnEdge[1])/2;
			edgeLengths.put(tmpNodeName, tmpLengthsAnEdge);
		}
		/**
		 * go through leaf nodes
		 */
		for(k=0; k<numOfLeafNodes; k++)
		{
			tmpNodeName = leafNodesList.get(k);
			tmpMatrix = tmpResults.get(tmpNodeName);
			tmpLengthsAnEdge = new double[3];
			tmpLengthsAnEdge[0] = calculateNonstationaryEdgeLength(tmpMatrix);
			tmpTransportMatrix = obtainTranspose(tmpMatrix);
			tmpLengthsAnEdge[1] = calculateNonstationaryEdgeLength(tmpTransportMatrix);
			tmpLengthsAnEdge[2] = (tmpLengthsAnEdge[0] + tmpLengthsAnEdge[1])/2;
			edgeLengths.put(tmpNodeName, tmpLengthsAnEdge);
		}
	}
   				
	/**
	 * calculate nonstationary edge length using formula (4) in the manuscript
	 */
	protected double calculateNonstationaryEdgeLength(double[][] inputMatrix) 
	{
		double resultEdgeLength=0.0; 
		
		double[] startingFrequencies = null;

		/**
		 * obtain frequencies
		 */
		startingFrequencies = computeNULLMarginalFrequencies(inputMatrix, 1);//always from ptc since ctp has been transporsed.
		double[][] piMatrix = null; 
	    /**
	     * obtain frequencies matrix \Pi
	     */
		piMatrix = convertVectorToDiag(startingFrequencies, numOfBases);
		/**
		 * declair temporary storage
		 */
		double[][] tmpMatrix1 = null;
		double[][] tmpMatrix; 
		ArrayList<double[]> tmpResults = null;
		/**
		 * \Pi^{-1}
		 */
		tmpMatrix1 = computeInverseOfDiagMatrix(piMatrix);
		
		/**
		 * tmpMatrix = Q
		 */
		tmpMatrix = inputMatrix;
		/**
		 * compute conditional probability matrix P
		 */
		double[][] originalCond = null;
		originalCond = computeMultiplicationOfTwoMatrices(tmpMatrix1, tmpMatrix);
		/**
		 * compute eigen decomposition of P
		 * Note: it is possible that P gets a complex valued result
		 */
		ArrayList<double[]> eigenResults = null;
		eigenResults = computeEigenDecomposition(originalCond);
		
		/**
		 * reorganize eigen decomposition
		 */
		ArrayList<double[]> reorganizedEigenResults = null;
		
		reorganizedEigenResults = reorganizeEigenResults(eigenResults);
		
		/**
		 * obtain real part of eigen value. 
		 * The "one" corresponds to frequencies.
		 * The imaginary part of eigen value is not useful
		 */
		double[] realPartEigenValues = null; 
		realPartEigenValues = eigenResults.get(0);
		/**
		 * obtain the largest eigenvalue index
		 */
		int indexPi;
		indexPi = findOutIndexOfPi(realPartEigenValues);
		/**
		 * obtain eigenvectors U^{-1} of P in order to obtain \pi_{new}
		 * since the left eigenvector of P which corresponds to the largest
		 * eigenvalue is always real, imaginary part is not very important.
		 */
		double[] realPartEigenVectors = null;
		realPartEigenVectors = reorganizedEigenResults.get(4);//real part of eigenvectors
		/**
		 * construct a numOfBases*numOfBases matrix
		 */
		tmpMatrix = convertFortranToJavaMatrix(realPartEigenVectors);
		/**
		 * obtain best fit \Pi matrix
		 */
		double[][] piNew = obtainNewPiVector(tmpMatrix, indexPi);
		
		//System.out.println("pi new: ");
		/**
		 * for debug
		 */
		//sumOverColumnsOfAMatrix(piNew);
		/**
		 * obtain symmetrized joint probablity matrix
		 */
		double[][] newJointProb = null;
		/**
		 * best fit joint = best fit \Pi * P
		 */
		newJointProb = computeMultiplicationOfTwoMatrices(piNew, originalCond);
		/**
		 * force symmetry
		 */
		newJointProb = symmetrizeAMatrix(newJointProb);
		/**
		 * compute sqrt(best fit \Pi) and inverse sqrt(best fit \Pi)
		 */
		double[][] piNewSqrt = null; 
		double[][] piNewSqrtInverse = null;
		piNewSqrt = computeSqrtOfDiagMatrix(piNew);
		piNewSqrtInverse = computeInverseOfDiagMatrix(piNewSqrt);
		/**
		 * P_{sym} = \pi^{-\frac{1}{2}} %*% F %*% \pi&{-\frac{1}{2}}
		 * 
		 */
		double[][] newProbSymmetric = null;
		/*
		 *  \pi^{-\frac{1}{2}} %*% F
		 */
		tmpMatrix1 = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newJointProb);
		/*
		 * \pi^{-\frac{1}{2}} %*% F %*% \pi&{-\frac{1}{2}}
		 */
		newProbSymmetric = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrtInverse);
		
		//System.out.println("P_sym:");
		//sumOverColumnsOfAMatrix(newProbSymmetric);
		/**
		 * computeEigenDecompositionRealSymMatrix return values
		 * [0]: eigenvectors;
		 * [1]: eigenvalues;
		 */
		ArrayList<double[]> eigenDecompositionOfReContructedCondMatrix = null;
		eigenDecompositionOfReContructedCondMatrix = computeEigenDecompositionRealSymMatrix(newProbSymmetric);
		/**
		 * eigen values
		 */
		double[] newEigenValuesOfReconstructedCondMatrix = null;
		newEigenValuesOfReconstructedCondMatrix = eigenDecompositionOfReContructedCondMatrix.get(0);
		/**
		 * if any negative eigenvalues, reset as zero.
		 */
		newEigenValuesOfReconstructedCondMatrix = resetEigenvalues(newEigenValuesOfReconstructedCondMatrix);
		/**
		 * take logarithm
		 */
		double[] newEigenValuesLogOfReconstructedCondMatrix = takeLogarithmResetEigenvalues(newEigenValuesOfReconstructedCondMatrix);
		//System.out.println(newEigenValuesOfReconstructedCondMatrix[0]+ " " + newEigenValuesOfReconstructedCondMatrix[1]+ " " + newEigenValuesOfReconstructedCondMatrix[2]+" " +  newEigenValuesOfReconstructedCondMatrix[3]);
		/**
		 * construct a new diagonal eigenvalues matrix
		 */
		double[][] diagonalNewEigenValuesLogOfReconstructedCondMatrix = null;
		diagonalNewEigenValuesLogOfReconstructedCondMatrix = convertVectorToDiag(newEigenValuesLogOfReconstructedCondMatrix, numOfBases);
		
		/**
		 * obtain right eigenvectors of P_{sym}
		 */
		double[] tmpVector = null;
		tmpVector = eigenDecompositionOfReContructedCondMatrix.get(1);
		/**
		 * convert right eigenvectors of P_{sym} into a java style matrix
		 */
		double[][] newEigenVectorsOfReconstructedCondMatrixSym = null;
		newEigenVectorsOfReconstructedCondMatrixSym = convertFortranToJavaMatrix(tmpVector);
		/**
		 * desymmetrizing right eigenvectors of P_{sym}
		 */
		double[][] newEigenVectorsOfReconstructedCondMatrix = null;
		newEigenVectorsOfReconstructedCondMatrix = computeMultiplicationOfTwoMatrices(piNewSqrtInverse, newEigenVectorsOfReconstructedCondMatrixSym);
		/**
		 * preparing inversing right eigenvectors of P_{sym}
		 * lapack asks for LU decomposited matrix for computing inverse matrix
		 */
		ContainerForNagArrayListForDoubleAndIntArray luOfEigenVectors = computeLURealNumber(tmpVector);
		int[] ipiv = luOfEigenVectors.getintArray();
		tmpResults = luOfEigenVectors.getAList();
		realPartEigenVectors = tmpResults.get(0);//use old vector for LU results of a real matrix
		ArrayList<double[]> inverseEigenVectors = computeInverseRealMatrix(realPartEigenVectors, ipiv);
		/**
		 * now tmpMatrix1 has U_{sym}^{-1}
		 */
		tmpMatrix1 = convertFortranToJavaMatrix(inverseEigenVectors.get(0));
		/**
		 * desymmetrizing left eigenvectors of P_{sym} (U^{-1}_{sym} * sqrt{\Pi})
		 */
		double[][] newEigenVectorsOfReconstructedCondMatrixInversed = computeMultiplicationOfTwoMatrices(tmpMatrix1, piNewSqrt);
		/**
		 * compute log cond.
		 */
		tmpMatrix = computeMultiplicationOfTwoMatrices(newEigenVectorsOfReconstructedCondMatrix, diagonalNewEigenValuesLogOfReconstructedCondMatrix);
		tmpMatrix = computeMultiplicationOfTwoMatrices(tmpMatrix, newEigenVectorsOfReconstructedCondMatrixInversed);
		//System.out.println("Q*: ");
		//sumOverColumnsOfAMatrix(tmpMatrix);
		
		/**
		 * set negative off-diagonal as zero
		 */
		tmpMatrix = handleNegativeOffDiagonalEntriesQ(tmpMatrix);
		//System.out.println("after handle negative off-diagonal: ");
		//sumOverColumnsOfAMatrix(tmpMatrix);
		
		/**
		 * the rate matrix should keep Q_{jj} = -\sum{j: j\neq i} Q_{i, j}
		 * after setting negative off-diagonals as zero, the diagonal elements will
		 * change. They should be recomputed based on Q_{jj} = -\sum{j: j\neq i} Q_{i, j}
		 */
		/**
		 * This is reset log cond
		 */
		double[][] logCond = resetRateMatrix(tmpMatrix);
		//System.out.println("recompute diagonal q: ");
		//sumOverColumnsOfAMatrix(tmpMatrix);
		
		double[] eigenDotH = new double[4];
		
		/**
		 * Largest eigenvalue
		 */
		eigenDotH[3] = 1;
		
		int i, j;
		
		/**
		 * The largest
		 */
		for(j=2; j>-1; j--)
		{
			eigenDotH[j] = (newEigenValuesOfReconstructedCondMatrix[j] - 1)/newEigenValuesLogOfReconstructedCondMatrix[j];
		}
		
		/**
		 * compute H matrix
		 */

		double[][] hMatrix = null;
		double[][] hEigenValueMatrix = null;
		
		hEigenValueMatrix = convertVectorToDiag(eigenDotH, 4);
		hMatrix = computeMultiplicationOfTwoMatrices(newEigenVectorsOfReconstructedCondMatrix, hEigenValueMatrix);
		hMatrix = computeMultiplicationOfTwoMatrices(hMatrix, newEigenVectorsOfReconstructedCondMatrixInversed);
		
		for(i=3; i>=0; i--)
		{
			for(j=3; j>=0; j--)
				resultEdgeLength += logCond[i][i]*startingFrequencies[j]*hMatrix[j][i];
		}
		
		return (-1)*resultEdgeLength;
	}	
	
	protected ArrayList<double[]> reorganizeEigenResults(ArrayList<double[]> eigenResults)
	{
		ArrayList<double[]> results = new ArrayList<double[]>();
		double[] wr = new double[4];
		double[] wi = new double[4];
		double[] vrr = new double[16];
		double[] vri = new double[16];
		double[] vlr = new double[16];
		double[] vli = new double[16];
		
		int j, k, isComplexPair;
		
		isComplexPair = 0;
		
		for(j=0; j<4; j++)
		{
			/**
			 * check whether the imaginary part of j^th eigenvalue is zero
			 * If yes, read; otherwise, complex
			 */
			if(eigenResults.get(1)[j] == 0)
			{
				/**
				 * real: copy all jth values
				 */
				wr[j] = eigenResults.get(0)[j];
				wi[j] = eigenResults.get(1)[j];
				for(k=0; k<numOfBases; k++)
				{
					vrr[(j)*numOfBases + k] = eigenResults.get(2)[(j)*numOfBases + k];
					vri[(j)*numOfBases + k] = 0.0;	
					vlr[(j)*numOfBases + k] = eigenResults.get(3)[(j)*numOfBases + k];
					vli[(j)*numOfBases + k] = 0.0;
				}
			}
			else
			{
				if(isComplexPair == 0)
				{
					/** 
					 * a + bi
					 * 
					 */
					wr[j] = eigenResults.get(0)[j];
					wi[j] = eigenResults.get(1)[j];
						
					for(k=0; k<numOfBases; k++)
					{
						vrr[(j)*numOfBases + k] = eigenResults.get(2)[(j)*numOfBases + k];
						vri[(j)*numOfBases + k] = eigenResults.get(2)[(j+1)*numOfBases + k];
						vlr[(j)*numOfBases + k] = eigenResults.get(3)[(j)*numOfBases + k];
						vli[(j)*numOfBases + k] = eigenResults.get(3)[(j+1)*numOfBases + k];
					}
					isComplexPair = 1;
				}	
				else
				{
					/**
					 * a-bi
					 */
					wr[j] = eigenResults.get(0)[j];
					wi[j] = eigenResults.get(1)[j];
					for(k=0; k<numOfBases; k++)
					{
						vrr[(j)*numOfBases + k] = eigenResults.get(2)[(j-1)*numOfBases + k];
						vri[(j)*numOfBases + k] = (-1)*eigenResults.get(2)[(j)*numOfBases + k];
						vlr[(j)*numOfBases + k] = eigenResults.get(3)[(j-1)*numOfBases + k];
						vli[(j)*numOfBases + k] = (-1)*eigenResults.get(3)[(j)*numOfBases + k];
					}
					isComplexPair = 0;
				}
				
			}
		}
		
		results.add(wr);
		results.add(wi);
		results.add(vrr);
		results.add(vri);
		results.add(vlr);
		results.add(vli);
		
		return results;
	}	
	
	protected void	writeEdgeLengths() 
	{
		int k, index;
	    int numOfLeafNodes = leafNodesList.size();
	    int numOfIntNodes = intNodesList.size();
	    double[] tmpLens; 
	    
	    String tmpLine, tmpNodeNameShort, tmpNodeName, tmpParentNodeName;
	    	
	    FileWriter myFw;
	    
	    NodeStructure tmpNode; 
	    
	    try
	    {	    		
	    	index = dataPath.lastIndexOf("/");
	    	myFw = new FileWriter(dataPath.substring(0,index) + "/edgeLengths.txt");
	    	
	    	/**
	    	 * Title line
	    	 */
	    	tmpLine = "edge" + "\t" + "forwarding" + "\t" + "backward" + "\n";
	    	myFw.write(tmpLine);
	    	
	    	/**
	    	 * go through internal nodes
	    	 */
	    	for(k=0; k<numOfIntNodes; k++)
	    	{
	    		tmpNodeName = intNodesList.get(k);
	    		tmpNode = phylogeneticTree.get(tmpNodeName);
	    		index = tmpNode.parentNode.indexOf("-");
	    		tmpParentNodeName = tmpNode.parentNode.substring(0, index);
	    		index = tmpNodeName.indexOf("-");
	    		tmpNodeNameShort = tmpNodeName.substring(0, index);
	    		tmpLens = edgeLengths.get(tmpNodeName);
	    		
	    		tmpLine = tmpParentNodeName + " -> " + tmpNodeNameShort + "\t" + tmpLens[0] + "\t" + tmpLens[1] + "\t" + tmpLens[2] + " " + "\n";
	    		
	    		myFw.write(tmpLine);
	    	}
	    	/**	
	    	 * go through leaf nodes
	    	 */	
	    	for(k=0; k<numOfLeafNodes; k++)
	    	{	
	    		tmpLine = "";
	    		tmpNodeName = leafNodesList.get(k);
	    		tmpNode = phylogeneticTree.get(tmpNodeName);
	    		index = tmpNode.parentNode.indexOf("-");
	    		tmpParentNodeName = tmpNode.parentNode.substring(0, index);
	    		index = tmpNodeName.indexOf("-");
	    		tmpNodeNameShort = tmpNodeName.substring(0, index);
	    		tmpLens = edgeLengths.get(tmpNodeName);
	    		
	    		tmpLine = tmpParentNodeName + " -> " + tmpNodeNameShort + "\t" + tmpLens[0] + "\t" + tmpLens[1] + "\t" + tmpLens[2] + "\t" + "\n";
	    		
	    		myFw.write(tmpLine);
	    	}
	    	
	    	myFw.close();
	    }
	    catch(Exception e)
	    {
	    	System.out.println("close file failed in writeDifferenceContainer");
	    	System.exit(-1);
	    }	
	}
		
	void writeTreeNewickFileBestRoot()
	{
		
		String bestRoot = getTheBestRoot();

		TreeMap<String, Double> edgeLengths = obtainEdgeLengthsForARoot(bestRoot);
		
		TreeMap<String, Node> reconstructedTree = shortestPathReconstructedTreesRootingAtANode.get(bestRoot).getRecontructedTree();
		
		writeTreeNewickFileRootingAtANode(bestRoot, edgeLengths, reconstructedTree, true);
	
	}
	protected void writeBestShortestPath(ArrayList<String> selectedRoots)
	{
		for(int j=0; j<selectedRoots.size(); j++)
		{
			String tmpNode = selectedRoots.get(j);
			if(j == 0)
				writeBestShortestPath(tmpNode, true);
			else
				writeBestShortestPath(tmpNode, false);
		}
	}
	
	protected ArrayList<String> obtainSelectedNodes(String inputLine) 
	{
		ArrayList<String> results = new ArrayList<String>();
		StringTokenizer aTokenizer = new StringTokenizer(inputLine, " ,");
		String tmp;
		
		while(aTokenizer.hasMoreTokens())
		{
			tmp = obtainSelectedNodeName(aTokenizer.nextToken());
			if(tmp == null)
			{
				System.err.println("Can't find selected node");
				System.exit(-1);
			}
			
			results.add(tmp);
		}
		
		return results;
	} 
	
	protected String obtainSelectedNodeName(String inputString) 
	{
		String result = null;
		int inputIndex, counter=0;
		
		if(Integer.parseInt(inputString) == counter)
		{
			result = new String(rootNode);
			return result;
		}
		else
		{
			counter++;
			int j;
			
			for(j=0; j<intNodesList.size(); j++)
			{
				if(Integer.parseInt(inputString) == counter)
				{
					result = new String(intNodesList.get(j));
					return result;
				}
				counter++;
			}
			
			for(j=0; j<leafNodesList.size(); j++)
			{
				if(Integer.parseInt(inputString) == counter)
				{
					result = new String(leafNodesList.get(j));
					return result;
				}
				counter++;
			}
		}
		return result;
	}
	
	boolean isInInternalNodesList(String inputNodeName)
	{
		
		for(int j=0; j<intNodesList.size(); j++)
		{
			if(inputNodeName.equals(intNodesList.get(j)))
				return true;
		}
		
		return false;
	}

	boolean isInLeafNodesList(String inputNodeName)
	{
		
		for(int j=0; j<leafNodesList.size(); j++)
		{
			if(inputNodeName.equals(leafNodesList.get(j)))
				return true;
		}
		
		return false;
	}

	String obtainTreeNotationCase1(String inputRootName, 
			                       TreeMap<String, Double> edgeLengths, 
			                       TreeMap<String, Node> recontructedTree,
			                       int whichCase)
	{
		//boolean ;
		String results = "", child, nodeType, lastVisitedNode, newRootChild = null;
		LinkedList<String> nodesVisited = new LinkedList<String>();//first in, last out
		int listSize = 0;
		double anEdgeLen = 0.0, lenRootToChild = 0.0;
		boolean previousNodeIsInternalNode = false, previousPopUpAnInternal = false;
		Node nd = null, nd1 = null;
		NodeStructure ns = null;
		
		nodesVisited.clear();
		
		/**
		 * whichCase
		 * 1. either new root is old root or leaf node
		 * 2. new root's old parent is the old root
		 */
		nd = recontructedTree.get(inputRootName);
		newRootChild = nd.getLeftChild();

		if(whichCase == 1)
		{
			/**
			 * new root is the old root or a leaf node
			 * get the edge from its child
			 */
			if(newRootChild == null)
			{
				newRootChild = nd.getRightChild();
			}
			lenRootToChild = edgeLengths.get(newRootChild).doubleValue();
			/**
			 * new root is the old root or a leaf node
			 * the left bracket and new root will be added first
			 */
			results += "(" + inputRootName.substring(0, inputRootName.indexOf('-')) + ":" + lenRootToChild/2 + ", (";
			
			nd1 = recontructedTree.get(newRootChild);
			nodeType = nd1.getNodeType();
			
			if(nodeType.equals("Leaf"))
			{
				/**
				 * Special case: one edge in the tree
				 * close tree notation and return;
				 */
				results += newRootChild.substring(0, newRootChild.indexOf('-')) + ":" + lenRootToChild/2 + ");";
				
				return results;
			}
			else
			{	
				nd.setLeftVisited(true);
				nd.setRightVisited(true);
				recontructedTree.put(inputRootName, nd);
		
				nodesVisited.add(inputRootName);
				nodesVisited.add(newRootChild);
			
			}
		}
		if(whichCase == 2)
		{
			/**
			 * new root's old parent is old root
			 * new root's parent is the old root
			 * get edge length with left child and use half as edge length 
			 * under parent. 
			 * set root node left and right visited
			 */
			lenRootToChild = edgeLengths.get(rootNode).doubleValue();
			results += "(" + rootNode.substring(0, rootNode.indexOf('-')) + ":" + lenRootToChild/2 + ", (";
			
			if(nd.getRightChild() == null && nd.getRightChild() == null)
			{
				/**
				 * special case: one edge in the tree
				 * close tree notation and return;
				 */
				results += inputRootName.substring(0, inputRootName.indexOf('-')) +  lenRootToChild/2 + ");";
				return results;
			}
			nd.setLeftVisited(true);
			recontructedTree.put(inputRootName, nd);
			child = nd.getLeftChild();
			nodesVisited.add(inputRootName);
			nodesVisited.add(child);
		}
		while(true)
		{
			listSize = nodesVisited.size();
	
			if(listSize == 0)
			{
				/**
				 * has gone through all nodes
				 */
				break;
			}
			
			/**
			 * retrive the last-in internal node
			 */
			lastVisitedNode = nodesVisited.getLast();
			nd = recontructedTree.get(lastVisitedNode);
			nodeType = nd.getNodeType();
			
			if(nodeType != "Leaf")//it is an internal node
			{
				if(!nd.getLeftVisited())
				{
					/**
					 * first visit. put left child into node list
					 */			
					child = nd.getLeftChild();
					nodesVisited.add(child);
					nd.setLeftVisited(true);
					recontructedTree.put(lastVisitedNode, nd);
					previousNodeIsInternalNode = true;
					if(previousPopUpAnInternal)
					{
						results += ", (";
						previousPopUpAnInternal = false;
					}
					else
					{
						results += "(";
					}
				}
			    else if(nd.getLeftVisited() && !nd.getRightVisited())
				{
					/**
					 * has finihsed visiting left child but not right
					 * go ahead visiting right child
					 * get right child and add into nodesVisited
					 */
					//results += ",";
					child = nd.getRightChild();
					nodesVisited.add(child);
					previousNodeIsInternalNode = false;
					nd.setRightVisited(true);
				}
				else if(nd.getLeftVisited() && nd.getRightVisited())
				{
					/**
					 * finished visiting left and right children
					 * get edge length 
					 * if this is the internal node connected with root
					 * use half length
					 */
					
					if(lastVisitedNode.equals(inputRootName))
					{
						/**
						 * finished iteration
						 * close tree notation
						 */
						//anEdgeLen = edgeLengths.get(rootNode);
						
						results += "):" + lenRootToChild/2+");" + "\n";
						nodesVisited.removeLast();
						if(nodesVisited.size() != 0)
						{
							/**
							 * unexpected 
							 */
							System.err.println("Unexpected case 5.");
							System.exit(-1);	
						}
							break;
					}
					else
					{
						anEdgeLen = edgeLengths.get(lastVisitedNode);
						results += "):" + anEdgeLen;
					}
					/**
					 * pop up last added internal node
					 */
					nodesVisited.removeLast();
					previousPopUpAnInternal = true;
				}
				else
				{
					/**
					 * unexpected 
					 */
					System.err.println("Unexpected case 6.");
					System.exit(-1);
				}
			}
			else
			{
				/**
				 * leaf node
				 * results added leaf node name and edge length
				 */
				anEdgeLen = edgeLengths.get(lastVisitedNode).doubleValue();
				results += lastVisitedNode.substring(0, lastVisitedNode.indexOf('-')) + ":" + anEdgeLen;
				if(previousNodeIsInternalNode)
				{
					results += ",";
				}
				nodesVisited.removeLast();
				previousNodeIsInternalNode = false;
			}
		}
		
		results += "\n";
		
		return results;
	}
	
	String obtaineTreeNotationCase2(String inputRootName, 
			                        TreeMap<String, Double> edgeLengths, 
			                        TreeMap<String, Node> reconstructedTree 
			                        )
	{
		//boolean ;
		String results = "", child, parent, nodeType, lastVisitedNode;
		LinkedList<String> nodesVisited = new LinkedList<String>();//first in, last out
		int listSize = 0;
		double anEdgeLen = 0.0;
		boolean previousNodeIsInternalNode = false, previousPopUpAnInternal = false;;
		Node nd = null;
		
		nodesVisited.clear();
		
		/**
		 * In this case, the new root's old parent is also its parent
		 * Thus, the algorithm will go through the path starting from 
		 * parent first; then go through two children
		 * The tricky is that it is necessary to check whether both of the "parent"
		 * and "children" paths have been gone through when meeting new root.
		 */
		
		nd = reconstructedTree.get(inputRootName);
		parent = nd.getParent();
		results += "(";
		/**
		 * use "leftVisited" indicated has been visiting "parent" path.
		 * This is a special treatment. It will not be expected to use to other 
		 * internal node. Thus, it is necessary to check whether the internal node is 
		 * the new root. 
		 */
		reconstructedTree.put(inputRootName, nd);
		
		nodesVisited.add(inputRootName);
		nodesVisited.add(parent);
		
		while(true)
		{
			listSize = nodesVisited.size();
			
			if(listSize == 0)
			{
				/**
				 * has gone through all nodes
				 */
				break;
			}
			
			/**
			 * retrive the last-in internal node
			 */
			lastVisitedNode = nodesVisited.getLast();
			nd = reconstructedTree.get(lastVisitedNode);
			nodeType = nd.getNodeType();
			
			if(nodeType != "Leaf")//it is an internal node
			{
				if(!nd.getLeftVisited())
				{
					child = nd.getLeftChild();
					nodesVisited.add(child);
					nd.setLeftVisited(true);
					reconstructedTree.put(lastVisitedNode, nd);
					previousNodeIsInternalNode = true;
					if(previousPopUpAnInternal || lastVisitedNode.equals(inputRootName))
					{
						results += ", (";
						previousPopUpAnInternal = false;
					}
					else
					{
						results += "(";
					}
				}
				else if(nd.getLeftVisited() && !nd.getRightVisited())
				{
					/**
					 * if it is a normal internal node and it 
					 * has finihsed visiting left child but not right
					 * go ahead visiting right child
					 * get right child and add into nodesVisited
					 */
					child = nd.getRightChild();
					nodesVisited.add(child);
					nd.setRightVisited(true);
					reconstructedTree.put(lastVisitedNode, nd);
					previousNodeIsInternalNode = false;
					nd.setRightVisited(true);
					
				}
				else if(nd.getLeftVisited() && nd.getRightVisited())
				{
					if((!lastVisitedNode.equals(inputRootName) && !lastVisitedNode.equals(parent)))
					{
						/**
						 * it is a normal internal node
						 * finished visiting left and right children
						 * get edge length 
						 * if this is the internal node connected with root
						 * use half length
						 */
						anEdgeLen = edgeLengths.get(lastVisitedNode);
						results += "):" + anEdgeLen;
						/**
						 * pop up last added internal node
						 */
						nodesVisited.removeLast();
						previousPopUpAnInternal = true;
					}
					else if(lastVisitedNode.equals(parent))
					{
						/**
						 * it is new root's parent and has finished traversing the path through its 
						 * parent. Close brackets plus half edge length between old parent
						 * and new root should be used. 
						 */
						anEdgeLen = edgeLengths.get(lastVisitedNode).doubleValue();
						results += "):" + anEdgeLen/2 + ", ";
						nodesVisited.removeLast();
						previousNodeIsInternalNode = true;
					}
					else if(lastVisitedNode.equals(inputRootName))
					{
						/**
						 * it is new root and it indicates that all 
						 * paths have gone through. it is time to close.  
						 */
						anEdgeLen = edgeLengths.get(parent).doubleValue();
						results += "):" + anEdgeLen/2 + ");" + "\n";
						nodesVisited.removeLast();
						if(nodesVisited.size() != 0)
						{
							System.err.println("Unexpected case 2.");
							System.exit(-1);
						}
						break;
					}
					else
					{
						System.err.println("Unexpected case 3.");
						System.exit(-1);
					}
				}
				else 
				{
					/**
					 * unexpected case terminates
					 */
					System.err.println("Unexpected case 4.");
					System.exit(-1);					
				}
			}
			else
			{
				/**
				 * leaf node
				 * results added leaf node name and edge length
				 */
				anEdgeLen = edgeLengths.get(lastVisitedNode).doubleValue();
				results += lastVisitedNode + ":" + anEdgeLen;
				if(previousNodeIsInternalNode || previousPopUpAnInternal)
				{
					results += ",";
				}
				nodesVisited.removeLast();
				previousNodeIsInternalNode = false;
			}
		}
		
		results += "\n";
		
		return results;
	}
	
	void writeBestNSGTRModelParametersAllNodes()
	{
		int j;
		String nodeName; 
		
		writeBestNSGTRModelParameters(rootNode, true);
		
		for(j=0; j<intNodesList.size(); j++)
		{
			nodeName = intNodesList.get(j);
			writeBestNSGTRModelParameters(nodeName, false);
		}
		
		for(j=0; j<leafNodesList.size(); j++)
		{
			nodeName = leafNodesList.get(j);
			writeBestNSGTRModelParameters(nodeName, false);
		}

	}
	
	void writeBestNSGTRModelParameters(ArrayList<String> selectedRoots)
	{
		int listSize = selectedRoots.size();
		int j;
		String nodeName;
		
		for(j=0; j<listSize; j++)
		{
			nodeName = selectedRoots.get(j);
			if(j==0)
			{
				writeBestNSGTRModelParameters(nodeName, true);
			}
			else
			{
				writeBestNSGTRModelParameters(nodeName, false);
			}
		}
	}
	
	
	void writeBestNSGTRModelParameters(String bestRoot, boolean isFirst)
	{
		TreeMap<String, double[][]> bestFitConds = shortestPathReconstructedTreesRootingAtANode.get(bestRoot).getBestFitNonStationaryConds();
		int index;
    	FileWriter myFw = null;
    	String tmpLine = "", tmpString;
    		
    	//tmpLine += "Rooting At " + bestRoot + "\n";
    	
		try
    	{

    		index = dataPath.lastIndexOf("/");
    		
    		if(!isFirst)
    		{
    			/**
    			 * append
    			 */
    			myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitNSGTRParameters.txt", true);
    		}
    		else 
    			myFw = new FileWriter(dataPath.substring(0,index) + "/bestFitNSGTRParameters.txt");
    		
    		tmpLine += "Rooting At Node " + bestRoot.substring(0, bestRoot.indexOf('-')) + "\n\n";
    		
    		tmpString= getNSGTRParametersForInternalNodes(bestFitConds, bestRoot);
    		tmpLine += tmpString;
    		tmpLine += getNSGTRParametersForLeafNodes(bestFitConds, bestRoot);
    		
    		myFw.write(tmpLine);
    		myFw.close();
    	}
		catch (Exception e) 
		{
			System.err.println(e);
			System.err.println("error in writeBestNSGTRModelParameters");
			System.exit(-1);
		}

	}
	
	String getNSGTRParametersForInternalNodes(TreeMap<String, double[][]> bestFitConds, String bestRoot)
	{
		String results = "", nodeName, oldParent;
		double[][] conds = null;
		containerNSGTRParameters aCon;
		
		ContainerShortestPathSearchResults shortPathResults = shortestPathReconstructedTreesRootingAtANode.get(bestRoot);
		TreeMap<String, Node> reconstructedTreeAtBestRoot = shortPathResults.getRecontructedTree();
		
		int j=0;
		
		for(j=0; j<intNodesList.size(); j++)
		{
			nodeName = intNodesList.get(j);
			oldParent = phylogeneticTree.get(nodeName).getParentNode();
			
			if(nodeName.equals(bestRoot) || !oldParent.equals(reconstructedTreeAtBestRoot.get(nodeName).getParent()))
			{
				/**
				 * three cases:
				 * 1. best root with old root as parent
				 * 2. best root with old parents: direction from bestRoot to parent. but parent holds the estimates
				 * 3. An internal node in an reverse direction
				 */
				results += "(" + nodeName.substring(0,nodeName.indexOf('-')) + ", " +  oldParent.substring(0, oldParent.indexOf('-')) + "): \n";
				conds = bestFitConds.get(oldParent);	
			}
			else
			{
				results += "(" + oldParent.substring(0,oldParent.indexOf('-')) + ", " +  nodeName.substring(0, nodeName.indexOf('-')) + "): \n";
				conds = bestFitConds.get(nodeName);
			}
			aCon = getNSGTRParemeters(conds);
			results += aCon.toString();
			results += "\n";
		}
		
		return results;
	}
	
	String getNSGTRParametersForLeafNodes(TreeMap<String, double[][]> bestFitConds, String bestRoot)
	{
		String results = "", nodeName, oldParent;
		double[][] conds = null;
		containerNSGTRParameters aCon;
		int j=0;
		
		for(j=0; j<leafNodesList.size(); j++)
		{
			nodeName = leafNodesList.get(j);
			oldParent = phylogeneticTree.get(nodeName).getParentNode();
			
			if(nodeName.equals(bestRoot))
			{
				results += "(" + nodeName.substring(0, nodeName.indexOf('-')) + ", " + oldParent.substring(0, oldParent.indexOf('-')) + "): \n";
				conds = bestFitConds.get(oldParent);
			}
			else
			{
				results += "(" + oldParent.substring(0, oldParent.indexOf('-')) + ", " + nodeName.substring(0, nodeName.indexOf('-')) + "): \n";
				conds = bestFitConds.get(nodeName);
			}
			
			aCon = getNSGTRParemeters(conds);	
			results += aCon.toString();
			results += "\n";
		}
		
		return results;
	}
	
	public void writeNewickFile()
	{
		String bestRoot = null;
		
		bestRoot = getTheBestRoot();	
		TreeMap<String, Double> edgeLengths;
		TreeMap<String, Node> reconstructedTree;
		ArrayList<String> selectedRoots;
		
		if(defaultPosition == true)
		{
			edgeLengths = obtainEdgeLengthsForARoot(bestRoot);
			reconstructedTree = shortestPathReconstructedTreesRootingAtANode.get(bestRoot).getRecontructedTree();
			writeTreeNewickFileRootingAtANode(bestRoot, edgeLengths, reconstructedTree, true);			
		}
		else
		{
			selectedRoots = obtainSelectedNodes(selectedRootingPosition);
			writeTreeNewickFileForSelectedRoots(selectedRoots);
		}
	}

	void writeTreeNewickFileForSelectedRoots(ArrayList<String> selectedRoots)
	{
		int j;
		String newRootName;
		
		TreeMap<String, Double> edgeLengths;
		TreeMap<String, Node> reconstructedTree;
		
		for(j=0; j<selectedRoots.size(); j++)
		{
			newRootName = selectedRoots.get(j);
			edgeLengths = obtainEdgeLengthsForARoot(newRootName);
			reconstructedTree = shortestPathReconstructedTreesRootingAtANode.get(newRootName).getRecontructedTree();
			
			if(j==0)
				writeTreeNewickFileRootingAtANode(newRootName, edgeLengths, reconstructedTree, true);
			else
				writeTreeNewickFileRootingAtANode(newRootName, edgeLengths, reconstructedTree, false);	
		}
	}
	
	TreeMap<String, Double> obtainEdgeLengthsForARoot(String inputRootName)
	{
		TreeMap<String, Double> results = new TreeMap<String, Double>();
		TreeMap<String, double[][]> estOfJoints = shortestPathReconstructedTreesRootingAtANode.get(inputRootName).getBestFitBHJoints();
		int j;
		String nodeName; 
		double[][] joints = null;
		double edgeLength;
		
		if(!inputRootName.equals(rootNode))
		{
			joints = estOfJoints.get(rootNode);
			edgeLength = calculateNonstationaryEdgeLength(joints);
			results.put(rootNode, new Double(edgeLength));
		}
		
		for(j=0; j<intNodesList.size(); j++)
		{
			nodeName = intNodesList.get(j);
			if(!nodeName.equals(inputRootName))
			{
				joints = estOfJoints.get(nodeName);
			
				edgeLength = calculateNonstationaryEdgeLength(joints);
				results.put(nodeName, new Double(edgeLength));
			}
		}
	
		for(j=0; j<leafNodesList.size(); j++)
		{
			nodeName = leafNodesList.get(j);
			if(!nodeName.equals(inputRootName))
			{
				joints = estOfJoints.get(nodeName);
				edgeLength = calculateNonstationaryEdgeLength(joints);
				results.put(nodeName, new Double(edgeLength));
			}
		}

		return results;
	}
	/**
	 * 
	 * This function uses the edge lengths along edges under specific root node. 
	 * 
	 * Starting from the root, a forwarding traversal algorithm will perform 
	 * for travelling through all edges.
	 * 
	 */
	void writeTreeNewickFileRootingAtANode(String inputRootName, 
			                               TreeMap<String, Double> edgeLengths,
			                               TreeMap<String, Node> reconstructedTree,
			                               boolean isFirst
			                               )
	{
	
		
		int index;
		String tmpLine = "";
    	FileWriter myFw = null;

		try
    	{

    		index = dataPath.lastIndexOf("/");
    		
    		if(!isFirst)
    		{
    			/**
    			 * append
    			 */
    			myFw = new FileWriter(dataPath.substring(0,index) + "/tree.txt", true);
    		}
    		else 
    			myFw = new FileWriter(dataPath.substring(0,index) + "/tree.txt");
    		
    		tmpLine += "Rooting at node "  + inputRootName + ": \n";
			myFw.write(tmpLine);
			
			tmpLine = "";//preparing writing the tree
			
			if(inputRootName.equals(rootNode) || isInLeafNodesList(inputRootName))
				tmpLine = obtainTreeNotationCase1(inputRootName, edgeLengths, reconstructedTree, 1);
			else if(rootNode.equals((String)phylogeneticTree.get(inputRootName).getParentNode()))
				tmpLine = obtainTreeNotationCase1(inputRootName, edgeLengths, reconstructedTree, 2);
			else 	
				tmpLine = obtaineTreeNotationCase2(inputRootName, edgeLengths, reconstructedTree);
			
			myFw.write(tmpLine);
			myFw.close();
		}
		catch (Exception e) 
		{
    		System.out.println("wrte file failed in writeTreeNewickFileRootingAtANode");
    		System.exit(-1);
		}
	}
	double[] getAZeroMatrix(int inputSize)
	{
		double[] results = new double[inputSize*inputSize];
		
		for(int j=0; j<inputSize*inputSize; j++)
		{
				results[j]=0.0;
		}
		
		return results;
	}
}	

	