/*
 * estimatePermutations.c
 *
 *  Created on: 2011-08-04
 *      Author: zoul
 */


#include <jni.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include"estimatePermutations.h"

typedef struct { double re,im; } Complex;

Complex* createComplexArray(jdouble*, jdouble*, jint, jint);


/*
 * Class:     estimatePermutations
 * Method:    callingDGEMM
 * Signature: (CCIIID[DI[DID[DI)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingDGEMM(JNIEnv*      jenv,
		                                                      jobject      jobj,
		                                                      jchar        jtransa,
		                                                      jchar        jtransb,
		                                                      jint         m,
		                                                      jint         n,
		                                                      jint         k,
		                                                      jdouble      alpha,
		                                                      jdoubleArray ja,
		                                                      jint         lda,
		                                                      jdoubleArray jb,
		                                                      jint         ldb,
		                                                      jdouble      beta,
		                                                      jdoubleArray jc,
		                                                      jint         ldc)
{

	jdouble *a;
	jdouble *b;
	jdouble *c;
	char *transa, *transb;

	transa = (char*)malloc(2*sizeof(char));
	transb = (char*)malloc(2*sizeof(char));
	transa[0] = jtransa;
	transa[1] ='\0';
	transb[0] = jtransb;
	transb[1] = '\0';
	

	/**
	 * extract java arrays
	 */
	a = (*jenv)->GetDoubleArrayElements(jenv, ja, 0);
	b = (*jenv)->GetDoubleArrayElements(jenv, jb, 0);
	c = (*jenv)->GetDoubleArrayElements(jenv, jc, 0);
	
	/**
	 * calling dgemm
	 */
	dgemm_(transa, transb, &m, &n, &k, &alpha, a, &lda, b, &ldb, &beta, c, &ldc);
	
	/**
	 * release memories
	 */
	(*jenv)->ReleaseDoubleArrayElements(jenv, ja, a, 0);
	(*jenv)->ReleaseDoubleArrayElements(jenv, jb, b, 0);
	(*jenv)->ReleaseDoubleArrayElements(jenv, jc, c, 0);
	
	free(transa);
	free(transb);

}

/*
 * Class:     estimatePermutations
 * Method:    callingDGEEV
 * Signature: (CCI[DI[D[D[DI[DI[DII)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingDGEEV(JNIEnv * jenv,
		                                                      jobject  	 jobj,
		                                                      jchar    jjobvl,
		                                                      jchar    jjobvr,
		                                                      jint     jn,
		                                                      jdoubleArray ja,
		                                                      jint         jlda,
		                                                      jdoubleArray jwr,
		                                                      jdoubleArray jwi,
		                                                      jdoubleArray jvl,
		                                                      jint         jldvl,
		                                                      jdoubleArray jvr,
		                                                      jint         jldvr,
		                                                      jdoubleArray jwork,
		                                                      jint         jlwork,
		                                                      jint         jinfo)
{
    jdouble      *a;
    jdouble      *wr;
    jdouble      *wi;
    jdouble      *vl;
    jdouble      *vr;
    jdouble      *work;

    char *jobvl, *jobvr;
    jobvl = (char*)malloc(2*sizeof(char));
    jobvl[0] = jjobvl;
    jobvl[1] ='\0';

    jobvr = (char*)malloc(2*sizeof(char));
    jobvr[0] = jjobvr;
    jobvr[1] ='\0';

    /**
     * extract java double arrays
     */
    a    = (*jenv)->GetDoubleArrayElements(jenv, ja, 0);
    wr   = (*jenv)->GetDoubleArrayElements(jenv, jwr, 0);
    wi   = (*jenv)->GetDoubleArrayElements(jenv, jwi, 0);
    vl   = (*jenv)->GetDoubleArrayElements(jenv, jvl, 0);
    vr   = (*jenv)->GetDoubleArrayElements(jenv, jvr, 0);
    work = (*jenv)->GetDoubleArrayElements(jenv, jwork, 0);

    /**
     * calling function f02ebf
     */
    dgeev_(&jjobvl, &jjobvr, &jn, a, &jlda, wr, wi, vl, &jldvl, vr, &jldvr, work, &jlwork, &jinfo);

    /**
     * release memory
     */
    (*jenv)->ReleaseDoubleArrayElements(jenv, ja, a, 0);
    (*jenv)->ReleaseDoubleArrayElements(jenv, jwr, wr, 0);
    (*jenv)->ReleaseDoubleArrayElements(jenv, jwi, wi, 0);
    (*jenv)->ReleaseDoubleArrayElements(jenv, jvl, vl, 0);
    (*jenv)->ReleaseDoubleArrayElements(jenv, jvr, vr, 0);
    (*jenv)->ReleaseDoubleArrayElements(jenv, jwork, work, 0);

    free(jobvl);
    free(jobvr);

}

/*
 * Class:     estimatePermutations
 * Method:    callingDGETRF
 * Signature: (II[DI[II)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingDGETRF(JNIEnv *     jenv,
															   jobject      jobj,
															   jint         jm,
															   jint         jn,
															   jdoubleArray ja,
															   jint         jlda,
															   jintArray    jipiv,
															   jint         jinfo)
{
	  jdouble *a;
	  jint    *ipiv;

	  /**
	   * extract arrays
	   **/
	  a = (*jenv)->GetDoubleArrayElements(jenv, ja, 0);
	  ipiv = (*jenv)->GetIntArrayElements(jenv, jipiv, 0);

	  /**
	   * calling f07adf
	   **/
	  dgetrf_(&jm, &jn, a, &jlda, ipiv, &jinfo);
	  /**
	   * release memory
	   **/
	  (*jenv)->ReleaseDoubleArrayElements(jenv, ja, a, 0);
	  (*jenv)->ReleaseIntArrayElements(jenv, jipiv, ipiv, 0);

}

/*
 * Class:     estimatePermutations
 * Method:    callingDSYEV
 * Signature: (CCI[DI[D[DII)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingDSYEV(JNIEnv *     jenv,
		                                                      jobject      jobj,
		                                                      jchar        jjobz,
		                                                      jchar        juplo,
		                                                      jint         jn,
		                                                      jdoubleArray ja,
		                                                      jint         jlda,
		                                                      jdoubleArray jw,
		                                                      jdoubleArray jwork,
		                                                      jint         jlwork,
		                                                      jint         jinfo)
{
	  jdouble* a;
	  jdouble* w;
	  jdouble* work;

	  char *job, *uplo;

	  job = (char*)malloc(2*sizeof(char));
	  uplo = (char*)malloc(2*sizeof(char));
	  job[0] = jjobz;
	  job[1] ='\0';
	  uplo[0] = juplo;
	  uplo[1] = '\0';

	  /**
	   * extract arrays
	   **/
	  a = (*jenv)->GetDoubleArrayElements(jenv, ja, 0);
	  w = (*jenv)->GetDoubleArrayElements(jenv, jw, 0);
	  work = (*jenv)->GetDoubleArrayElements(jenv, jwork, 0);

	  /**
	   * calling F02FAF
	   **/
	  dsyev_(job, uplo, &jn, a, &jlda, w, work, &jlwork, &jinfo);

	  /**
	   * release memories
	   */
	  (*jenv)->ReleaseDoubleArrayElements(jenv, ja, a, 0);
	  (*jenv)->ReleaseDoubleArrayElements(jenv, jw, w, 0);
	  (*jenv)->ReleaseDoubleArrayElements(jenv, jwork, work, 0);

	  free(job);
	  free(uplo);

}

/*
 * Class:     estimatePermutations
 * Method:    callingDGETRI
 * Signature: (I[DI[I[DII)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingDGETRI(JNIEnv *     jenv,
		                                                       jobject      jobj,
		                                                       jint         jn,
		                                                       jdoubleArray ja,
		                                                       jint         jlda,
		                                                       jintArray    jipiv,
		                                                       jdoubleArray jwork,
		                                                       jint         jlwork,
		                                                       jint         jinfo)
{
	  jdouble *a, *work;
	  jint* ipiv;
	  /**
	   * allocate memory
	   **/
	  work = (jdouble*)malloc(jlwork*sizeof(jdouble));

	  /**
	   * extract arrays
	   **/
	  a = (*jenv)->GetDoubleArrayElements(jenv, ja, 0);
	  ipiv = (*jenv)->GetIntArrayElements(jenv, jipiv, 0);

	  /**
	   * calling f07ajf
	   **/
	  dgetri_(&jn, a, &jlda, ipiv, work, &jlwork, &jinfo);

	  /**
	   * release memory
	   **/
	  (*jenv)->ReleaseDoubleArrayElements(jenv, ja, a, 0);
	  (*jenv)->ReleaseIntArrayElements(jenv, jipiv, ipiv, 0);

	  free(work);

}

/*
 * Class:     estimatePermutations
 * Method:    callingZGETRF
 * Signature: (II[D[DI[II)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingZGETRF(JNIEnv *     jenv,
		                                                       jobject      jobj,
		                                                       jint         jm,
		                                                       jint         jn,
		                                                       jdoubleArray jar,
		                                                       jdoubleArray jai,
		                                                       jint         jlda,
		                                                       jintArray    jipiv,
		                                                       jint         jinfo)
{
    jdouble *ar;
    jdouble *ai;
    jint    *ipiv;

    int j, arraySize;

    Complex *complexNumberArray;

    /**
     * extract arrays
     */
    ar   = (*jenv)->GetDoubleArrayElements(jenv, jar, 0);
    ai   = (*jenv)->GetDoubleArrayElements(jenv, jai, 0);
    ipiv = (*jenv)->GetIntArrayElements(jenv, jipiv, 0);

    /**
     * obtain complex number array
     */
    complexNumberArray = createComplexArray(ar, ai, jm, jn);

    /**
     * calling f07arf
     */
    zgetrf_(&jm, &jn, (Complex*)complexNumberArray, &jlda, ipiv, &jinfo);

    arraySize = jm*jn;
    for(j=0; j<arraySize; j++)
    {
    	ar[j] = complexNumberArray[j].re;
    	ai[j] = complexNumberArray[j].im;
    }

    /**
     * release memory
     */
    (*jenv)->ReleaseDoubleArrayElements(jenv, jar, ar, 0);
    (*jenv)->ReleaseDoubleArrayElements(jenv, jai, ai, 0);
    (*jenv)->ReleaseIntArrayElements(jenv, jipiv, ipiv, 0);

    free(complexNumberArray);
}

/*
 * Class:     estimatePermutations
 * Method:    callingZGETRI
 * Signature: (I[D[DI[III)V
 */
JNIEXPORT void JNICALL Java_estimatePermutations_callingZGETRI(JNIEnv *     jenv,
		                                                       jobject      jobj,
		                                                       jint         jn,
		                                                       jdoubleArray jar,
		                                                       jdoubleArray jai,
		                                                       jint         jlda,
		                                                       jintArray    jipiv,
		                                                       jint         jlwork,
		                                                       jint         jinfo)
{
	jdouble *ar;
	jdouble *ai;
	jint    *ipiv;
	Complex *a;
	Complex *work;

	int j, arraySize;

	/**
	 * extract java arrays
	 */
	ar = (*jenv)->GetDoubleArrayElements(jenv, jar, 0);
	ai = (*jenv)->GetDoubleArrayElements(jenv, jai, 0);
	ipiv = (*jenv)->GetIntArrayElements(jenv, jipiv, 0);

	/**
	 * obtain the complex array for a
	 */
	a = createComplexArray(ar, ai, jn, jn);
	/**
	 * allocate memory for WORK
	 */
	work = (Complex*)malloc(jlwork*sizeof(Complex));

	/**
	 * calling F07AWF
	 */
	zgetri_(&jn, a, &jlda, ipiv, work, &jlwork, &jinfo);

	/**
	 * copy values to ar and ai
	 */
	arraySize = jn*jlda;

	for(j=0; j<arraySize; j++)
	{
		ar[j] = (jdouble)a[j].re;
		ai[j] = (jdouble)a[j].im;
	}

	/**
	 * release memories
	 */
	(*jenv)->ReleaseDoubleArrayElements(jenv, jar, ar, 0);
	(*jenv)->ReleaseDoubleArrayElements(jenv, jai, ai, 0);
	(*jenv)->ReleaseIntArrayElements(jenv, jipiv, ipiv, 0);

	free(a);
	free(work);


}

/*Complex* createComplexArray(jdouble*, jdouble*, jint, jint);*/
Complex* createComplexArray(jdouble* jrealPartArray,
		                    jdouble* jimaginaryPartArray,
		                    jint     jrowSize,
		                    jint     jcolumnSize)
{
	Complex * results;

	results = (Complex*)malloc(jrowSize*jcolumnSize*sizeof(Complex));

	int j, k;

	for(j=0; j<jrowSize*jcolumnSize; j++)
	{
		results[j].re = jrealPartArray[j];
		results[j].im = jimaginaryPartArray[j];

	}

	return results;
}
